#==============================================================================
# �package      DE_EPAGES::Design::UI::Shop
# �state        private
#==============================================================================
package DE_EPAGES::Design::UI::Shop;
use base DE_EPAGES::Shop::UI::Shop;

use strict;
use DE_EPAGES::Core::API::Error qw( Error ExistsError GetError );
use DE_EPAGES::Core::API::File qw( DeleteFile fsplit );
use DE_EPAGES::Core::API::Image qw( ResizeImage TestImage );
use DE_EPAGES::Core::API::PerlTools qw( ElementOf );
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Core::API::TestParameter qw( TestInteger );
use DE_EPAGES::Core::API::Url qw (UrlEncodeValue);
use DE_EPAGES::Design::API::Style qw( AddStyle DeleteTemporaryStyles ExistsStyle Style );
use DE_EPAGES::Design::API::StyleGroup qw( GetSubStyleGroups );
use DE_EPAGES::Object::API::Factory qw( Transaction LoadRootObject );
use DE_EPAGES::Presentation::API::Upload qw ( Upload );
use DE_EPAGES::Shop::API::JWT qw ( AddEditorTokenToURL );

#==============================================================================
# Event handlers
#==============================================================================

sub SetCookiePreviewStyle {
    my $self = shift;
    my ($Servlet) = @_;

    my ($PreviewStyle) = $Servlet->form->value('PreviewStyle');
    $Servlet->cookies('PreviewStyle', $PreviewStyle);
}

sub RemoveCookiePreviewStyle {
    my $self = shift;
    my ($Servlet) = @_;

    $Servlet->cookies('PreviewStyle', undef);
}

sub ResetStyle {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;
    my $StyleTemplateID = $Servlet->params('StyleTemplateID');

    return unless defined $StyleTemplateID;
    TestInteger('StyleTemplateID', $StyleTemplateID);
    my $StyleTemplate = LoadRootObject()->loadObject($StyleTemplateID, 'StyleTemplate');
    if (defined $StyleTemplate) {
        if ($Servlet->params('ACCEPT') or not ExistsStyle ($Shop, $StyleTemplate) ) {
            $Servlet->vars('RESET',0);
        }
        else {
            $Servlet->vars('RESET',1);
            return;
        }

        Transaction ( sub {

            if( ExistsStyle ($Shop, $StyleTemplate) ) {
                my $Style = Style($Shop, $StyleTemplate);
                $Style->duplicate;
                $Style->reset($StyleTemplate);
            }

            #override/add Style
            AddStyle($Shop, $StyleTemplate);
        } );
    }
    return;
}

sub DuplicateStyles {
    my $self = shift;
    my ( $Servlet ) = @_;

    my $Object = $Servlet->object();
    my $Form = $Servlet->form;
    my $hValues = $Form->form($Object, 'ListedObjects');
    my @ListObjectIDs = map { $_->{'ListObjectID'} } @{$hValues->{'ListObjectIDs'}};
    my $aStyles = $Object->loadObjects(\@ListObjectIDs, 'Style');

    # copy styles
    foreach my $Style ( @$aStyles ) {
        # don't allow to save the object of someone else
        $Servlet->testPermission($Style, 'Save');
        $Style->duplicate;
    }

    return;
}

sub DeleteListedStyles {
    my $self = shift;
    my ( $Servlet ) = @_;
    my $Shop = $Servlet->object();

    my $Form = $Servlet->form;
    my $hValues = $Form->form($Shop, 'ListedObjects');
    my @ListObjectIDs = map { $_->{'ListObjectID'} } @{$hValues->{'ListObjectIDs'}};

    # if the list contains the DefaultStyle, it is not allowed to delete the Styles
    $Form->addFormError({
        Reason => 'DontDeleteDefaultStyle',
        Name => 'ListObjectID'
    }) if ElementOf($Shop->get('DefaultStyle')->id, @ListObjectIDs);

    $Form->executeFormError;
    $self->DeleteListedObjects( $Servlet );
    return;
}

sub SaveStylePreview {
    my $self = shift;
    my ($Servlet) = @_;
    my $Form = $Servlet->form;

    my $StyleID = $Form->value('StyleID');
    my $Shop = $Servlet->object;
    my $Style = $Shop->loadObject( $StyleID, 'Style' );

    my $FileName = $Form->value('NewFile');
    my (undef, undef, $Extension) = fsplit($FileName);
    my $FilePath = $Style->get('PublicPath');
    my $OrgImage = $FilePath . '/' . $FileName;
    my $ImageSmall = $FilePath . '/stylepreview' . $Extension;
    my $StylePreviewImage = $Style->get('WebPath').'/stylepreview' . $Extension;

    # write post data into file
    my (undef, $hResponseContent) = Upload($Shop, $Form, {'FormName' => 'SaveFiles', 'InputName' => 'NewFile', 'Index' => 0, 'Path' => $FilePath . '/'});
    $Servlet->vars('ResponseContent', $hResponseContent);

    # check if file is valid image
    eval {
        TestImage($OrgImage);
    };
    if (ExistsError()) {
        DeleteFile($OrgImage);
        my $Error = GetError();
        Error($Error) unless ($Error->code eq 'INVALID_IMAGE_FILE' || $Error->code eq 'IMAGE_EXCEEDS_MEGAPIXELS');
        $Servlet->form->executeFormError({
            'Reason'=> $Error->code,
            'Form'  => 'SaveFiles',
            'Name'  => 'NewFile',
        });
        return;
    }

    # remove original preview image
    DeleteFile($FilePath . '/' . $Style->get('StylePreview')) if defined $Style->get('StylePreview');

    eval {
        ResizeImage($OrgImage, $ImageSmall, 174); ## no critic qw(ProhibitMagicNumbers)
        $Style->set({'StylePreview' => 'stylepreview' . $Extension});
        DeleteFile($OrgImage);
    };
    if ( ExistsError() ){
        my $Error = GetError();
        if (ElementOf ($Error->code, 'FILE_NOT_FOUND', 'IMAGE_MAGICK', 'UNKNOWN_IMAGE_SIZE', 'IMAGE_MAGICK')){
            Warning('ResizeImageFailed', {
                'Code'      => $Error->code,
                'FileName'  => $OrgImage,
                'StyleID'   => $Style->id,
            });
        } else {
            Error($Error);
        }
    }

    $Servlet->vars('ResponseContent', {
        'uploadedFile' => $StylePreviewImage,
    });

    return;
}

sub JSONGetStyleGroups {
    my $self = shift;
    my ($Servlet) = @_;

    my $aStyleGroups = LoadRootObject()->get('StyleGroups');
    my $ahInfo = [];
    foreach my $group (@$aStyleGroups) {
        push @$ahInfo, GetSubStyleGroups($group);
    }
    $Servlet->writeContentOutput($ahInfo);
    $Servlet->params('ExitAfterEvent', 1);
    return;
}

sub _StyleTemplateLoop ($$) {
    my ($Map, $WebPath) = @_; # StyleGroupMap or StyleTemplateVariation

    my $StyleTemplate = $Map->get('StyleTemplate');
    my $Title = $StyleTemplate->get('NameOrAlias');
    if ($StyleTemplate->parent->parent->instanceOf('StyleTemplate')) {
        $Title = $StyleTemplate->parent->parent->get('NameOrAlias') . " ($Title)";
    }
    return {
        objectId        => $Map->id,
        title           => $Title,
        styleDir        => $WebPath . '/' . $StyleTemplate->get('StyleDir'),
        styleTemplateId => $StyleTemplate->id,
        customizeLevel  => $StyleTemplate->get('CustomizeLevel')
    };
}

sub byPosition {
    return $a->get('Position') <=> $b->get('Position');
}

sub JSONGetStyleGroupMaps {
    my $self = shift;
    my ($Servlet) = @_;

    my $System = LoadRootObject();
    my $StyleGroup = $System->loadObject($Servlet->params('StyleGroupID'), 'StyleGroup');

    my $WebPath = $System->get('StoreRoot');
    my @Maps = ();
    my $aStyleTemplateIDs = [];
    while (defined $StyleGroup) {
        foreach my $StyleGroupMap (sort byPosition @{$StyleGroup->get('StyleGroupMaps')}) {
            my $hMap = _StyleTemplateLoop($StyleGroupMap, $WebPath);
            unless (ElementOf($hMap->{'styleTemplateId'}, @$aStyleTemplateIDs)) {
                my @Variations = map { _StyleTemplateLoop($_, $WebPath) }
                                 sort byPosition
                                 @{$StyleGroupMap->get('StyleTemplateVariations')};
                unshift @Variations, $hMap;
                push @Maps, {
                    %$hMap,
                    'variations' => \@Variations
                };
                push @$aStyleTemplateIDs, $hMap->{'styleTemplateId'};
            }
        }
        $StyleGroup = $StyleGroup->parentByClass('StyleGroup');
    }
    $Servlet->writeContentOutput(\@Maps);
    $Servlet->params('ExitAfterEvent', 1);
}

sub JSONSaveLogo {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    $Servlet->form->{Values}->{'ScaleImageSize'}->[0] = $Shop->get('DefaultStyle')->get('ScaleLogoSize');
    $self->SaveLogo($Servlet);
    my $FileName = $Servlet->vars('Logo');
    my $hValues = {
      uploadedFile => $Shop->get('WebPath').'/'.$FileName,
      uploadedObjectFile => $FileName
    };
    $Servlet->writeContentOutput($hValues);
    $Servlet->params('ExitAfterEvent', 1);
}

sub ViewMyAccount {
    my $self = shift;
    my ($Servlet) = @_;

    $self->_redirectToUserAccount($Servlet);
}

sub _redirectToUserAccount {
    my $self = shift;
    my ($Servlet, $ViewActionPrefix) = @_;

    $ViewActionPrefix //= '';

    # redirect to SSL if SSL is available, but not requested
    unless ($Servlet->isSecureConnection()) {
        my $Shop = $Servlet->object->getSite();

        my $LocaleID = $Servlet->locale;
        my $Location = $Shop->get('WebServerScriptUrlSSL') . "/$LocaleID/";

        my @Params;
        push @Params, 'ObjectPath=' . UrlEncodeValue($Shop->get('Path'));
        push @Params, 'ViewAction=' . $ViewActionPrefix . 'ViewMyAccount';
        push @Params, 'LastViewAction=' . $Servlet->params('LastViewAction') if $Servlet->params('LastViewAction');
        push @Params, 'HideNotice=' . $Servlet->params('HideNotice') if $Servlet->params('HideNotice');

        $Location .= '?' . join('&', @Params) if scalar(@Params) > 0;

        $Servlet->redirect($Location);
        return;
    }

    my $User = $Servlet->user;

    if (defined $User and not $User->get('IsAnonymous')) {
        $Servlet->vars('ObjectID',$User->id);
        $Servlet->vars('ViewAction', $ViewActionPrefix . 'ViewMyAccount');
    }
    else {
        # change ViewAction to ViewRegistrationRedirect
        $Servlet->vars('ViewAction', $ViewActionPrefix . 'ViewRegistrationRedirect');
    }
}

sub UnityViewMyAccount {
    my $self = shift;
    my ($Servlet) = @_;

    $self->_redirectToUserAccount($Servlet, 'Unity');
}

sub View {
    my $self = shift;
    my ($Servlet) = @_;

    DeleteTemporaryStyles($Servlet->object);
    $self->SUPER::View($Servlet);
    return;
}

sub ViewMobileSF {
    my $self = shift;
    my ($Servlet) = @_;

    if ($Servlet->object->get('LoginRequired')) {
        $Servlet->redirectToSecureConnection(
            $Servlet->object->get('WebServerScriptUrlMobileSSL')
        );
    }

    $self->SUPER::ViewCached($Servlet);
    return;
}

sub ViewEditor {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop              = $Servlet->object->getSite();
    my $TargetURLUnitySSL = $Shop->get('WebUrlUnitySSL');
    my $TargetPathUnity   = $Servlet->params('TargetPathUnity');

    # create the correct url for the given objetct
    my $UniyObjectID      = $Servlet->params('UnityObjectID');
    if( $UniyObjectID ) {
        my $EditorObject    = $Shop->loadObject($UniyObjectID, 'ContentObject');
        $TargetURLUnitySSL  = $EditorObject->get('WebUrlUnitySSL');
    }

    # inject /editor after hostname (and port)
    $TargetURLUnitySSL =~ s/^(\w+:\/\/[^\/\?#]+)/$1\/editor/;

    # inject target path into URL if needed (like /l/contact)
    # for opening the editor with the correct path
    # e.g. subdomainrouting : http://enmhoehn121020161401.mhoehn-vm-centos.intern.epages.de:7000/l/privacy
    # e.g. without subdomainrouting : http://mhoehn-vm-centos.intern.epages.de:7000/l/privacy?shop=enmhoehn121020161401
    if(defined $TargetPathUnity){
        $TargetURLUnitySSL =~ s:(\?|$):$TargetPathUnity$1:;
    }

    # add token to the end of the URL
    $TargetURLUnitySSL = AddEditorTokenToURL($Servlet, $TargetURLUnitySSL);

    # write correct URL to the Servlet
    $Servlet->params('TargetURLUnitySSL', $TargetURLUnitySSL);

    $self->SUPER::View($Servlet);
    return;
}

sub ViewThemeManager {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop              = $Servlet->object->getSite();
    my $TargetURLUnitySSL = $Shop->get('WebUrlUnitySSL');

    $TargetURLUnitySSL =~ s:(\?|$):\/thememanager$1:;

    # add token to the end of the URL
    $TargetURLUnitySSL = AddEditorTokenToURL($Servlet, $TargetURLUnitySSL);

    $Servlet->params('TargetURLUnitySSL', $TargetURLUnitySSL);

    $self->SUPER::View($Servlet);
    return;
}

1;
