#========================================================================================
# package      DE_EPAGES::Installer::API::PatchBuilder
# state        public
# description  This modul generates and change all files necessary for patch.
#               These files can be find under ../Patches of each cartridge.
#               The patch change the file ../API/Cartridge.pm in each cartridge.
#               Each patch increase the version number of each cartridge, even if there are no changes.
#               So the ePages version number and the cartridge version numbers are linked.
#
#========================================================================================
package DE_EPAGES::Installer::API::PatchBuilder;

use strict;
use feature 'switch';

use Config;
use Getopt::Long;
use Data::Dumper;
use DE_EPAGES::Core::API::CVS;
use DE_EPAGES::Core::API::File qw (
    CopyFile
    CreateDirectory
    DeleteDirectory
    DeleteFile
    DetectFileEncoding
    ExistsDirectory
    ExistsFile
    GetFileContent
    GetFileMD5Digest
    GetFileNames
    GetSubDirectories
    RunOnDirectory
    WriteFile);
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Installer::API::CVSTagHandler;
use DE_EPAGES::Installer::API::Object::PatchCodeGenerator;
use DE_EPAGES::Installer::API::XMLFilePatcher;
use DE_EPAGES::Installer::API::PatchXMLExportDriver;
use DE_EPAGES::Object::API::Factory qw(SplitObjectPath);
use DE_EPAGES::Object::API::XMLCompare qw (CompareXML);
use DE_EPAGES::Object::API::XMLStructureExportDriver;

## no critic qw(RequireExtendedFormatting ProhibitLongLines)

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder = DE_EPAGES::Installer::API::PatchBuilder->new(
#                   'OldReleaseTag' => $OldReleaseTag,
#                   'NewReleaseTag' => $NewReleaseTag,
#                   'Verbose' => 1,
#                   'ForceReview' => 1,
#                   'DontExportDeletedDirectories' => 0,
#               );
#----------------------------------------------------------------------------------------
# description  Creates a new PatchBuilder object
#----------------------------------------------------------------------------------------
# input        $hOptions | keys
#                   <ul>
#                       <li>OldReleaseTag - old release tag, e.g. 'release_6_9_1'
#                       <li>NewReleaseTag - new release tag, e.g. 'release_6_9_2'
#                       <li>Verbose       - print debug information (0 or 1)
#                       <li>ForceReview   - add dies to Cartridge.pm to force manual check
#                       <li>DontExportDeletedDirectories - set to 1 if deleted cartridge
#                           directories should not exported, e.g. for language pack patch
#                       <li>reconstruct   - 1 to reconstruct from CVS
#                                           ignore already existing patch files
#                       <li>CommitBranch - name of branch to commit
#                       <li>AutoMode - PatchBuilder operates in HEAD Branch
#                       <li>UseGeneratedBranch - ony allowed in combination with AutoMode, operates in branches HEAD and B_generated_patch_code
#                   </ul>
#               | hash
# return       $PatchBuilder | PatchBuilder object | object
#========================================================================================
sub new {
    my $class = shift;
    my ($hOptions) = @_;
    TestString( 'OldReleaseTag', $hOptions->{'OldReleaseTag'} );
    TestString( 'NewReleaseTag', $hOptions->{'NewReleaseTag'} );
    die "\nmissing option 'AutoMode'" unless (defined $hOptions->{'AutoMode'});
    die "\n'UseGeneratedBranch' only allowed with option 'AutoMode'" if (
        (defined $hOptions->{'UseGeneratedBranch'}) and
        ($hOptions->{'UseGeneratedBranch'}) and
        not ($hOptions->{'AutoMode'})
    );
    my $self = bless $hOptions, $class;

    $self->{'CompanyName'} = 'DE_EPAGES' if (not defined($self->{'CompanyName'}));
    $self->{'PatchStructure'} = {};

    #set old tag
    $self->{'OldReleaseOrHotFixTag'} = $self->{'OldReleaseTag'};

    return $self;
}

#========================================================================================
# function     exportReleaseFolders
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportReleaseFolders($CVSPath, $OldDir, $NewDir);
#----------------------------------------------------------------------------------------
# description  step 1
#               export two versions of one cartridge, one tagged with old release and
#               one tagged with the new release tag to different folders
#----------------------------------------------------------------------------------------
# input        $CVSPath | cvs path from cartridge that should be patcht | string
# input        $OldDir | directory to export cartridge version tagged with old release tag | string
# input        $NewDir | directory to export cartridge version tagged with new release tag | string
#========================================================================================
sub exportReleaseFolders {
    my $self = shift;
    my ($CVSPath, $OldDir, $NewDir) = @_;

    my $OriginalOldReleaseOrHotFixTag = $self->{'OldReleaseOrHotFixTag'};

    #print message
    $self->printMessage( "\n >step-1: export old and new version of cartridge to\n   -$OldDir and\n   -$NewDir\n" );

    #check if to use hot fix tag
    if ($self->{'UseOldHotFixTag'}){

        #check for hotfix tag
        my $CVSTagHandler = $self->DE_EPAGES::Installer::API::CVSTagHandler->new();

        #print message
        $self->printMessage( "\n >WARNING: changed old target tag from '$OriginalOldReleaseOrHotFixTag' to '" . $self->{'OldReleaseOrHotFixTag'} . "' \n" );

        #set tag to build from
        $self->{'OldReleaseOrHotFixTag'} = $CVSTagHandler->getLatestCorrespondingHotFixTag(
            $CVSPath . '/API/Cartridge.pm',
            $self->{'OldReleaseTag'}
        );
    }

    #export cartridge versions tagged with old and new release
    $self->exportReleaseFolder($CVSPath, $OldDir, $self->{'OldReleaseOrHotFixTag'});
    $self->exportReleaseFolder($CVSPath, $NewDir, $self->{'NewReleaseTag'});

    return;
}

#========================================================================================
# function     generatePatchFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->generatePatchFiles($OldFolder, $NewFolder, $CartridgeFolder);
#----------------------------------------------------------------------------------------
# description  step 2
#               - compare old and new version of cartridge
#               - generates all patch files under ../Patches/Patch../Database/XML
#               - (if patch files already builded, only additional tags where added
#                  manually changes with stay)
#----------------------------------------------------------------------------------------
# input        $OldFolder | folder of cartridgeversion tagged with old release tag | string
# input        $NewFolder | folder of cartridgeversion tagged with new release tag | string
# input        $CartridgeFolder | folder of cartridge, where should commit | string
# input        $useGit | 1, if you use the Git version, 0 otherwise | bool
#========================================================================================
sub generatePatchFiles {
    my $self = shift;
    my ($OldFolder, $NewFolder, $CartridgeFolder, $useGit) = @_;

    $self->printMessage("\n##### Generate patch files#####\n");

    #get name of folder to save patch files
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);
    my $CommitBranch = $self->{'CommitBranch'};

    #Git replacement does not seem necessary at this point
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    #create directory unless exists
    CreateDirectory($TargetFolder) unless (ExistsDirectory($TargetFolder));

    if(!$useGit) {
        #add target folder to cvs
        RunOnDirectory(
            $CartridgeFolder,
            sub {
                my $RelativeTargetName = $TargetFolder;
                $RelativeTargetName =~ s/\Q$CartridgeFolder\E\///;

                #add folder Patches
                $CVS->command('add', undef, ['Patches'], {}, {'skiperrors' => 1});

                #add target folder
                $CVS->command('add', undef, [$RelativeTargetName], {}, {'skiperrors' => 1});
            }
        );

        #update target folder if commit branch is set
        RunOnDirectory(
            $TargetFolder,
            sub {
                if (
                    defined $CommitBranch and
                    $CommitBranch ne ''
                ){
                    #message
                    $self->printMessage("\nCommit Branch selected, updating file '$TargetFolder' to branch '$CommitBranch'\n");

                    #update Cartridge.pm to special commit branch
                    $CVS->command( 'update', { '-d', '-P', '-r', $CommitBranch }, undef ) ;
                }
            }
        );
    }

    #remove all previous patch files if reconstruct is enabled
    $self->_cleanPreviousPatchFiles($TargetFolder) if ($self->{'reconstruct'});

    #print Message
    $self->printMessage("\n >step-2: collect informations from old and new version of Cartridge, save patch files at\n   -$TargetFolder\n");

    #handle files that changed from old to new release tag
    $self->_handleChangedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle files that only can find in new cartridge version
    $self->_handleAddedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle files that only can find in old cartridge version
    $self->_handleDeletedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle directories that only can find in old cartridge version
    $self->_handleDeletedDirectories($CartridgeFolder, $OldFolder, $NewFolder)
    unless (exists $self->{'DontExportDeletedDirectories'} and $self->{'DontExportDeletedDirectories'} == 1);
}

sub _cleanPreviousPatchFiles {
    my $self = shift;
    my ($Folder) = @_;

    if(ExistsDirectory($Folder)){
        foreach my $File (@{GetFileNames($Folder, {'recursive' => 1})}){
            #ignore files that ignored by cvs
            next if ($File =~ /\/CVS\//);
            print "\n  delete file '$File'";
            DeleteFile($File) if ExistsFile($File);
            print "\n";
        }
    }
}

#========================================================================================
# function     createPatchXMLFromPatchFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->createPatchXMLFromPatchFiles($CartridgePath, $CartridgePackage, $BaseName );
#----------------------------------------------------------------------------------------
# description  step 3
#               creates the patch.xml file from the files in the patch directory.
#----------------------------------------------------------------------------------------
# input        $CartridgePath | target directory for the cartridge | string
# input        $CartridgePackage | cartridge package, e.g. 'DE_EPAGES::Object' | string
# input        $BaseName | (optional; default: 'patch.xml') output file name | string
# return       $Created | true if an XML file was created | boolean
#========================================================================================
sub createPatchXMLFromPatchFiles{
    my $self = shift;
    my ($CartridgePath, $CartridgePackage, $BaseName) = @_;

    #generate names
    $BaseName =  'patch.xml' unless defined $BaseName;
    my $CartridgePatchPath = _GetPatchTargetName($CartridgePath, $self->oldVersion);
    my $FileName = $CartridgePatchPath .'/'. $BaseName;
    my $CartridgePatchDBXMLPath = $CartridgePatchPath . '/Database/XML';

    #print message
    $self->printMessage( "\n >step-3: generate or actualize\n   -$CartridgePatchPath\/patch.xml" );

    #save information in hash structure, necessary for export patch.xml
    $self->{'PatchStructure'}->{'PatchCartridge'} = {
            'OldRelease' => $self->{'OldReleaseOrHotFixTag'},
            'NewRelease' => $self->{'NewReleaseTag'},
            'NewVersion' => $self->newVersion,
            'Cartridge'  => $CartridgePackage,
            'OldVersion' => $self->oldVersion,
    };

    #check if patch folder exists
    if (ExistsDirectory($CartridgePatchDBXMLPath)){

        #for each file in patch folder
        foreach my $DBXMLFile (@{GetFileNames($CartridgePatchDBXMLPath)}){
            #ignore files that ignored by cvs
            next if ($DBXMLFile =~ /\/\.\#[a-zA-Z\.\_0-9]*/);

            my $DBXMLFilePatchPath = $DBXMLFile;
            $DBXMLFile =~ s/^\Q$CartridgePatchDBXMLPath\E\///;
            $DBXMLFilePatchPath =~ s/^\Q$CartridgePath\E\///;

            #save information in hash structure, necessary for export patch.xml
            $self->_collectPatchXMLInfoFromFiles($DBXMLFile, $DBXMLFilePatchPath);
        }
    }
    else {
        #no patch folder found, no db content changes have to add to patch.xml
        $self->printMessage("\n    no patch folder found for $CartridgePath");
    }

    #stop if nothing is to do
    return 0 if (
        (#$self->{'PatchStructure'}->{'PatchDBStructure'} doesnt exists
            (not exists $self->{'PatchStructure'}->{'PatchDBStructure'}) or
            (#only warnings exists in $self->{'PatchStructure'}->{'PatchDBStructure'}
                exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'} and
                scalar keys %{$self->{'PatchStructure'}->{'PatchDBStructure'}} == 1
            )
        ) and
        (not exists $self->{'PatchStructure'}->{'Patch'}) and
        (not exists $self->{'PatchStructure'}->{'PatchCartridgeDir'})
    );

    #export patch.xml
    my $PatchExportDriver = DE_EPAGES::Installer::API::PatchXMLExportDriver->new($self->{'PatchStructure'});
    if (not ExistsFile($FileName)){#patch.xml doesnt exists, patcht first time
        $PatchExportDriver->exportXML($FileName);
    }
    else{#an older version of patch.xml file already exists
        #generate temporary new name
        my $NewFileName = $FileName;
        $NewFileName =~ s/(\Q$BaseName\E)$/\.\#\(new\)$1/;

        #export to temporary file
        $PatchExportDriver->exportXML($NewFileName);

        if (CompareXML($NewFileName, $FileName) and CompareXML($FileName,$NewFileName)){
            #no changes detected - delete the new file
            DeleteFile($NewFileName);
        }
        elsif (CompareXML($NewFileName, $FileName)){#new patch.xml has no additional tags or changed values, nothing to do
            print "    [Warning:] difference in patch.xml file detected, please check manually\n";
            print "       -original $FileName\n";
            print"        -new      $NewFileName\n";
        }
        else{#new patch.xml has additional tags or attributes
            my $OldFileName = $FileName;
            $OldFileName =~ s/(\Q$BaseName\E)$/\.\#\(old\)$1/;

            #get xml structure of old and new patch.xml files
            my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
            my $OrigXMLStruct = $ImportHandler->getXMLStructure($FileName);
            my $AddXMLStruct = $ImportHandler->getXMLStructure($NewFileName);

            #merge new and old patch.xml, positions could be destroyed, it is necessary to correct the positions
            $OrigXMLStruct->assimilateXMLStructure($AddXMLStruct,{'overwrite' => '1'});
            _CorrectPositionsOfPatchXMLXMLStruct($OrigXMLStruct);

            #save old file
            CopyFile($FileName,$OldFileName);
            DeleteFile($NewFileName);


            #export combined patch.xml
            my $ExportDriver = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
            $ExportDriver->exportXML($FileName, $OrigXMLStruct);
            print "    [Warning:] difference in patch.xml file detected, please check manually\n";
            print "       -original $OldFileName\n";
            print "       -new      $FileName\n";
        }
    }

    return 1;
}

#========================================================================================
# function     patchCartridgePm
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->patchCartridgePm($CartridgePath)
#----------------------------------------------------------------------------------------
# description  step 4
#               patch the ../API/Cartridge.pm file
#----------------------------------------------------------------------------------------
# input        $CartridgePath | target directory for the cartridge | string
# input        $GenerateWarnings | if 0 no warnings will generate in Cartridge.pm | bool
# input        $useGit | 1, if Git version is used, 0 otherwise | bool
#========================================================================================
sub patchCartridgePm{
    my $self = shift;
    my ($CartridgePath,$GenerateWarnings,$useGit) = @_;

    #get Info
    my $CommitBranch = $self->{'CommitBranch'};
    my $CartridgePmPath = $CartridgePath . '/API/Cartridge.pm';
    my $sCartridgePm = GetFileContent($CartridgePmPath, {'usebom' => 1 }); #use byte order mark
    my $Encoding = DetectFileEncoding($CartridgePmPath);
    my $OriginalContent = $$sCartridgePm;
    my $OldVersion = $self->oldVersion;
    my $NewVersion = $self->newVersion;
    my $OldVersionUnderscored = $OldVersion;
    my $NewVersionUnderscored = $NewVersion;
    my $CodeGenerator = DE_EPAGES::Installer::API::Object::PatchCodeGenerator->new();
    my $ControlCode = $CodeGenerator->getControlCode();
    $OldVersionUnderscored =~ s/\./\_/g;
    $NewVersionUnderscored =~ s/\./\_/g;


    if(!$useGit) {
        if (
            defined $CommitBranch and
            $CommitBranch ne ''
        ){
            #message
            $self->printMessage("\nCommit Branch selected, updating file '$CartridgePath/API/Cartridge.pm' to branch '$CommitBranch'\n");

            #update Cartridge.pm to special commit branch
            $self->cleanUpdateToBranch($CartridgePath . '/API', $CommitBranch, 'Cartridge.pm');
        }
    }

    $GenerateWarnings = 1 unless defined $GenerateWarnings;

    #print message
    $self->printMessage("\n >step-4: patch $CartridgePmPath");

    #paste control code
    $$sCartridgePm =~ s{(
        \#[\=]+\n                 #comment line
        \#\s\function\s+new   #'# function     new'
    )}
    {$ControlCode$1}x if ($self->{'ForceReview'} eq '1');

    #register versions
    if (not $$sCartridgePm =~ /\s+\'Version\'\s+\=\>\s\'\Q$NewVersion\E/){
        #update Version
        $$sCartridgePm =~ s/(\s+\'Version\'\s+\=\>\s\')[0-9\.]+(\'\,)/$1$NewVersion$2/;
    }
    if (not $$sCartridgePm =~ /\s+\'Patches\'\s+\=\>\s\[[0-9\.\,\']*\Q$OldVersion\E/){
        #register old version
        $$sCartridgePm =~ s/(\s+\'Patches\'\s+\=\>\s\[[0-9\.\,\']+)(\]\,)/$1\,\'$OldVersion\'$2/;
        $$sCartridgePm =~ s/(\s+\'Patches\'\s+\=\>\s\[)(\])/$1\'$OldVersion\'$2/;
    }

    #check for changes
    if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}){
        my $ExecutePatchXML = '';

        #create warning
        my $PatchBuilderWarning = '';
        if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}){

            if ($GenerateWarnings){
                foreach my $DontKnowWhatToDoFile (keys %{$self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}}){

                    if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} and
                        $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} eq 'Changed')
                    {
                        $PatchBuilderWarning = $PatchBuilderWarning . "    die \"changes in $DontKnowWhatToDoFile cant automate at these time.\"\;\n";
                    }
                    elsif (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} and
                        $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} eq 'Deleted')
                    {
                        $PatchBuilderWarning = $PatchBuilderWarning . "    die \"delete $DontKnowWhatToDoFile cant automate at these time.\"\;\n";
                    }
                    else
                    {
                        die 'cant generate warning for constalation warning ' .
                            $DontKnowWhatToDoFile . ' -> ' .
                            Dumper($self->{'PatchStructure'}->{'PatchDBStructure'}->
                                {'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}) . "\n";
                    }
                }
            }
        }

        #create code generator object and placeholder
        my $DBStructureChangeCode = '';

        #check if patch.xml should be executed
        if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}){
            #define function call to execute patch.xml
            $ExecutePatchXML = "\n    \$self->importPatchDir('Patch_$OldVersionUnderscored', 'PatchDBStructure');\n";
        }

        #generate code to handle changes in sql files
        if (scalar keys %{$self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'mysql'}} >= 1){
            $DBStructureChangeCode .= $CodeGenerator->getCodeForDiffInSQLFiles($self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'mysql'}, 'mysql');
        }
        $DBStructureChangeCode =~ s/\s+my\s+\$dbi\s+\=\s+GetCurrentDBHandle\(\)\;/FIRST_DBI_POSITION/;
        $DBStructureChangeCode =~ s/\s+my\s+\$dbi\s+\=\s+GetCurrentDBHandle\(\)\;//g;
        $DBStructureChangeCode =~ s/FIRST_DBI_POSITION/    my \$dbi \= GetCurrentDBHandle\(\)\;\n/;

        my $FunctionHeader = "\n#========================================================================================\n" .
                "# function     patchDBStructure_$OldVersionUnderscored\n" .
                "# state        private\n" .
                "#----------------------------------------------------------------------------------------\n" .
                "# syntax       \$Installer->patchDBStructure_$OldVersionUnderscored;\n" .
                "#----------------------------------------------------------------------------------------\n" .
                "# description  Patches cartridge db structure in active store (database connection)\n" .
                "#               from version '$OldVersion' to version '$NewVersion'.\n" .
                "#----------------------------------------------------------------------------------------\n" .
                "# return       \$Version | version of cartridge after patch function, '$NewVersion' | string\n" .
                "#========================================================================================\n";

        #check for rebuild from CVS
        if ($self->{'reconstruct'}){
            #remove old patchDBStructure function
            $$sCartridgePm =~ s/\Q$FunctionHeader\E//g;
            $$sCartridgePm =~ s/sub\spatchDBStructure\_\Q$OldVersionUnderscored\E\s\{.*?return\s\'\Q$NewVersion\E\'\;.*?\}\n//s;
        }

        if (not $$sCartridgePm =~ /sub\spatchDBStructure\_\Q$OldVersionUnderscored\E\s\{/){
            #define function
            my $FunctionCode =
                $FunctionHeader .
                "sub patchDBStructure_$OldVersionUnderscored {\n" .
                "    my \$self = shift;\n" .
                "    #dont use API function, to avoid inconistencies between perl modules and database content\n" .
                "$PatchBuilderWarning\n" .
                "$DBStructureChangeCode\n" .
                "$ExecutePatchXML\n" .
                "    return '$NewVersion';\n" .
                "}\n";

            #write function to Cartridge.pm file content
            $$sCartridgePm =~ s/(\n1\;)/$FunctionCode$1/ if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'});
        }
    }

    #generate options for encoding
    my $hFlags = undef;
    if(defined $Encoding){
        $hFlags = { 'usebom' => 1, 'encoding' => $Encoding };
    }

    #write Cartridge.pm file
    WriteFile($CartridgePmPath,$sCartridgePm, $hFlags) if (
        ($OriginalContent ne $$sCartridgePm)

        and not(
            #exists in original Cartridge.pm
            ($OriginalContent =~ /sub\spatchDBStructure\_\Q$OldVersionUnderscored\E\s\{/)
            and
            #not exists in generated new Cartridge.pm
            (not($$sCartridgePm =~ /sub\spatchDBStructure\_\Q$OldVersionUnderscored\E\s\{/))
        )
    );
}

#========================================================================================
# function     cleanUpdateToBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->cleanUpdateToBranch($Path,$Branch)
#----------------------------------------------------------------------------------------
# description  clean update to branch or tag
#----------------------------------------------------------------------------------------
# input        $Path | path to update | string
# input        $Branch | branch to update from | string
# input        $Part | files or folders | string
#========================================================================================
sub cleanUpdateToBranch{
    my $self = shift;
    my ($Path,$Branch,$Part) = @_;

    #get cvs handler
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    RunOnDirectory(
        $Path,
        sub {

            #message
            $self->printMessage("\nupdating folder '$Path' to branch '$Branch'\n");

            #update Cartridge.pm to special commit branch
            $CVS->command( 'update', { '-d', '-C', '-P', '-r', $Branch }, [$Part] ) ;
        }
    );
}

#========================================================================================
# function     updateToBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->updateToBranch($Path,$Branch)
#----------------------------------------------------------------------------------------
# description  update to branch or tag
#----------------------------------------------------------------------------------------
# input        $Path | path to update | string
# input        $Branch | branch to update from | string
# input        $Part | files or folders | string
#========================================================================================
sub updateToBranch{
    my $self = shift;
    my ($Path,$Branch,$Part) = @_;

    #get cvs handler
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    RunOnDirectory(
        $Path,
        sub {

            #message
            $self->printMessage("\nupdating folder '$Path' to branch '$Branch'\n");

            #update Cartridge.pm to special commit branch
            $CVS->command( 'update', { '-d', '-P', '-r', $Branch }, [$Part] ) ;
        }
    );
}

#========================================================================================
# function     setCommitBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->setCommitBranch($Branch)
#----------------------------------------------------------------------------------------
# description  actualize the commit branch
#----------------------------------------------------------------------------------------
# input        $Branch | branch to commit | string
#========================================================================================
sub setCommitBranch {
    my $self = shift;
    my ($CommitBranch) = @_;

    $self->{'CommitBranch'} = $CommitBranch;

    return;
}

#========================================================================================
# function     addPatchFolderToCVS
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->addPatchFolderToCVS($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 5
#               add files under ../Patches to cvs
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub addPatchFolderToCVS {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    #get info
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    if (ExistsDirectory($TargetFolder)){
        RunOnDirectory(
            $CartridgeFolder,
            sub {


                my $RelativeTargetName = $TargetFolder;
                $RelativeTargetName =~ s/\Q$CartridgeFolder\E\///;


                $CVS->command('add', undef, [$RelativeTargetName], {}, {'skiperrors' => 1});

                foreach my $Folder (@{GetSubDirectories($TargetFolder, {'recursive' => 1})}){
                    next if ($Folder =~ /\/CVS/);
                    my $RelativeName = $Folder;
                    $RelativeName =~ s/\Q$CartridgeFolder\E\///;

                    $CVS->command('add', undef, [$RelativeName], {}, {'skiperrors' => 1});
                }

                foreach my $File (@{GetFileNames($TargetFolder, {'recursive' => 1})}){
                    next if ($File =~ /\/CVS\//);
                    next if ($File =~ /\.\#/);
                    my $RelativeFileName = $File;
                    $RelativeFileName =~ s/\Q$CartridgeFolder\E\///;

                    $CVS->command('add', undef, [$RelativeFileName], {}, {'skiperrors' => 1});
                    $CVS->command('update', undef, [$RelativeFileName], {}, {'skiperrors' => 1});
                }
                return;
            }
        );
    }
    return;
}

sub _getPatchTag{
    my $self = shift;
    my ($ReleaseTag) = @_;

    $ReleaseTag =~ s/release/patch/;

    return $ReleaseTag;
}

#========================================================================================
# function     commitPatchFolderToCVS
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->commitPatchFolderToCVS($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 5
#               add files under ../Patches to cvs
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub commitPatchFolderToCVS {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot add patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    my $CommitString = 'automated patch building from \'' . $self->{OldReleaseTag} . '\' to \'' . $self->{NewReleaseTag} . '\'';

    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {

                #commit
                $CVS->command( 'commit', { '-m', $CommitString } );

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     updateCVSPatchReleaseTagsOnPatchFolder
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->updateCVSPatchReleaseTagsOnPatchFolder($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 9
#               update tags
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub updateCVSPatchReleaseTagsOnPatchFolder {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot add patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);
    my $CVSPath = _GetCVSPath($TargetFolder);

    #get tag handler
    my $CVSTagHandler = DE_EPAGES::Installer::API::CVSTagHandler->new();

    #get control tag
    my $NewReleaseTag = $self->{NewReleaseTag};
    my $PatchTag = $self->_getPatchTag($NewReleaseTag);

    die "\n patch and release tag are same '$PatchTag'" if ($PatchTag eq $NewReleaseTag);

    my $ReleaseControlTag = $CVSTagHandler->getControlTag($NewReleaseTag);
    my $PatchControlTag = $CVSTagHandler->getControlTag($PatchTag);
    my $CommitBranch = $self->{'CommitBranch'};
    $CommitBranch = 'HEAD' if ($CommitBranch eq 'B_generated_patch_code');

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {
                #set necessary tags on generated files
                $CVSTagHandler->setTagByServer($NewReleaseTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($ReleaseControlTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchControlTag, $CommitBranch, $CVSPath);

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     updateCVSPatchReleaseTagsOnCartridgePM
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->updateCVSPatchReleaseTagsOnCartridgePM($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 9
#               tag Cartridge.pm file
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub updateCVSPatchReleaseTagsOnCartridgePM {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot commit patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = $CartridgeFolder . '/API';

    #get tag handler
    my $CVSTagHandler = DE_EPAGES::Installer::API::CVSTagHandler->new();

    #get control tag
    my $NewReleaseTag = $self->{NewReleaseTag};
    my $PatchTag = $self->_getPatchTag($NewReleaseTag);

    die "\n patch and release tag are same '$PatchTag'" if ($PatchTag eq $NewReleaseTag);

    my $ReleaseControlTag = $CVSTagHandler->getControlTag($NewReleaseTag);
    my $PatchControlTag = $CVSTagHandler->getControlTag($PatchTag);
    my $CommitBranch = $self->{'CommitBranch'};
    $CommitBranch = 'HEAD' if ($CommitBranch eq 'B_generated_patch_code');

    my $CVSPath = _GetCVSPath($TargetFolder . '/Cartridge.pm');

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {

                #set necessary tags on generated files
                $CVSTagHandler->setTagByServer($NewReleaseTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($ReleaseControlTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchControlTag, $CommitBranch, $CVSPath);

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     commitCartridgePMToCVS
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->commitCartridgePMToCVS($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 5
#               add files under ../Patches to cvs
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub commitCartridgePMToCVS {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot commit patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = $CartridgeFolder . '/API';
    my $CommitString = 'automated patch building from \'' . $self->{OldReleaseTag} . '\' to \'' . $self->{NewReleaseTag} . '\'';

    #get cvs handler
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {
                #cvs add unless is added
                $CVS->command('update', undef, [] );

                #commit
                $CVS->command( 'commit', { '-m', $CommitString }, ['Cartridge.pm'] );

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     exportReleaseFolder
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportReleaseFolder($CVSPath, $Dir, $Tag);
#----------------------------------------------------------------------------------------
# description  part of step 1
#               export two versions of one cartridge, one tagged with old release and
#               one tagged with the new release tag to different folders
#----------------------------------------------------------------------------------------
# input        $CVSPath | cvs path from cartridge that should be patcht | string
# input        $OldDir | directory to export cartridge version tagged with $Tag | string
# input        $Tag | tag of cartridge, most any release tag | string
#========================================================================================
sub exportReleaseFolder {
    my $self = shift;
    my ($CVSPath, $Dir, $Tag) = @_;

    DeleteDirectory($Dir, { 'recursive'=>1});
    CreateDirectory($Dir);

    my $OS = $Config{'osname'};
    my $CVS = DE_EPAGES::Core::API::CVS->new({ '-q' => '' }, {'print' => 1});

    #check for operating system
    if ($OS eq 'MSWin32'){
        #use absolute pathes
        $CVS->command('export', {'-d' => $Dir, '-r'=>$Tag}, [ $CVSPath ]);
    }
    else{
        #use relative pathes

        #get pathes
        my $PathEnd = SplitObjectPath($Dir)->[-1];
        my $PathMainPart = $Dir;
        $PathMainPart =~ s/[\\|\/]\Q$PathEnd\E$//;

        #run on directory
        RunOnDirectory(
            $PathMainPart,
            sub {
                $CVS->command('export', {'-d' => $PathEnd, '-r'=>$Tag}, [ $CVSPath ]);
            }
        );
    }

    return;
}

#========================================================================================
# function     verbose
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->verbose()
#----------------------------------------------------------------------------------------
# description  return object attribute verbose
#----------------------------------------------------------------------------------------
# return       $Verbose | object attribute | boolean
#========================================================================================
sub verbose {
    return shift->{'Verbose'};
}

#========================================================================================
# function     printMessage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->printMessage('message to test');
#----------------------------------------------------------------------------------------
# description  print message if verbose is true
#----------------------------------------------------------------------------------------
# input        $Message | message to print | string
#========================================================================================
sub printMessage {
    my $self = shift;
    my ($Message) = @_;

    if( $self->verbose ) {
        print $Message, "\n";
    }
}

#========================================================================================
# function     newVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $NewVersion = $self->newVersion();
#----------------------------------------------------------------------------------------
# description  return object attribute, new version
#----------------------------------------------------------------------------------------
# return       $NewVersion | object attribute | string
#========================================================================================
sub newVersion {
    my $self = shift;
    return GetVersionByReleaseTag( $self->{'NewReleaseTag'} );
}

#========================================================================================
# function     oldVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $OldVersion = $self->oldVersion();
#----------------------------------------------------------------------------------------
# description  return object attribute, old version
#----------------------------------------------------------------------------------------
# return       $OldVersion | object attribute | string
#========================================================================================
sub oldVersion {
    my $self = shift;
    return GetVersionByReleaseTag( $self->{'OldReleaseOrHotFixTag'} );
}

#========================================================================================
# function     _handleChangedFiles
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_handleChangedFiles($CartridgeFolder, $OldFolder, $NewFolder);
#----------------------------------------------------------------------------------------
# description  part of step 2
#               handle changed files, generates patch xml files and
#               collect information under $self->{'PatchStructure'}, necessary to export patch.xml
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | folder of cartridge, where should commit | string
# input        $OldFolder | folder of cartridgeversion tagged with old release tag | string
# input        $NewFolder | folder of cartridgeversion tagged with new release tag | string
#========================================================================================
sub _handleChangedFiles {
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    #get old files, by run on old directory
    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    #get new files, by run on new directory
    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    #get all filenames that can find in the new and in the old directory
    my $aIntersectionItems = _GetIntersectionItems($aOldFiles,$aNewFiles);

    #remove files that are equal in old and new folder
    my $aChangedFiles = _RemoveFilesNotChanged($aIntersectionItems, $OldFolder, $NewFolder);

    #get patch target folder
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    #get cartridge name
    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    foreach my $FileName (@$aChangedFiles) {#for each changed file
        $FileName=~ s/^\///;

        given ($FileName) {
            when (/^Database\/XML\/.*\.xml/) {
                my $OldFile = $OldFolder . '/' . $FileName;
                my $NewFile = $NewFolder . '/' . $FileName;
                $self->_handleChangedXMLFile( $CartridgeName, $OldFile, $NewFile, $FileName, $TargetFolder );
            }
            when (/^Database\/Sybase\/Procedures\/.*/) {
                #generate relative name to patch file
                my $RelTargetName = $TargetFolder . '/' . $FileName;
                $RelTargetName =~ s/.*\/(Patches\/Patch\_.*)$/$1/;

                #register files for handle in patch.xml file
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}->{'02-ProceduresAndOther'}->{$RelTargetName}->{'CVS'} = 'Changed';

                #copy changed procedure file in patch folder
                CopyFile(
                    $NewFolder . '/' . $FileName,
                    $TargetFolder . '/' . $FileName,
                    {'create_directory' => 1}
                );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\/.*/ ) {
                next if($FileName =~ /delete\_.*?\.sql$/);

                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$FileName}->{'CVS'} = 'Changed';

                #register files, that have to be code generated for
                if ($FileName =~ /^Database\/Sybase\/.*/){
                    $self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'Sybase'}->{$OldFolder . '/' . $FileName} = $NewFolder . '/' . $FileName;
                }
                elsif ($FileName =~ /^Database\/mysql\/.*/){
                    $self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'mysql'}->{$OldFolder . '/' . $FileName} = $NewFolder . '/' . $FileName;
                }
            }
            when ( /^Data\/WebRoot\/.*/ ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyWebRootFile'}->{$FileName}->{'CVS'} = 'Changed';
            }
            when ( /^Data\/javascript\/.*/ ) {
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyJavascriptFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Public\/.*/ ) {
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyPublicFile'}->{$FileName}->{'CVS'} = 'Changed';

                unless( $FileName =~ /\bjavascript\b/ ||
                    $FileName =~ /\.js$/ ||
                    $FileName =~ /\.css$/
                ) {
                    my $OldFile = $OldFolder . '/' . $FileName;
                    my $MD5 = GetFileMD5Digest($OldFile);
                    $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyPublicFile'}->{$FileName}->{'OldMD5Digest'} = $MD5;
                }
            }
            when ( /^Data\/Private\/.*/ ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyPrivateFiles'}->{$FileName}->{'CVS'} = 'Changed';
            }
            when ( /^Data\/Scheduler\/.*/ ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySchedulerFiles'}->{$FileName}->{'CVS'} = 'Changed';
            }
            when ( /^Data\/SharedConfig\/.*/ ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySharedConfigFiles'}->{$FileName}->{'CVS'} = 'Changed';
            }
        }
    }
}

#funktion handhabt alle hinzugefgten Files
sub _handleAddedFiles{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    my $aImportFiles = _GetMissingItems($aOldFiles,$aNewFiles);

    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    foreach my $FileName (@$aImportFiles) {
        $FileName=~ s/^\///;
        given ($FileName) {
            when ( /^Database\/XML\/.*\.xml/ ) {
                my $NewFile = $NewFolder . '/' . $FileName;
                $self->_handleNewXMLFile( $CartridgeName, $NewFile, $FileName, $TargetFolder );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\/Tables/ ) { #tables should come before procedures in patch.xml file
                next if($FileName =~ /delete\_.*?\.sql$/);

                #generate relative name to patch file
                my $RelTargetName = $TargetFolder . '/' . $FileName;
                $RelTargetName =~ s/.*\/(Patches\/Patch\_.*)$/$1/;

                #register files for handle in patch.xml file
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}->{'01-Tables'}->{$RelTargetName}->{'CVS'} = 'New';

                #copy added file in patch folder
                CopyFile(
                    $NewFolder . '/' . $FileName,
                    $TargetFolder . '/' . $FileName,
                    {'create_directory' => 1}
                );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\/.*/ ) {
                #generate relative path to patch file
                my $RelTargetName = $TargetFolder . '/' . $FileName;
                $RelTargetName =~ s/.*\/(Patches\/Patch\_.*)$/$1/;

                #register files for handle in patch.xml file
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}->{'02-ProceduresAndOther'}{$RelTargetName}->{'CVS'} = 'New';

                #copy added file in patch folder
                CopyFile(
                    $NewFolder . '/' . $FileName,
                    $TargetFolder . '/' . $FileName,
                    {'create_directory' => 1}
                );
            }
            when ( /^Data\/WebRoot\/.*/){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyWebRootFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/javascript\/.*/){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyJavascriptFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Public\/.*/){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyPublicFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Private\/.*/){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyPrivateFiles'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Scheduler\/.*/){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySchedulerFiles'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/SharedConfig\/.*/){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySharedConfigFiles'}->{$FileName}->{'CVS'} = 'New';
            }
        }
    }
}

sub _handleNewXMLFile {
    my $self = shift;
    my ($CartridgeName, $NewFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /\/\.\#.*/);

    #check for displaced files
    if (
        $FileName =~ /\/\QDemoContent.xml\E$/ or
        $FileName =~ /\/\QDefaultShop.xml\E$/ or
        $FileName =~ /\/\QDefaultShopUnity.xml\E$/ or
        $FileName =~ /\/\QEbayRegions.xml\E$/ or
        $FileName =~ /\/\QShopTypes.xml\E$/ or
        $FileName =~ /\/ShopTypes.*\.xml$/ or
        $FileName =~ /\/\QStandardsLanguagePacks.xml\E$/ or
        $FileName =~ /\/\QDistributorStore.xml\E$/ or
        $FileName =~ /\/\QDistributors.xml\E$/ or
        $FileName =~ /\/\QProviders.xml\E$/ or
        $FileName =~ /\/\QDistributor.xml\E$/ or
        $FileName =~ /\/\QDHLProductList.xml\E$/
    ){
        print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
        return;
    }
    #very special case for FlexProvider
    elsif ($FileName =~ /\/DistributorShopTypes.*\.xml$/){
        print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
        unless($CartridgeName eq 'FlexProvider'){
            return;
        }
    }

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $OldFile = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Installer/Scripts/empty.xml';

    my $PatchFile = $FileName;
    $PatchFile =~ s/\QDatabase\E\/\QXML\E\///;

    #make import file
    my $OrigPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    my $TmpPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    my $OldPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 0);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 0);

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -previous $OrigPatchFile\n      -new $TmpPatchFile\n";
            }
        }
    }
}

sub _handleChangedXMLFile {
    my $self = shift;
    my ($CartridgeName, $OldFile, $NewFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /\/\.\#.*/);

    #check for displaced files
    if (
        $FileName =~ /\/\QDemoContent.xml\E$/ or
        $FileName =~ /\/\QDefaultShop.xml\E$/ or
        $FileName =~ /\/\QDefaultShopUnity.xml\E$/ or
        $FileName =~ /\/\QEbayRegions.xml\E$/ or
        $FileName =~ /\/\QShopTypes.xml\E$/ or
        $FileName =~ /\/ShopTypes.*\.xml$/ or
        $FileName =~ /\/\QStandardsLanguagePacks.xml\E$/ or
        $FileName =~ /\/\QDistributorStore.xml\E$/ or
        $FileName =~ /\/\QDistributors.xml\E$/ or
        $FileName =~ /\/\QDistributor.xml\E$/ or
        $FileName =~ /\/\QProviders.xml\E$/ or
        $FileName =~ /\/\QDHLProductList.xml\E$/
    ){
        print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
        return;
    }
    #very special case for FlexProvider
    elsif ($FileName =~ /\/DistributorShopTypes.*\.xml$/){
        print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
        unless($CartridgeName eq 'FlexProvider'){

            return;
        }
    }

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $PatchFile = $FileName;
    $PatchFile =~ s/\QDatabase\E\/\QXML\E\///;

    #make delete file
    my $OrigPatchFile = $TargetFolder . '/Database/XML/del_' . $PatchFile;
    my $TmpPatchFile = $TargetFolder . '/Database/XML/del_' . $PatchFile;
    my $OldPatchFile = $TargetFolder . '/Database/XML/del_' . $PatchFile;
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 1);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 1);

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                #no changes have to add
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -previous $OldPatchFile\n      -new $OrigPatchFile\n";
            }
        }
    }

    #make import file
    $OrigPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $TmpPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $OldPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 0);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 0);

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    [Warning:] new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one
            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -new $OrigPatchFile\n      -previous $OldPatchFile\n";
            }
        }
    }
}

#function handles all deleted files
sub _handleDeletedFiles{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    my $aDeleteFiles = _GetMissingItems($aNewFiles,$aOldFiles);

    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    foreach my $FileName (@$aDeleteFiles) {
        $FileName=~ s/^\///;
        given( $FileName ) {
            when ( /^Database\/XML\/.*\.xml/ ) { #../Database/XML/*.xml files
                my $OldFile = $OldFolder . '/' . $FileName;
                #register file name for delete in patch.xml
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';

                #build patch.xml files
                $self->_handleDeletedXMLFile( $CartridgeName, $OldFile, $FileName, $TargetFolder );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\/.*/ ) {
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/WebRoot\/.*/){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteWebRootFile'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/javascript\/.*/){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchDeleteJavascriptFile'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/Public\/.*/){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchDeletePublicFile'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/Private\/.*/){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeletePrivateFiles'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/Scheduler\/.*/){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteSchedulerFiles'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/.*/){
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            default {
                $self->{'PatchStructure'}->{'PatchCartridgeDirFinalPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
        }
    }
}

sub _handleDeletedXMLFile {
    my $self = shift;
    my ($CartridgeName, $OldFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /\/\.\#.*/);

    #check for displaced files
    if (
        $FileName =~ /\/\QDemoContent.xml\E$/ or
        $FileName =~ /\/\QDefaultShop.xml\E$/ or
        $FileName =~ /\/\QDefaultShopUnity.xml\E$/ or
        $FileName =~ /\/\QEbayRegions.xml\E$/ or
        $FileName =~ /\/\QShopTypes.xml\E$/ or
        $FileName =~ /\/ShopTypes.*\.xml$/ or
        $FileName =~ /\/\QStandardsLanguagePacks.xml\E$/ or
        $FileName =~ /\/\QDistributorStore.xml\E$/ or
        $FileName =~ /\/\QDistributors.xml\E$/ or
        $FileName =~ /\/\QDistributor.xml\E$/ or
        $FileName =~ /\/\QProviders.xml\E$/ or
        $FileName =~ /\/\QDHLProductList.xml\E$/
    ){
        print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
        return;
    }
    #very special case for FlexProvider
    elsif ($FileName =~ /\/DistributorShopTypes.*\.xml$/){
        print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
        unless($CartridgeName eq 'FlexProvider'){
            return;
        }
    }

    my $NewFile = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Installer/Scripts/empty.xml';

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $PatchFile = $FileName;
    $PatchFile =~ s/\QDatabase\E\/\QXML\E\///;

    #make delete file
    my $OrigPatchFile = "$TargetFolder/Database/XML/del_$PatchFile";
    my $TmpPatchFile = "$TargetFolder/Database/XML/del_$PatchFile";
    my $OldPatchFile = "$TargetFolder/Database/XML/del_$PatchFile";
    $TmpPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([a-zA-Z\.0-9\_]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 1);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 1);
        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Conflict:] patch already build\n      -previous $OrigPatchFile\n      -new $TmpPatchFile\n";
            }
        }
    }
}

sub _handleDeletedDirectories{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFolders = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetSubDirectories( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFolders = RunOnDirectory( $NewFolder, sub { GetSubDirectories( undef, {'recursive' => 1} ) } );

    my $aDeleteFolders = _GetMissingItems($aNewFolders,$aOldFolders);
    foreach my $FolderName (@$aDeleteFolders){
        $FolderName =~ s/^\///;
        $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}->{$FolderName}->{'CVS'} = 'Deleted';
    }
    #ignore subfolders
    if (exists $self->{'PatchStructure'}->{'PatchCartridgeDir'} and $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}){
        foreach my $FolderName (@$aDeleteFolders){
            foreach my $SubFolder (keys %{$self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}}){
                delete $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}->{$SubFolder}
                if $SubFolder =~ /^\Q$FolderName\E\/.+/;
            }
        }
    }
    #ignore files from deleted folders
    if (exists $self->{'PatchStructure'}->{'PatchCartridgeDir'} and $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}){
        foreach my $FolderName (keys %{$self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}}){
            $FolderName =~ s/^\///;
            foreach my $File (keys %{$self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}}){
                delete $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$File}
                if $File =~ /^\Q$FolderName\E\/.*/;
            }
        }
    }

}
#HilfsFunktion ermittelt alle Namen der Ordner die fehlen
sub _GetMissingItems {
    my ($aReferenceNames,$aControlNames) = @_;

    my $hRef;
    my @Result = ();
    my $CleanName;

    foreach my $RefName (@$aReferenceNames){
        $CleanName = $RefName;
        $CleanName =~ s/^\.//;
        $hRef->{$CleanName} = '1';
    }

    foreach my $ControlName (@$aControlNames){
        $CleanName = $ControlName;
        $CleanName =~ s/^\.//;
        push(@Result, $CleanName) if ($hRef->{$CleanName} ne '1');
    }

    return \@Result;
}

#return names of elements exists in both arrays
sub _GetIntersectionItems {
    my ($aReferenceNames,$aControlNames) = @_;

    my $hRef;
    my @Result = ();
    my $CleanName;

    foreach my $RefName (@$aReferenceNames){
        $CleanName = $RefName;
        $CleanName =~ s/^\.//;
        $hRef->{$CleanName} = '1';
    }

    foreach my $ControlName (@$aControlNames){
        $CleanName = $ControlName;
        $CleanName =~ s/^\.//;
        push(@Result, $CleanName) if ($hRef->{$CleanName} eq '1');
    }

    return \@Result;
}

sub _RemoveFilesNotChanged{
    my ($aIntersectionItems, $OldDir, $NewDir) = @_;

    my @Result;

    foreach my $FileName (@$aIntersectionItems){
        my $CheckSumOld = GetFileMD5Digest($OldDir . $FileName);
        my $CheckSumNew = GetFileMD5Digest($NewDir . $FileName);
        push(@Result, $FileName) if ($CheckSumOld ne $CheckSumNew);
    }
    return \@Result;
}

sub _collectPatchXMLInfoFromFiles{
    my $self = shift;
    my ($DBXMLFile, $DBXMLFilePatchPath) = @_;

    #onPatchImportObjects
    given ($DBXMLFile) {
        when ( /^\Qimp_Attributes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'01-Attributes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_LastUpdateProfiles\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'02-LastUpdateProfiles'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Actions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }

        #exceptions please remove if files renamed
        when ( /^\Qimp_GoogleAnalyticsActions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_EtrackerActions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Permissions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'04-Permissions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_PageTypes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'05-PageTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_TemplateTypes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'06-TemplateTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Forms\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'07-Forms'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Features\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'08-Features'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_ClientApplication\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'09-ClientApplication'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_WidgetGroups\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'10-WidgetGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_WidgetTypes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'11-WidgetTypes'}->{$DBXMLFilePatchPath} = {};
        }
        #when ( /^\Qimp_ColorSets\E.*\.xml/){
        #   $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'12-ColorSets'}->{$DBXMLFilePatchPath} = {};
        #}
        when ( /^\Qimp_NavBars\E.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'13-NavBars'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_NavElement\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'14-NavElements'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Search\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'15-Searchs'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_MailType\E.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'16-MailTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Shipping\E.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'17-Shipping'}->{$DBXMLFilePatchPath} = {};
        }
        #when ( /^\Qimp_DefaultShipping\E.*\.xml/){
        #   $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'18-Shipping'}->{$DBXMLFilePatchPath} = {};
        #}
        when ( /^\Qimp_PaymentLogo\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'19-PaymentLogo'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_PaymentType\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'20-PaymentTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_PortalSites\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'21-PortalSites'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_System\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'22-Systems'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Units\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'23-Units'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Taxes\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'24-Taxes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_StyleTemplates\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'25-StyleTemplates'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_StyleGroups\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'26-StyleGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_Countries\_3166\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'27-Countries'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_CountrySubdivisions\_.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'28-CountrySubdivisions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_TaxMatrixDigital\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'29-TaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_DefaultTaxMatrixDigital\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'30-DefaultTaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_TaxMatrixUS\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'31-TaxMatrixUS'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_EcoParticipation\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Helptopics\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_EtrackerActions\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_EbayEnvironment\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qimp_Locales\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_TaxTypes\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_CreditCheckProviders\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_Currencies\_4217\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_OrderProcessSteps\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_Languages.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_Locales.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_CustomizableProducts.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_DistributorShopTypes\.xml$}x){
            #$self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_ImageDBGroups.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_GadgetGroups.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_EbayCallTypes.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_AvataxTaxExemption\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_ERPSystemSoap\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_ERPSystemLexware\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_ERPSystem\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_UnityMBOGuidedTourGeneralConfig\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_UnityMBOGuidedTourLegalConfig\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_UnityMBOGuidedTourProductsConfig\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_DefaultSystemTemplate\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_ShopNotifications\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_PaymentGroups.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_FlexShopType.*\.xml$}x){
            #ignore
        }
        when ( m{^imp\_DHLIntrashipService\.xml$}x) {
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^imp\_Fonts\.xml$}x) {
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        #onPatchImportShopNotificationPendingOrder
        when ( /^\Qimp_ShopNotificationPendingOrder\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportShopNotificationPendingOrder'}->{$DBXMLFilePatchPath} = {};
        }
        #onPatchImportShopNotificationEbayNoOAuthToken
        when ( /^\Qimp_ShopNotificationEbayNoOAuthToken\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportShopNotificationEbayNoOAuthToken'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchImportShopNotificationCookieNotice
        when ( /^\Qimp_ShopNotificationCookieNotice\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportShopNotificationCookieNotice'}->{$DBXMLFilePatchPath} = {};
        }

        # ADD NEW onPatchImportObjects 32-Other *BEFORE* THIS LINE

        #EPG-15720 ignore Translation.*.xml files in patch.xml, automatic delete and import now in Installer.pm
        when ( /^\Qimp_Translation\E\..*\.(\w\w|\w\w-\w\w)\.xml$/){
            #$self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'30-Translations'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchDeleteObjects
        when ( /^\Qdel_Attributes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'01-Attributes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_LastUpdateProfiles\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'02-LastUpdateProfiles'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Actions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }

        #exceptions please remove if files are renamed
        when ( /^\Qdel_GoogleAnalyticsActions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_EtrackerActions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }

        when ( /^\Qdel_Permissions\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'04-Permissions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_PageTypes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'05-PageTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_TemplateTypes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'06-TemplateTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Forms\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'07-Forms'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Features\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'08-Features'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_ClientApplication\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'09-ClientApplication'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_WidgetGroups\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'10-WidgetGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_WidgetTypes\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'11-WidgetTypes'}->{$DBXMLFilePatchPath} = {};
        }
        #when ( /^\Qdel_ColorSets\E.*\.xml/){
        #   $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'12-ColorSets'}->{$DBXMLFilePatchPath} = {};
        #}
        when ( /^\Qdel_NavBars\E.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'13-NavBars'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_NavElement\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'14-NavElements'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Search\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'15-Searchs'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_MailType\E.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'16-MailTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Shipping\E.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'17-Shipping'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_DefaultShipping\E.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'18-Shipping'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_PaymentLogo\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'19-PaymentLogo'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_PaymentType\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'20-PaymentTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_PortalSites\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'21-PortalSites'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_System\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'22-Systems'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Units\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'23-Units'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Taxes\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'24-Taxes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_StyleTemplates\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'25-StyleTemplates'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_StyleGroups\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'26-StyleGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_Countries\_3166\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'27-Countries'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_CountrySubdivisions\_.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'28-CountrySubdivisions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del\_TaxMatrixDigital\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'29-TaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del\_DefaultTaxMatrixDigital\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'30-DefaultTaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del\_TaxMatrixUS\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'31-TaxMatrixUS'}->{$DBXMLFilePatchPath} = {};
        }

        # ADD NEW onPatchDeleteObjects 32-Other *AFTER* THIS LINE
        when ( m{^del\_DHLIntrashipService\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_FlexShopType.*\.xml$}x){
            #ignore
        }
        when ( m{^del\_PaymentGroups.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_ShopNotifications\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_UnityMBOGuidedTourLegalConfig\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_UnityMBOGuidedTourProductsConfig\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_UnityMBOGuidedTourGeneralConfig\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_ERPSystem\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_ERPSystemLexware\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_ERPSystemSoap\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_AvataxTaxExemption\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_EcoParticipation\E.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_Helptopics\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_EtrackerActions\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_EtrackerActions\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_TaxTypes\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_CreditCheckProviders\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_Currencies\_4217\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_OrderProcessSteps\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_Languages\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_Locales\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_CustomizableProducts\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_DistributorShopTypes\.xml$}x){
            #$self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_ImageDBGroups.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_GadgetGroups.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_EbayCallTypes.*\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^\Qdel_EbayEnvironment\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( m{^del\_DefaultSystemTemplate\.xml$}x){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        #EPG-15720 ignore Translation.*.xml files in patch.xml, automatic delete and import now in Installer.pm
        when ( /^\Qdel_Translation\E\..*\.(\w\w|\w\w-\w\w)\.xml$/){
        #    $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'30-Translations'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchDeleteHooks
        when ( /^\Qdel_Hooks\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteHooks'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchImportHooks
        when ( /^\Qimp_Hooks\E.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportHooks'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchDeleteShopChildDependencies
        when ( /^\Qdel_ShopChildDependencies\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteShopChildDependencies'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchImportShopChildDependencies
        when ( /^\Qimp_ShopChildDependencies\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportShopChildDependencies'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchDeleteShopDependencies
        when ( /^\Qdel_Dependencies\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteDependencies'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchImportDependencies
        when ( /^\Qimp_Dependencies\E\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportDependencies'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchImportShopNotificationLegalDocuments
        when ( /^\Qimp_ShopNotificationLegalDocuments\E\.xml$/ ) {
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportShopNotificationLegalDocuments'}->{$DBXMLFilePatchPath} = {};
        }

        default {
            die "can't interpret $DBXMLFile";
        }
    }
}

sub _CorrectPositionsOfPatchXMLXMLStruct{
    my ($XMLStruct) = @_;

    foreach my $Element (@{$XMLStruct->getElements()}){
        given ($Element) {
            ## no critic qw( ProhibitComplexRegexes )
            when ($_ =~ /^\/PatchCartridge\{OldRelease\-\>(?:lp_|cs_|test_)?release\_[0-9\_]*\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '000001')
            }
            when ($_ =~ /\/PatchDBStructure\{PatchDBStructure\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010000')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010100')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}.*\/\"{0,1}File\{.*\/Tables\/.*\.sql/){#"
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010101')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}.*\/\"{0,1}File\{.*\/Procedures\/.*\.sql/){#"
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010102')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}.*\/\"{0,1}File\{.*\/Views\/.*\.sql/){#"
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010102')
            }
            when ($_ =~ /\/Patch\{Patch\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020000')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020100')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteDependencies\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteShopChildDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020200')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteShopChildDependencies\}.*/){
                #ignore
            }
            #when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteDependencies\}$/){
            #    $XMLStruct->setXMLStructAttribute($Element,'Position', '020300')
            #}
            #when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteDependencies\}.*/){
            #    #ignore
            #}
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteHooks\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020300')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020400')
            }

            # ADD NEW onPatchDeleteObjects *AFTER* THIS LINE
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*DHLIntrashipService\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020401')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PaymentGroups[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020402')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ShopNotifications\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020403')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourProductsConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020404')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourLegalConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020405')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourGeneralConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020406')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ERPSystem\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020407')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ERPSystemLexware\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020408')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ERPSystemSoap\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020409')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*AvataxTaxExemption\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020410')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*EbayCallTypes\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020411')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Countries_3166\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020412')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*CustomizableProducts\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020413')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Locales\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020414')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Languages\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020415')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*OrderProcessSteps\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020416')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*CreditCheckProviders\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020417')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Currencies_4217\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020418')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*EtrackerActions\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020419')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Helptopics\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020420')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*EcoParticipation[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020421')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Taxes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020422')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Units[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020423')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*System[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020424')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PortalSites[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020425')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PaymentType[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020426')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PaymentLogo[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020427')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Shipping[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020428')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*MailType[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020429')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Search[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020430')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*NavElement[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020431')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*NavBars[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020432')
            }
            #when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ColorSets[a-zA-Z0-9\_]*\.xml/){
            #    $XMLStruct->setXMLStructAttribute($Element,'Position', '020423')
            #}
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Features[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020433')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*WidgetTypes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020434')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*WidgetGroups[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020435')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Forms[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020436')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*StyleGroups[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020437')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*StyleTemplates[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020438')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*TemplateTypes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020439')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PageTypes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020440')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Permissions[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020441')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Actions[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020442')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*LastUpdateProfiles[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020443')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Attributes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020444')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteHooks\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportHooks\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020500')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportHooks\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020600')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportDependencies\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020700')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Attributes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020701')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*LastUpdateProfiles[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020702')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Actions[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020703')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Permissions[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020704')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PageTypes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020705')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*TemplateTypes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020706')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*StyleTemplates[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020707')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*StyleGroups[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020708')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Forms[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020709')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Features[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020710')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*WidgetGroups[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020711')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*WidgetTypes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020712')
            }
            #when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ColorSets[a-zA-Z0-9\_]*\.xml/){
            #    $XMLStruct->setXMLStructAttribute($Element,'Position', '020710')
            #}
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*NavBars[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020713')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*NavElement[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020714')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Search[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020715')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*MailType[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020716')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Shipping[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020717')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PaymentLogo[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020718')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PaymentType[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020719')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PortalSites[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020720')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*System[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020721')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Units[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020722')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Taxes[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020723')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EcoParticipation[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020724')
            }
            #when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*s\.xml/){
            #    $XMLStruct->setXMLStructAttribute($Element,'Position', '020723')
            #}
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Helptopics\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020725')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EtrackerActions\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020726')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EbayEnvironment\.xml/){ #"
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020727')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Locales\.xml/){ #"
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020728')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*CreditCheckProviders\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020729')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Currencies_4217\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020730')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*OrderProcessSteps\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020731')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Languages\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020732')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Countries_3166\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020733')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*CustomizableProducts\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020734')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EbayCallTypes\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020735')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*AvataxTaxExemption\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020736')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ERPSystemSoap\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020737')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ERPSystemLexware\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020738')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ERPSystem\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020739')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourGeneralConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020740')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourLegalConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020741')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourProductsConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020742')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotifications\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020743')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PaymentGroups[a-zA-Z0-9\_]*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020744')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*DHLIntrashipService\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207445')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationPendingOrder\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207446')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationEbayNoOAuthToken\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207447')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationCookieNotice\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207448')
            }

            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationLegalDocuments\.xml/) {
                $XMLStruct->setXMLStructAttribute($Element, 'Position', '0207449')
            }
            # ADD NEW onPatchImportObjects *BEFORE* THIS LINE

            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportShopChildDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021000')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchImportShopChildDependencies\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteWebRootFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021100')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteWebRootFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeletePrivateFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021300')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeletePrivateFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteSchedulerFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021400')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteSchedulerFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteSharedConfigFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021450')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteSharedConfigFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyWebRootFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021500')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyWebRootFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyPrivateFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021600')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyPrivateFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopySchedulerFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021700')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopySchedulerFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopySharedConfigFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021750')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopySharedConfigFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchCartridgeDir\{PatchCartridgeDir\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '030000')
            }
            when ($_ =~ /\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '030100')
            }
            when ($_ =~ /\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '030200')
            }
            when ($_ =~ /\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}.*/){
                #ignore
            }


            when ($_ =~ /\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '040000')
            }
            when ($_ =~ /\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '040100')
            }
            when ($_ =~ /\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '040200')
            }
            when ($_ =~ /\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}.*/){
                #ignore
            }

            when ($_ =~ /\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '050000')
            }
            when ($_ =~ /\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '050100')
            }
            when ($_ =~ /\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '050200')
            }
            when ($_ =~ /\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}.*/){
                #ignore
            }


            when ($_ =~ /\/PatchStoreTypeDir\{PatchStoreTypeDir\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060000')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeletePublicFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060100')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeletePublicFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteJavascriptFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060200')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchDeleteJavascriptFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyPublicFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060300')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyPublicFile\}.*/){
                #ignore
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyJavascriptFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060400')
            }
            when ($_ =~ /\/PatchFunction\{Method\-\>onPatchCopyJavascriptFile\}.*/){
                #ignore
            }
            default{
                die "\nunknown element $Element in xml structure of patch.xml"
            }
        }
    }
    return;
}

sub GetVersionByReleaseTag{
    my ($ReleaseTag) = @_;
    my $CartridgeVersion = $ReleaseTag;
    $CartridgeVersion =~ s/^(?:lp_|cs_|test_)?release_//;
    $CartridgeVersion =~ s/\_/\./g;

    return $CartridgeVersion;
}

sub _GetPatchTargetName {
    my ($CartridgeFolder, $OldCartridgeVersion) = @_;

    $OldCartridgeVersion =~ s/\./\_/g;

    return $CartridgeFolder . '/Patches/Patch_' . $OldCartridgeVersion;
}

sub _GetCVSPath {
    my ($Folder) = @_;

    my $CVSPath =$Folder;

    $CVSPath =~ s/.*(\\|\/)DE_EPAGES//g;

    die "\nunable to replace chunk from path, cannot create CVS path" if ($CVSPath eq $Folder);

    return 'RND/epages/epages5/AllOS/Cartridges/DE_EPAGES' . $CVSPath;
}

sub addCartridgeTag{
    my $self = shift;
    my ($PatchFile, $CartridgeName) = @_;

    my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
    my $XMLStruct = $ImportHandler->getXMLStructure($PatchFile);

    foreach my $Element (@{$XMLStruct->getElements()}){
        next if($XMLStruct->getXMLStructAttribute($Element,'Position') eq '1');
        $XMLStruct->setXMLStructAttribute($Element,'Position',$XMLStruct->getXMLStructAttribute($Element,'Position') + 1 );
    }

    my $CompanyName = $self->{'CompanyName'};

    $XMLStruct->addElement(
        '/Cartridge',
        {},
        {
            'Position' => '2',
            'TagName' => 'Cartridge',
            'Class' => 'Cartridge',
            'NonEffectiveForSubPathes' => '1'
        },
        {'Package' => "$CompanyName::" . $CartridgeName}
    );

    my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
    $XMLStructExporter->exportXML($PatchFile,$XMLStruct);
}

1;
