#========================================================================================
# package      DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandler
# state        public
#----------------------------------------------------------------------------------------
# description  This the install handler for all patch install functions.
#========================================================================================
package DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandler;
use base qw( Exporter );

use strict;
use warnings;
use diagnostics;
use feature 'switch';

use Config::IniFiles;
use Data::Dumper;
use Date::Format qw(time2str);
use File::Basename qw ( dirname basename fileparse );
use File::Path;
use JSON::XS;
use List::MoreUtils qw ( any );
use Proc::Background;
use Sort::Versions;
use Term::ReadKey;
use Time::HiRes qw (time);

use DE_EPAGES::Cartridge::API::XML::CartridgesExportDriver qw ( ExportCartridges );
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::File qw (
    CopyDirectory
    CopyFile
    CreateDirectory
    DeleteDirectory
    DeleteFile
    ExistsDirectory
    ExistsFile
    GetFileContent
    GetFileContentLines
    GetInfoDirectory
    MoveFile
    WriteFile
);
use DE_EPAGES::Core::API::CartridgeConfig qw(CartridgeConfig);
use DE_EPAGES::Core::API::IniConfig;
use DE_EPAGES::Core::API::SystemCommand qw(SystemCommand);
use DE_EPAGES::Database::API::Config qw ( GetBusinessUnits StoreType );
use DE_EPAGES::Database::API::Connection qw ( RunOnStore );
use DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBContent;
use DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBStructure;
use DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerFiles;
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadRootObject ExistsObjectByPath);
use DE_EPAGES::Object::API::XML::ObjectExportDriver qw ( ExportObjects );
use DE_EPAGES::Object::API::XMLStructureExportDriver;
use DE_EPAGES::Object::API::XMLStructureImportHandler;
use DE_EPAGES::ShopConfiguration::API::XML::CartridgeRefExportDriver qw ( ExportCartridgeRefs );
use DE_EPAGES::Trigger::API::XML::HookExportDriver qw ( ExportHooks );
use DE_EPAGES::WebInterface::API::ServerConfig qw ( ServerConfig );
use DE_EPAGES::XML::API::XML::CartridgeInstallerImportHandler qw(GetStoreTypeCartridges);

use constant LINE_OFFSET => 30;
use constant MAX_PARALLEL_PATCH_SECTIONS => 20;

#known db exceptions
use constant DBPatchExceptions => {
  'ASPoolDB'            => 1,
  'Backup'              => 1,
  'EbayCategories'      => 1,
  'PatchSystemInfoDB'   => 1,
  'System'              => 1,
};

#categories
## no critic (ProhibitAutomaticExportationNonDAL)
our @EXPORT = qw(FileCategories DBStructureCategories DBContentCategories
  DBXMLGeneralStoreSections DBXMLExclusiveStoreSections DBXMLExclusiveSiteSections
);
 ## use critic

use constant FileCategories => [
    'Cartridges',
    'Perl',
    'RequestRouter',
    'Service',
    'Stores',
    'WebAdapter',
    'WebRoot',
    'Config',
    'Monitor',
    'DiffCartridgesWebRoot',
    'j',
    'CTMPLFolder',
    'var-epages'
];

use constant DBStructureCategories => [
    'DBSchema',
    'ASPoolDB',
    'epagesj',
    'EbayCategoriesDB',
    'System'
];

use constant DBContentCategories => [
    'ShopChildDependencies',
    'BCP',
    'MySQLConfig',
    'Locales'
];

use constant DBXMLGeneralStoreSections => [
    'XML-Cartridges',
    'XML-Classes',
    'XML-ClientApplications',
    'XML-Hooks',
    'XML-MailTypes',
    'XML-MailTypeTemplates',
    'XML-PageTypes',
    'XML-RestServices',
    'XML-System',
    'XML-Units',
    'XML-WebServices'
];

use constant DBXMLExclusiveStoreSections => [
    'XML-ContentElements',
    'XML-ContentElementGroups',
    'XML-CreditCheckProviders',
    'XML-EcoParticipationCategories',
    'XML-EbayEnvironments',
    'XML-EbayTemplates',
    'XML-Features',
    'XML-FileTypes',
    'XML-ImageDBGroups',
    'XML-MailTypeTemplateGroups',
    'XML-NavBars',
    'XML-NavElements',
    'XML-NavElementGroups',
    'XML-OrderProcessSteps',
    'XML-PagerSettings',
    'XML-PaymentLogos',
    'XML-PaymentTypes',
    'XML-Portals',
    'XML-ShippingTypes',
    'XML-ShopMailTypeTemplates',
    'XML-ShopTypeRefs',
    'XML-StyleBodyImages',
    'XML-StyleHeaderImages',
    'XML-StyleGroups',
    'XML-StyleTemplates',
    'XML-SystemFonts',
    'XML-ExternalShipping',
    'XML-TaxMatrices',
    'XML-TaxTypes',
    'XML-WidgetGroups',
    'XML-WidgetTypes'
];

use constant DBXMLExclusiveSiteSections => [
    'XML-CartridgeRefs',
    'XML-DistributorUsers',
    'XML-FeatureRefs',
    'XML-FeaturePacks',
    'XML-Stores',
    'XML-Users'
];

use constant DBXMLCategories => [
    @{ +DBXMLGeneralStoreSections },
    @{ +DBXMLExclusiveStoreSections },
    @{ +DBXMLExclusiveSiteSections },
];

sub new {
  my $class   = shift;
  my ($hOptions) = @_;
  my $self = bless $hOptions, $class;
  return $self;
}

#========================================================================================
# function     exportReport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->exportReport( $OutputFolder, $Role, $aReportSections, $aStore, $SkriptDir, $Background, $Recursive );
#----------------------------------------------------------------------------------------
# description  exports the report of the report sections
#               will ask permission to delete output folder
#               sets standard values for role and report section if undef
#               is able to call background processes
#----------------------------------------------------------------------------------------
# input        $OutputFolder | target folder for output | string
# input        $Role | role, regulate the type of output (optional) | string
# input        $aReportSections | report sections, regulate the type of output (optional) | ref array string
# input        $aStore | stores, regulate the type of output (optional) | ref array string
# input        $SkriptDir | directory of getSystemInfo script | string
# input        $Background | use background processes to fetch every caregory simultaneously | boolean
# input        $Recursive | this process is called recursive | boolean
#========================================================================================
sub exportReport {
    my $self = shift;
    my ( $OutputFolder, $Role, $aReportSections, $aStore, $SkriptDir, $Background, $Recursive ) = @_;

    print $aReportSections->[0] . time2str("  %T\n", time) if $aReportSections;

    #get operation system name
    my $OS = $^O;

    #set parallel work variable
    $self->{'work_parallel'} = $Background;

    #if $aReportSections is undef: report section default case, exportReport for all sections
    $aReportSections = [
        @{ +FileCategories },
        'DBXMLExclusiveStoreSections',
        @{ +DBStructureCategories },
        'DBXMLExclusiveSiteSections',
        @{ +DBContentCategories },
        'DBXMLGeneralStoreSections',
    ] unless $aReportSections;

    #test whether passed stores is empty
    unless ($aStore) {
        #get all stores
        my $aDatabases = GetBusinessUnits({'conf' => 1});

        #for each existing database
        foreach my $Database (@$aDatabases) {

            #skip for known exceptions
            next if ( DBPatchExceptions->{$Database} );

            #add Store
            push (@$aStore, $Database);
        }
    }

    #default output path
    unless ($OutputFolder){
        if ( $OS eq 'linux' ) {
            $OutputFolder = '/temp/SystemInfo';
        }
        else{
            $OutputFolder = 'C:/temp/SystemInfo';
        }
    }

    #ref array containing the three valid roles
    my $aRoles = [ 'QA', 'RND' ];

    #get amount of report sections
    my $Count = @$aReportSections;

    #check if parallel work is forced
    if ( $self->{'work_parallel'} ) {
        #count given rport sections
        $Count = @$aReportSections;

        #check if more than one report section exists
        if ( $Count > 1 ) {
            #call multiple funktion
            $self->_exportByCallingParallelProcessesInBack( $OutputFolder, $aRoles, $SkriptDir, $aReportSections, $Role );
        }
        else {
            #call single function
            $self->_exportFromParallelProcess( $Recursive, $OutputFolder, $aRoles, $aReportSections->[0], $Role, $aStore );
        }
    }
    else {
        #dont use parrallel functions
        $self->_exportSequential( $OutputFolder, $aRoles, $aReportSections, $aStore, $Role );
    }
}

#========================================================================================
# function     _exportSequential
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_exportSequential( $OutputFolder, $aRoles, $aReportSections, $aStore, $Role ) ;
#----------------------------------------------------------------------------------------
# description  exports the report of the report sections sequentially in one process
#               will ask permission to delete output folder
#               sets standard values for role and report section if undef
#----------------------------------------------------------------------------------------
# input        $OutputFolder | target folder for output | string
# input        $aRoles | array of all roles | ref arry string
# input        $aReportSections | report sections, regulate the type of output (optional) | ref array string
# input        $aStore | stores, regulate the type of output (optional) | ref array string
# input        $Role | role, regulate the type of output (optional) | string
#========================================================================================
sub _exportSequential {
    my $self = shift;
    my ( $OutputFolder, $aRoles, $aReportSections, $aStore, $Role ) = @_;

    #aks user permission to delete output directory
    $self->_removeOutputDir( $OutputFolder, $aRoles );

    #get intersection of passed report sections and all file categories
    my $aIntersection = $self->_intersection( $aReportSections, \@{ FileCategories() } );

    #initialize file report handler
    my $FileReportHandler = DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerFiles->new( $aIntersection, $OutputFolder, $aStore );

    #get intersection of passed report sections and all dbstructure categories
    $aIntersection = $self->_intersection( $aReportSections, \@{ DBStructureCategories() } );

    #initialize file db structure report handler
    my $DBStructureReportHandler = DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBStructure->new( $aIntersection, $aStore );

    #get intersection of passed report sections and all dbcontent categories
    $aIntersection = $self->_intersection( $aReportSections, \@{ DBContentCategories() } );

    #initialize file db dbcontent report handler
    my $DBContentReportHandler = DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBContent->new( $aIntersection, $aStore );

    #export xml files
    foreach my $Section (@$aReportSections){
        given ($Section){
            when('DBXMLGeneralStoreSections'){
                #call xml export function
                $self->_exportGeneralStoreSectionsXMLContent($OutputFolder, $aStore);
            }
            when('DBXMLExclusiveStoreSections'){
                #call xml export function
                $self->_exportExclusiveStoreSectionsXMLContent($OutputFolder, $aStore);
            }
            when('DBXMLExclusiveSiteSections'){
                #call xml export function
                $self->_exportExclusiveSiteSectionsXMLContent($OutputFolder, $aStore);
            }
            when (DBXMLCategories){
                #call xml export function
                $self->_exportXMLContent($OutputFolder, $Section, $aStore);
            }
        }
    }

    #if $Role is undef: role default case, print report for all roles
    unless ( defined $Role ) {

        #for each role
        foreach my $Role (@$aRoles) {
            $self->_printExport( $OutputFolder, $Role, $aReportSections, $aStore, $FileReportHandler, $DBStructureReportHandler, $DBContentReportHandler );
        }
    }

    #test if $Role is valid
    elsif ( ( !grep { $_ eq $Role } @$aRoles ) ) {
        die "$Role is not a valid Role.\n";
    }
    else {
        $self->_printExport( $OutputFolder, $Role, $aReportSections, $aStore, $FileReportHandler, $DBStructureReportHandler, $DBContentReportHandler );
    }

    #delete temp directory
    if ( -e "$OutputFolder/Temp" ) {
        DeleteDirectory( "$OutputFolder/Temp", { 'recursive' => 1 } );
    }

    _printRunTime($FileReportHandler, 'File Category');
    _printRunTime($DBContentReportHandler, 'DBContent Category');
    _printRunTime($DBStructureReportHandler, 'DBStructure Category');
    _printRunTime($self, 'Writing');
}

sub _countRunningProcesses {
    my $self = shift;
    my ($aProcesses) = @_;

    #get placeholders
    my $hRunningProcesses = {};

    #for every process
    foreach my $Process (@$aProcesses){
        #check if process is already running
        if ($Process->alive()){
            $hRunningProcesses->{$Process->pid()} = 1;
        }
    }

    #return the count
    return scalar keys %{$hRunningProcesses};
}

sub _isSectionExecutable{
    my $self = shift;
    my ($Section) = @_;

    given ($Section) {

        when ('RequestRouter') {
            return ExistsDirectory($ENV{'EPAGES'} . "/$Section");
        }
        when ('XML-DistributorUsers'){
            #return null unless site exists
            return 0 unless ($self->_storeExists('Site'));

            my $Return;

            RunOnStore(
                Store      => 'Site',
                DBPassword => undef,
                Sub        => sub {
                    $Return = ExistsObjectByPath('/Providers/Distributor');
                }
            );

            return $Return;
        }
        default {
            return 1;
        }
    }
}

sub _updateRunTime {
    my $self = shift;
    my ($StartTime, $Section) = @_;

    my $RunTime = $self->{'runtime'}->{ "Write $Section" } // 0;
    $self->{'runtime'}->{ "Write $Section" } = sprintf( '%.3f', time - $StartTime + $RunTime);
}

sub _printRunTime {
    my ($Handler, $Heading) = @_;
    return unless $Handler;
    my $hTime = $Handler->{'runtime'};
    return unless scalar keys %$hTime;

    print "\n$Heading:\n" if $Heading;
    foreach my $Section ( keys %$hTime ) {
        print "$Section:" . ' ' x ( LINE_OFFSET - length($Section) ) . "$hTime->{$Section}\n";
    }
}

#========================================================================================
# function     _exportByCallingParallelProcessesInBack
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_exportByCallingParallelProcessesInBack( $OutputFolder, $aRoles, $SkriptDir, $aReportSections, $Role );
#----------------------------------------------------------------------------------------
# description  exports the report of the report sections parallel by calling background processes
#               will ask permission to delete output folder
#----------------------------------------------------------------------------------------
# input        $OutputFolder | target folder for output | string
# input        $aRoles | array of all roles | ref arry string
# input        $SkriptDir | directory of getSystemInfo script | string
# input        $aReportSections | report sections, regulate the type of output (optional) | ref array string
# input        $Role | role, regulate the type of output (optional) | string
#========================================================================================
sub _exportByCallingParallelProcessesInBack {
    my $self = shift;
    my ( $OutputFolder, $aRoles, $SkriptDir, $aReportSections, $Role ) = @_;

    #aks user permission to delete output directory
    $self->_removeOutputDir( $OutputFolder, $aRoles );

    #get perl.exe path
    my $PerlExe = $^X;

    #set systeminfo script path
    my $SytemInfoScript = "$SkriptDir/getSystemReport.pl";
    my $Command;

    #proccesses array
    my $aProccesses;

    #hash containing process id as keys and section names as values
    my $hProcesssNames;

    #create stack array
    my @SectionStack = @$aReportSections;

    my $WaitIntervall = 3;

    #do until stack is empty
    while( scalar @SectionStack ) {

        #check for limit of simultanious processes
        if( $self->_countRunningProcesses($aProccesses) < MAX_PARALLEL_PATCH_SECTIONS ) {

            #take the first Store from stack
            my $Section = shift(@SectionStack);

            #case 1: execute for a certain role
            if ( defined $Role ) {
                #get commandline
                $Command = "$PerlExe $SytemInfoScript -section $Section -role $Role -outputpath $OutputFolder -recursive -p";
            }

            #case 2: execute for all roles
            else {
                #get commandline
                $Command = "$PerlExe $SytemInfoScript -section $Section -outputpath $OutputFolder -recursive -p";
            }

            #start parallel process
            my $Process = Proc::Background->new($Command);

            #save process in array
            push (@$aProccesses, $Process);

            #save process id and section name in hash
            $hProcesssNames->{ $Process->pid() } = $Section;

            print "\nstarted background process '" . $Process->pid() . '\' to get section ' . $hProcesssNames->{ $Process->pid() } . " \n";

        }
        else{
            print "\n\nwait $WaitIntervall s to create new parallel sub processes \n";
            sleep $WaitIntervall;
        }
    }

    #for each process in  the process array
    foreach my $Process (@$aProccesses) {

        print "\nwait for background process '" . $Process->pid() . '\' to finish section ' . $hProcesssNames->{ $Process->pid() } . " \n";

        #wait for background process
        $Process->wait();
    }

    #foreach section
    foreach my $Section (@$aReportSections) {

        die "\nSection $Section was not successful, sub process crashed" unless (ExistsFile("$OutputFolder/Temp/Sections/$Section.txt"));

        #get file content to control if section is succussfully finished
        my $rContent = GetFileContent("$OutputFolder/Temp/Sections/$Section.txt");

        #test if section successfully done
        if ( !$rContent =~ /$Section: done/ ) {
            die "\nSection $Section was not successful, sub process crashed";
        }
        else {
            print "Section $Section was successful.\n";
        }
    }

    #delete temp directory
    if ( -e "$OutputFolder/Temp" ) {
        DeleteDirectory( "$OutputFolder/Temp", { 'recursive' => 1 } );
    }
}

#========================================================================================
# function     _exportFromParallelProcess
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_exportFromParallelProcess( $Recursive, $OutputFolder, $aRoles, $ReportSection, $Role, $aStore );
#----------------------------------------------------------------------------------------
# description  exports the report of the report sections as a called background process
#               will ask permission to delete output folder
#----------------------------------------------------------------------------------------
# input        $Recursive | this process is called recursive | boolean
# input        $OutputFolder | target folder for output | string
# input        $aRoles | array of all roles | ref arry string
# input        $ReportSection | report section | string
# input        $Role | role, regulate the type of output (optional) | string
# input        $aStore | stores, regulate the type of output (optional) | ref array string
#========================================================================================
sub _exportFromParallelProcess {
    my $self = shift;
    my ( $Recursive, $OutputFolder, $aRoles, $ReportSection, $Role, $aStore ) = @_;

    #test if this is a recursive call
    unless ($Recursive) {

        #aks user permission to delete output directory
        $self->_removeOutputDir( $OutputFolder, $aRoles );
    }

    #check if section is executable
    if ($self->_isSectionExecutable($ReportSection)){
        my $FileReportHandler;
        my $DBStructureReportHandler;
        my $DBContentReportHandler;

        given ( $ReportSection ) {
            when (FileCategories) {
                #initialize file report handler
                $FileReportHandler = DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerFiles->new( [$ReportSection], $OutputFolder, $aStore );
            }
            when (DBStructureCategories) {
                #initialize file db structure report handler
                $DBStructureReportHandler = DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBStructure->new( [$ReportSection], $aStore );
            }
            when (DBContentCategories) {
                #initialize file db dbcontent report handler
                $DBContentReportHandler = DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBContent->new( [$ReportSection], $aStore );
            }
            when('DBXMLGeneralStoreSections'){
                #call xml export function
                $self->_exportGeneralStoreSectionsXMLContent($OutputFolder, $aStore);
            }
            when('DBXMLExclusiveStoreSections'){
                #call xml export function
                $self->_exportExclusiveStoreSectionsXMLContent($OutputFolder, $aStore);
            }
            when('DBXMLExclusiveSiteSections'){
                #call xml export function
                $self->_exportExclusiveSiteSectionsXMLContent($OutputFolder, $aStore);
            }
            when (DBXMLCategories){
                #call xml export function
                $self->_exportXMLContent($OutputFolder, $ReportSection, $aStore);
            }
        }

        #if $Role is undef: role default case, print report for all roles
        unless ( defined $Role ) {

            #for each role
            foreach my $Role (@$aRoles) {
                $self->_printExport( $OutputFolder, $Role, [$ReportSection], $aStore, $FileReportHandler, $DBStructureReportHandler, $DBContentReportHandler );
            }
        }
        elsif ( ( !grep { $_ eq $Role } @$aRoles ) ) {
            #role is not valid
            die "$Role is not a valid Role.\n";
        }
        else {
            #valid role print export
            $self->_printExport( $OutputFolder, $Role, [$ReportSection], $aStore, $FileReportHandler, $DBStructureReportHandler, $DBContentReportHandler );
        }

        #delete temp directory
        if ( -e "$OutputFolder/Temp" && !$Recursive ) {
            DeleteDirectory( "$OutputFolder/Temp", { 'recursive' => 1 } );
        }

        _printRunTime($_) foreach ($FileReportHandler, $DBContentReportHandler, $DBStructureReportHandler);
    }
    elsif ($self->{'show_missing_section'}){
        die "\ncant call section '$ReportSection' because it is not executeable\n";
    }
    else {
        print "$ReportSection:" . ' ' x ( LINE_OFFSET - length($ReportSection) ) . "skipped\n";
    }

    my $Content = "$ReportSection: done\n";
    WriteFile("$OutputFolder/Temp/Sections/$ReportSection.txt", \$Content, {create_directory=>1, append => 1});
}

sub _exportGeneralStoreSectionsXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore) = @_;

    foreach my $Section ( @{ +DBXMLGeneralStoreSections } ){
        $self->_exportXMLContent($OutputFolder, $Section, $aStore);
    }
}

sub _exportExclusiveStoreSectionsXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore) = @_;

    foreach my $Section ( @{ +DBXMLExclusiveStoreSections } ){
        $self->_exportXMLContent($OutputFolder, $Section, $aStore);
    }
}

sub _exportExclusiveSiteSectionsXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore) = @_;

    foreach my $Section ( @{ +DBXMLExclusiveSiteSections } ){
        if ($self->_isSectionExecutable($Section)){
            $self->_exportXMLContent($OutputFolder, $Section, $aStore);
        }
        elsif ($self->{'show_missing_section'}){
            die "\ncant call section '$Section' because it is not executeable\n";
        }
        else{
            print "skipping $Section\n";
        }
    }
}


sub _exportXMLContent{
    my $self = shift;
    my ($OutputFolder, $Section, $aStore) = @_;

    my $StartTime = time;

    given ( $Section ) {
        when ('XML-Cartridges') {
            #export Cartridges XML Content
            $self->_exportCartridgesXMLContent($OutputFolder, $aStore);
        }
        when ('XML-Hooks') {
            #export Cartridges XML Content
            $self->_exportHooksXMLContent($OutputFolder, $aStore);
        }
        when ('XML-System') {
            #export Cartridges XML Content
            $self->_exportSystemXMLContent($OutputFolder, $aStore);
        }
        when ('XML-DistributorUsers'){
            #export distributorUsers XML Content
            $self->_exportDistributorUsersXMLContent($OutputFolder, $aStore);
        }
        when ('XML-CartridgeRefs'){
            #export Cartridges XML Content
            $self->_exportCartridgeRefsXMLContent($OutputFolder, $aStore);
        }
        when ('XML-ExternalShipping'){
            #export Cartridges XML Content
            $self->_exportExternalShippingXMLContent($OutputFolder, $aStore);
        }
        when ('XML-TaxMatrices'){
            #export Cartridges XML Content
            $self->_exportTaxMatricesXMLContent($OutputFolder, $aStore);
        }
        when (DBXMLGeneralStoreSections){
            #export Cartridges XML Content
            $self->_exportGeneralStoreSectionXMLContent($OutputFolder, $aStore, $Section);
        }
        when (DBXMLExclusiveStoreSections){
            #export Cartridges XML Content
            $self->_exportExclusiveStoreSectionXMLContent($OutputFolder, $aStore, $Section);
        }
        when (DBXMLExclusiveSiteSections){
            #export Cartridges XML Content
            $self->_exportExclusiveSiteSectionXMLContent($OutputFolder, $aStore, $Section);
        }
    }

    $self->_updateRunTime($StartTime, $Section);
    return;
}

sub _storeExists{
    my $self = shift;
    my ($StoreName) = @_;

    my $aStore = GetBusinessUnits({'conf' => 1});

    return 1 if ($StoreName ~~ @$aStore);
    return 0;
}

sub _exportCartridgesXMLContent{
    my $self = shift;
    my ( $OutputFolder, $aStore ) = @_;

    my $Encoding = 'utf-8';

    foreach my $StoreName (@$aStore){
        RunOnStore(
            Store      => $StoreName,
            DBPassword => undef,
            Sub        => sub {

                #set export filepath for cartridges.xml
                my $FilePath = "$OutputFolder/Temp/$StoreName/XML/Cartridges.xml";
                print "export Cartridges.xml for $StoreName\n";

                #export cartridges.xml
                ExportCartridges( $FilePath, $Encoding );
            }
        ) if ($self->_storeExists($StoreName));
    }
}

sub _exportHooksXMLContent{
    my $self = shift;
    my ( $OutputFolder, $aStore ) = @_;

    my $Encoding = 'utf-8';

    foreach my $StoreName (@$aStore){
        RunOnStore(
            Store      => $StoreName,
            DBPassword => undef,
            Sub        => sub {

                #set export filepath for hooks.xml
                my $FilePath = "$OutputFolder/Temp/$StoreName/XML/Hooks.xml";
                print "export Hooks.xml for $StoreName\n";

                #export hooks.xml
                ExportHooks( $FilePath, $Encoding );
            }
        ) if ($self->_storeExists($StoreName));
    }
}

sub _exportSystemXMLContent{
    my $self = shift;
    my ( $OutputFolder, $aStore ) = @_;

    foreach my $StoreName (@$aStore){
        #set writer for system.xml
        my $Writer = 'DE_EPAGES::XML::API::NewLineSAXWriter';

        #set export filepath
        my $FilePath = "$OutputFolder/Temp/$StoreName/XML/System.xml";

        print "export System.xml for $StoreName\n";

        #export system.xml
        $self->_exportMainXML( $FilePath, $StoreName, ['/'], '/', $Writer, '0' );
    }
}

sub _exportGeneralStoreSectionXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore, $Section) = @_;

    $Section =~ s/^XML-//;

    foreach my $StoreName (@$aStore){
        #set export filepath
        my $FilePath = "$OutputFolder/Temp/$StoreName/XML/$Section.xml";
        print "export $Section.xml for $StoreName\n";

        #export xml files
        $self->_exportMainXML( $FilePath, $StoreName, [$Section], '/' );
    }
}

sub _exportExclusiveStoreSectionXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore, $Section) = @_;

    $Section =~ s/^XML-//;

    foreach my $StoreName (@$aStore){

        next if ($StoreName eq 'Site');

        #set export filepath
        my $FilePath = "$OutputFolder/Temp/$StoreName/XML/$Section.xml";
        print "export $Section.xml for $StoreName\n";

        #export xml files
        $self->_exportMainXML( $FilePath, $StoreName, [$Section], '/' );
    }
}

sub _exportCartridgeRefsXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore, $Section) = @_;

    $Section =~ s/^XML-//;

    my $Encoding = 'utf-8';

    foreach my $StoreName (@$aStore){

        #only export on Site
        next unless ($StoreName eq 'Site');

        #run on site
        RunOnStore(
            Store      => $StoreName,
            DBPassword => undef,
            Sub        => sub {

                #set export filepath for cartridgerefs.xml
                my $FilePath = "$OutputFolder/Temp/$StoreName/XML/CartridgeRefs.xml";

                print "export CartridgeRefs.xml\n";

                #export cartridgerefs.xml
                ExportCartridgeRefs( $FilePath, $Encoding );
            }
        ) if ($self->_storeExists($StoreName));
    }
}

sub _exportExternalShippingXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore) = @_;

    foreach my $StoreName (@$aStore){

        #only export on Store
        next if ($StoreName eq 'Site');

        #run on site
        RunOnStore(
            Store      => $StoreName,
            DBPassword => undef,
            Sub        => sub {

                #set export filepath for externalshipping.xml
                my $FilePath = "$OutputFolder/Temp/$StoreName/XML/ExternalShipping.xml";

                print "export ExternalShipping.xml\n";

                #export externalshipping.xml
                $self->_exportMainXML( $FilePath, $StoreName, [ 'ShippingFulfillmentTypes', 'ShippingPickupStations' ], '/' );
            }
        ) if ($self->_storeExists($StoreName));
    }
}

sub _exportTaxMatricesXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore) = @_;

    my $Password;
    my $Encoding  = 'utf-8';
    my $Recursive = 1;
    my $WithDefaults;
    my $WithoutGuids        = 1;
    my $WithoutCreationDate = 1;
    my $WithoutFilePathPart;
    my $WriterPackage;

    my %PARAMS = (
        'encoding'            => $Encoding,
        'recursive'           => $Recursive,
        'withdefaults'        => $WithDefaults,
        'withoutguids'        => $WithoutGuids,
        'withoutcreationdate' => $WithoutCreationDate,
        'withoutfilepathpart' => $WithoutFilePathPart,
        'writer'              => $WriterPackage,
    );

    foreach my $StoreName (@$aStore){

        #only for Stores
        next if ($StoreName eq 'Site');

        RunOnStore(
            Store      => $StoreName,
            DBPassword => $Password,
            Sub        => sub {
                #set export filepath for externalshipping.xml
                my $FilePath = "$OutputFolder/Temp/$StoreName/XML/TaxMatrices.xml";

                my $TaxMatrixClass    = LoadClassByAlias('TaxMatrix');
                my $aTaxMatrixObjects = LoadRootObject->childrenByClass($TaxMatrixClass);
                my @TaxMatrixObjectNames;

                foreach my $TaxMatrixObject (@$aTaxMatrixObjects) {
                    push( @TaxMatrixObjectNames, $TaxMatrixObject->get('Alias') );
                }

                @TaxMatrixObjectNames = sort @TaxMatrixObjectNames;

                ExportObjects( '/', $FilePath, \%PARAMS, \@TaxMatrixObjectNames, \&_transForm );

            },
        ) if ($self->_storeExists($StoreName));
    }

    return;
}

sub _exportDistributorUsersXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore, $Section) = @_;

    $Section =~ s/^XML-//;

    foreach my $StoreName (@$aStore){

        next unless ($StoreName eq 'Site');

        #set export filepath
        my $FilePath = "$OutputFolder/Temp/$StoreName/XML/DistributorUsers.xml";
        print "export DistributorUsers.xml for $StoreName\n";

        #export xml files
        $self->_exportMainXML( $FilePath, $StoreName, ['Users'], '/Providers/Distributor' );
    }
}

sub _exportExclusiveSiteSectionXMLContent{
    my $self = shift;
    my ($OutputFolder, $aStore, $Section) = @_;

    $Section =~ s/^XML-//;

    foreach my $StoreName (@$aStore){

        next unless ($StoreName eq 'Site');

        #set export filepath
        my $FilePath = "$OutputFolder/Temp/$StoreName/XML/$Section.xml";
        print "export $Section.xml for $StoreName\n";

        #export xml files
        $self->_exportMainXML( $FilePath, $StoreName, [$Section], '/' );
    }
}

#========================================================================================
# function     _printExport
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_printExport( $OutputFolder,$Role, $aReportSections, $aStore );
#----------------------------------------------------------------------------------------
# description  stores the report of the passed report sections into textfile
#----------------------------------------------------------------------------------------
# input        $OutputFolder | target folder for output | string
# input        $Role | role, regulate the type of output (optional) | string
# input        $aReportSections | report sections, regulate the type of output (optional) | ref array string
# input        $aStore | stores, regulate the type of output (optional) | ref array string
# input        $FileReportHandler | handler | ref.hash
# input        $DBStructureReportHandler | handler | ref.hash
# input        $DBContentReportHandler | handler | ref.hash
#========================================================================================
sub _printExport { ## no critic (ProhibitExcessComplexity)

    my $self = shift;
    my ( $OutputFolder, $Role, $aReportSections, $aStore, $FileReportHandler,
         $DBStructureReportHandler, $DBContentReportHandler ) = @_;

    foreach my $ReportSection (@$aReportSections) {

        my $StartTime = time;

        # --------------------------------------------------------------------
        # FileCategories
        # --------------------------------------------------------------------

        if ($ReportSection eq 'DiffCartridgesWebRoot') {
            my $hFiles = $FileReportHandler->getReport($ReportSection);
            foreach my $Store ( keys %{ $hFiles->{$Role} } ) {
                #write DiffCartridgesWebRoot.txt
                my $FilePath = "$OutputFolder/$Role/$Store/$ReportSection.txt";
                my $Content = join '', @{$hFiles->{$Role}->{$Store}};
                WriteFile($FilePath, \$Content, {create_directory=>1} );
            }
        }
        elsif ($ReportSection ~~ FileCategories) {
            my $hFiles = $FileReportHandler->getReport($ReportSection,$Role);
            my $FilePath = "$OutputFolder/$Role/Files/$ReportSection.txt";

            #clean up QA report hash
            if ($Role eq 'QA' and $ReportSection eq 'Cartridges'){
                $self->_normalizeReportCartridgeFilesHashQA($hFiles->{$Role});
            }
            if ($Role eq 'QA' and $ReportSection eq 'Config'){
                $self->_normalizeReportConfigHashQA($hFiles->{$Role});
            }
            if ($Role eq 'QA' and $ReportSection eq 'j'){
                $self->_normalizeReportEpagesjHashQA($hFiles->{$Role});
            }
            if ($Role eq 'QA' and $ReportSection eq 'Stores'){
                $self->_normalizeReportStoresFilesHashQA($hFiles->{$Role});
            }
            if ($Role eq 'QA' and $ReportSection eq 'var-epages'){
                $self->_normalizeReportVarEpagesHashQA($hFiles->{$Role});
            }
            if ($Role eq 'QA' and $ReportSection eq 'WebRoot'){
                $self->_normalizeReportWebRootHashQA($hFiles->{$Role});
            }
            if ($Role eq 'QA' and $ReportSection eq 'CTMPLFolder'){
                $self->_normalizeReportCTMPLHashQA($hFiles->{$Role});
            }

            #print hash to file
            $self->_saveFileData( $FilePath, $hFiles->{$Role} ) if defined $hFiles->{$Role};
        }

        if ( $ReportSection eq 'Cartridges' ) {
            #handle Cartridges.xml
            my $Source = $ENV{'EPAGES_CARTRIDGES'} . '/Cartridges.xml';
            my $Destination = "$OutputFolder/$Role/Files/Cartridges.xml";
            CopyFile( $Source, $Destination, { 'create_directory' => 1 } );
            $self->_sortCartridgesXML($Destination) if $Role eq 'QA';
        }
        if ( $ReportSection eq 'Monitor' ) {
            CopyDirectory( "$OutputFolder/Temp/Files/$ReportSection",
                           "$OutputFolder/$Role/Files/$ReportSection" );
            #delete .6.13.0-patch from history patch
            my $File = "$OutputFolder/$Role/Files/$ReportSection/.6.13.0-patch";
            DeleteFile($File) if -e $File and $Role eq 'QA';
        }
        if ( $ReportSection eq 'Config' ) {
            CopyDirectory( "$OutputFolder/Temp/Files/$ReportSection",
                           "$OutputFolder/$Role/Files/$ReportSection" );
            #normalize config files for QA report
            $self->_normalizeReportConfigFilesQA("$OutputFolder/$Role/Files/$ReportSection") if $Role eq 'QA';
        }
        if ( $ReportSection eq 'j' ) {
            CopyDirectory( "$OutputFolder/Temp/Files/$ReportSection",
                           "$OutputFolder/$Role/Files/$ReportSection" );
            # delete epagesj.conf.rpmsave from history patch
            my $File = "$OutputFolder/$Role/Files/$ReportSection/conf/epagesj.conf.rpmsave";
            DeleteFile($File) if -e $File and $Role eq 'QA';
            # canonicalize epagesj.conf and ep6WebService.conf in QA directory
            if ($Role eq 'QA') {
                _canonicalizeIni("$OutputFolder/QA/Files/$ReportSection/conf/epagesj.conf",
                                 [qw(host hosts masterUrl slaveUrls password)]);
                my $FileName = "$OutputFolder/QA/Files/$ReportSection/conf/ep6WebService.conf";
                my $rContent = GetFileContent($FileName);
                $$rContent =~ s,//.+?/,//ANONYMOUS/,g;
                WriteFile($FileName, $rContent);
            }
        }

        # --------------------------------------------------------------------
        # DBStructureCategories (SQLite)
        # --------------------------------------------------------------------

        if ( $ReportSection ~~ [qw(EbayCategoriesDB)] ) {
            #set export filepath for output file
            my $FilePath = "$OutputFolder/$Role/$ReportSection/DBSchema/schema.txt";

            #get hash containing dbschema data
            my $hSQLLite = $DBStructureReportHandler->getReport($ReportSection);

            #reduce hash to array
            my $aSQLLite = $hSQLLite->{$Role};

            #print array to file
            $self->_saveSQLLite( $aSQLLite, $FilePath );
        }

        # --------------------------------------------------------------------
        # DBStructureCategories (MySQL)
        # --------------------------------------------------------------------

        elsif ( $ReportSection ~~ DBStructureCategories ) {
            my $hDBSchemaFiles = $DBStructureReportHandler->getReport('DBSchema');
            #normalize QA Report
            $self->_normalizeReportDBStructureHashQA($hDBSchemaFiles) if $Role eq 'QA';

            #reduce hash to current role
            $hDBSchemaFiles = $hDBSchemaFiles->{$Role};

            #hash contains stores as first keys
            foreach my $Store ( keys %$hDBSchemaFiles ) {

                #hash contains name of different outputs as second keys
                $self->_saveDBSchemaData("$OutputFolder/$Role/$Store/DBSchema/$_.txt",
                                         $hDBSchemaFiles->{$Store}->{$_})
                    foreach keys %{ $hDBSchemaFiles->{$Store} };
            }
        }

        # --------------------------------------------------------------------
        # DBContentCategories
        # --------------------------------------------------------------------

        if ( $ReportSection eq 'ShopChildDependencies' ) {
            foreach my $Store (@$aStore) {
                #all stores without site
                next if $Store eq 'Site';

                #set export filepath for output file
                my $FilePath = "$OutputFolder/$Role/$Store/ShopChildDependencies.txt";

                #get hash containing ShopChildOrderList and DependencyList
                my $hShopChildDependencies = $DBContentReportHandler->getReport($ReportSection);

                #reduce hash to current role
                $hShopChildDependencies = $hShopChildDependencies->{$Role};

                #print hash to file in dependence on role
                $self->_saveShopChildDependencies( $FilePath, $hShopChildDependencies->{$Store} );
            }
        }

        if ( $ReportSection eq 'BCP' ) {
           my $hBCP = $DBContentReportHandler->getReport($ReportSection);

           #reduce hash to current role
           $hBCP = $hBCP->{$Role};

           #hash contains stores as first keys
           foreach my $Store ( keys %$hBCP ) {
               #hash contains name of different outputs as second keys
               $self->_saveArray("$OutputFolder/$Role/$Store/BCP/$_.txt", $hBCP->{$Store}->{$_})
                   foreach keys %{ $hBCP->{$Store} };
           }
        }

        if ( $ReportSection eq 'MySQLConfig' ) {
            my $hMySQL = $DBContentReportHandler->getReport($ReportSection);

            #reduce hash to current role
            $hMySQL = $hMySQL->{$Role};

            #set export filepath for output file
            my $FilePath = "$OutputFolder/$Role/mysql_config.txt";
            my $Content = "###SYSTEM_REPORT: QA_UNCOMPAREABLE_CONTENT###\n";

            if ($Role ne 'QA') {
                #print column names
                $Content = $hMySQL->{'Column_Names'} . "\n";
                #save data in mysql_config.txt
                foreach my $Key ( sort keys %$hMySQL ) {
                    next if ($Key eq 'Column_Names');
                    $Content .= "$Key,$hMySQL->{$Key}\n";
                }
            }
            WriteFile($FilePath, \$Content, {create_directory=>1} );
        }

        if ( $ReportSection eq 'Locales' ) {
            my $hLocalesInfo = $DBContentReportHandler->getReport($ReportSection);

            #reduce hash to current role
            my $hRoleLocalesInfo = $hLocalesInfo->{$Role};

            #set export filepath for output file
            my $FilePath = "$OutputFolder/$Role/$ReportSection.txt";
            my $Content = "Locales:\n";

            #save data in Locales.txt
            foreach my $Store ( sort keys %$hRoleLocalesInfo ){
                $Content .= "    >$Store\n";
                foreach my $LocaleID ( sort keys %{$hRoleLocalesInfo->{$Store}} ) {
                    $Content .= "        +$LocaleID\n";
                    foreach my $Key ( sort keys %{$hRoleLocalesInfo->{$Store}->{$LocaleID}} ){
                        my $Value = $hRoleLocalesInfo->{$Store}->{$LocaleID}->{$Key} // '';
                        $Value = '###SYSTEM_REPORT: QA_UNCOMPAREABLE_KEY###'
                            if $Key eq 'LanguageID' and $Role eq 'QA';
                        $Content .= "            -$Key='$Value'\n";
                    }
                }
            }
            WriteFile($FilePath, \$Content, {create_directory=>1} );
        }

        #-------------------------------------------------------------------
        # DBXMLCategories (create $OutputFolder/$Role/$Store/XML)
        #-------------------------------------------------------------------
        if ( $ReportSection ~~ [qw(DBXMLGeneralStoreSections DBXMLExclusiveStoreSections
                                   DBXMLExclusiveSiteSections)] ) {
            foreach my $Store (@$aStore) {
                next if $Store eq 'Site' and $ReportSection =~ /ExclusiveStore/;
                next if $Store ne 'Site' and $ReportSection =~ /ExclusiveSite/;
                my $SourceDir = "$OutputFolder/Temp/$Store/XML";
                my $TargetDir = "$OutputFolder/$Role/$Store/XML";
                CreateDirectory($TargetDir);
                # do not overwrite possibly existing files
                SystemCommand("cd \"$SourceDir\" && /bin/cp -rn . \"$TargetDir\"");
            }
        }
        elsif ( $ReportSection ~~ DBXMLCategories ) {
            my $FilePart = $ReportSection;
            $FilePart =~ s/^XML-//;

            foreach my $Store (@$aStore) {
                next if $Store eq 'Site' and $ReportSection ~~ DBXMLExclusiveStoreSections;
                next if $Store ne 'Site' and $ReportSection ~~ DBXMLExclusiveSiteSections;
                # do not overwrite possibly existing file
                next if -f "$OutputFolder/$Role/$Store/XML/$FilePart.xml";
                my $HasError;
                if (-f "$OutputFolder/Temp/$Store/XML/$FilePart.xml.error") {
                    CopyFile("$OutputFolder/Temp/$Store/XML/$FilePart.xml.error",
                             "$OutputFolder/$Role/$Store/XML/$FilePart.xml.error",
                             {create_directory => 1});
                    $HasError = 1;
                }
                next if $HasError and not -f "$OutputFolder/Temp/$Store/XML/$FilePart.xml";
                CopyFile("$OutputFolder/Temp/$Store/XML/$FilePart.xml",
                         "$OutputFolder/$Role/$Store/XML/$FilePart.xml",
                         {create_directory => 1});
            }
        }

        #-------------------------------------------------------------------
        # DBXMLCategories (change $OutputFolder/QA/$Store/XML)
        #-------------------------------------------------------------------
        next unless $Role eq 'QA';

        if ( $ReportSection ~~ [qw(XML-System DBXMLGeneralStoreSections)] ) {
            foreach my $Store (@$aStore) {
                my $File = "$OutputFolder/QA/$Store/XML/System.xml";
                $self->_normalizeSystemXMLQA($File);
            }
        }
        if ( $ReportSection ~~ [qw(XML-Stores DBXMLExclusiveSiteSections)] ) {
            my $File = "$OutputFolder/QA/Site/XML/Stores.xml";
            $self->_normalizeStoresXMLQA($File) if -f $File;
        }
        if ( $ReportSection ~~ [qw(XML-CartridgeRefs DBXMLExclusiveSiteSections)] ) {
            my $File = "$OutputFolder/QA/Site/XML/CartridgeRefs.xml";
            $self->_normalizeCartridgeRefsXMLQA($File) if -f $File;
        }
        if ( $ReportSection ~~ [qw(XML-ShopTypeRefs DBXMLExclusiveStoreSections)] ) {
            foreach my $Store (@$aStore) {
                next if $Store eq 'Site';
                my $File = "$OutputFolder/QA/$Store/XML/ShopTypeRefs.xml";
                $self->_normalizeShopTypeRefsXMLQA($File) if -f $File;
            }
        }
        if ( $ReportSection eq 'XML-ClientApplications' ) {
            foreach my $Store (@$aStore) {
                my $File = "$OutputFolder/QA/$Store/XML/ClientApplications.xml";
                $self->_normalizeClientApplicationsXMLQA($File) if -f $File;
            }
        }

        #-------------------------------------------------------------------
        # update runtime
        #-------------------------------------------------------------------
        $self->_updateRunTime($StartTime, $ReportSection);
    }
}

#========================================================================================
# function     _saveSQLLite
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_saveSQLLite( $aSQLLite );
#----------------------------------------------------------------------------------------
# description  saves sql lite db schema data
#----------------------------------------------------------------------------------------
# input        $aSQLLite | two dimensional array |  ref array
# input        $FilePath | target file for output | string
#========================================================================================
sub _saveSQLLite {
  my $self = shift;
  my ( $aSQLLite, $FilePath ) = @_;

  my $Filehandle;

  #create directory, if Filepath doesn't exist
  unless ( -e "$FilePath" ) {

    #get path to the file
    my ($Directories) = ( fileparse( $FilePath, '\..+$' ) )[1];
    CreateDirectory($Directories);
  }
  open( $Filehandle, '>>', $FilePath ) or Error("Could not open FilePath for writing! $!");    ## no critic (InputOutput::RequireBriefOpen)

  foreach my $row (@$aSQLLite) {
    print $Filehandle "$row->[0]|" . "$row->[1]|" . "$row->[2]|";
    print $Filehandle $self->_prettysql( $row->[3] );
    print $Filehandle "\n";

  }
  close($Filehandle) or Error("Could not close FilePath! $!");

}

#========================================================================================
# function     _prettysql
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_prettysql( $Sql );
#----------------------------------------------------------------------------------------
# description  preperate SQL to fit in one line
#----------------------------------------------------------------------------------------
# input        $Sql | unprepared SQL statement | string
# return       $Sql | SQL statement without new lines, unecessary whitespace,
#                      spaces and in lowercase | string
#========================================================================================
sub _prettysql {

  #/Patch/PatchDiff/exportsqllitedbschema.pl
  my $self = shift;
  my ($Sql) = @_;

  #initialize variable
  $Sql='' unless defined $Sql;

  #Remove new lines in SQL statements
  $Sql =~ s/\n//g;

  #Remove unnecessary whitespace (two or more will be merged)
  $Sql =~ s/\s{2,}/ /g;

  #Remove spaces around ',', '(' and ')'
  $Sql =~ s/\s*(\(|\)|,)\s*/$1/g;

  #convert to lowercase
  return lc $Sql;

}

#========================================================================================
# function     _intersection
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_intersection( $aArray1, $aArray2 );
#----------------------------------------------------------------------------------------
# description  returns the intersection of both passed arrays
#----------------------------------------------------------------------------------------
# input        $aArray1 | first array | ref array
# input        $aArray2 | second array | ref array
# return       $aIntersection | intersection between first and second array | ref array
#========================================================================================
sub _intersection {
  my $self = shift;
  my ( $aArray1, $aArray2 ) = @_;
  my $aIntersection = ();
  my %count         = ();

  #count quantity of each element in both arrays jointly
  foreach my $element ( @$aArray1, @$aArray2 ) {
    $count{$element}++;
  }

  #for each element in the joined hash
  foreach my $element ( keys %count ) {

    #if the quantity of current element is > 1, then it is in both arrays
    if ( $count{$element} > 1 ) { push @$aIntersection, $element; }
  }

  #return result array
  return $aIntersection;
}

#========================================================================================
# function     _normalizeSystemXMLQA
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_normalizeSystemXMLQA( $FileName );
#----------------------------------------------------------------------------------------
# description  removes lines from System.xml, that are not important for QA
#----------------------------------------------------------------------------------------
# input        $FileName | file path | string
#========================================================================================
sub _normalizeSystemXMLQA {
    my $self = shift;
    my ($SysFile) = @_;

    #save file content in a string
    my $rContent = GetFileContent($SysFile);

    foreach my $KeyToDelete (qw(DomainName EbayLastUpdate ShopzillaInstanceID
                                EPagesJBaseURL PatchInstalledOn PatchVersion
                                UnityBaseURL UnityApiUrlScheme UnityUrlScheme)){
        $$rContent =~ s,.*\b$KeyToDelete=.*\n,,;
    }

    #save changes
    WriteFile( $SysFile, $rContent );
}

#========================================================================================
# function     _normalizeStoresXMLQA
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_normalizeStoresXMLQA( $FileName );
#----------------------------------------------------------------------------------------
# description  removes stores with different aliases than Store or Site from the passed
#               xml file (e.g. Stores.xml)
#----------------------------------------------------------------------------------------
# input        $FileName | file path | string
#========================================================================================
sub _normalizeStoresXMLQA {
    my $self = shift;
    my ($FileName) = @_;

    #import xml file
    my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
    my $ExportDriver = DE_EPAGES::Object::API::XMLStructureExportDriver->new();

    #transform xml into object
    my $XMLStruct = $ImportHandler->getXMLStructure($FileName);

    foreach my $Element (@{$XMLStruct->getElements()}){
        if ($Element =~ m,/Stores/([^/]+)$, and $1 ne 'Site') {
            #anonymize:
            #<Store WebServiceProxy='http://cd-vm-test-01/epages/Store.soap'
            # ConfigurationServerPassword='e7C0IHdL' WebServicePassword='e7C0IHdL'
            # Alias='cd-vm-test-01' Name='cd-vm-test-01'>
            foreach my $Attribute (qw(ConfigurationServerPassword Name Store
                                      WebServicePassword WebServiceProxy)) {
                next unless $XMLStruct->existsTagAttribute($Element,$Attribute);
                $XMLStruct->setTagAttribute($Element, $Attribute, 'ANONYMOUS');
            }
            $XMLStruct->setIDValue($Element, 'Alias', 'ANONYMOUS')
                if $XMLStruct->existsIDPart($Element, 'Alias');
        } elsif ($Element =~ m,/Stores/(.+?)/StoreFeature,
                 and $1 ne 'Site' and $self->_isFlexSiteInstallation()) {
            #delete <StoreFeature /> for flex site
            $XMLStruct->deleteElement($Element);
        }
    }

    #export the patch delete file
    $ExportDriver->exportXML($FileName, $XMLStruct);
}

#========================================================================================
# function     _normalizeClientApplicationsXMLQA
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_normalizeClientApplicationsXMLQA( $FileName );
#----------------------------------------------------------------------------------------
# description  Removes infomation from the given XML file which is not relevant for
#               the patch diff evaluation, e.g. the passwords for Shippo and OneSaas.
#----------------------------------------------------------------------------------------
# input        $FileName | path to the ClientApplications.xml file | string
#========================================================================================
sub _normalizeClientApplicationsXMLQA {
    my $self = shift;
    my ($FileName) = @_;

    #import xml file
    my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
    my $ExportDriver = DE_EPAGES::Object::API::XMLStructureExportDriver->new();

    #transform xml into object
    my $XMLStruct = $ImportHandler->getXMLStructure($FileName);

    foreach my $Element (@{$XMLStruct->getElements()}) {
        next unless $XMLStruct->existsTagAttribute($Element, 'Password');
        $XMLStruct->setTagAttribute($Element, 'Password', 'ANONYMOUS');
    }

    #export the patch delete file
    $ExportDriver->exportXML($FileName, $XMLStruct);
}

#========================================================================================
# function     _normalizeShopTypeRefsXMLQA
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_normalizeShopTypeRefsXMLQA( $FileName );
#----------------------------------------------------------------------------------------
# description  removes old shop types for flex from the passed
#               xml file (e.g. Stores.xml)
#----------------------------------------------------------------------------------------
# input        $FileName | file path | string
#========================================================================================
sub _normalizeShopTypeRefsXMLQA {
    my $self = shift;
    my ($FileName) = @_;

    #import xml file
    my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
    my $ExportDriver = DE_EPAGES::Object::API::XMLStructureExportDriver->new();

    #transform xml into object
    my $XMLStruct = $ImportHandler->getXMLStructure($FileName);

    #for all elements
    my $FlexShopType = $self->_getFlexStoreShopType();
    foreach my $Element (@{$XMLStruct->getElements()}) {
        #delete <ShopTypeRef Alias='Flex_VERSION' /> for all versions other than current
        $XMLStruct->deleteElement($Element)
            if $Element =~ m|^/ShopTypeRefs/Flex_.*| and $Element ne "/ShopTypeRefs/$FlexShopType";
    }

    #export the patch delete file
    $ExportDriver->exportXML($FileName, $XMLStruct);
}

#========================================================================================
# function     _normalizeCartridgeRefsXMLQA
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_normalizeCartridgeRefsXMLQA( $FileName );
#----------------------------------------------------------------------------------------
# description  replaces value of attribute Stores by uncompareable content
#----------------------------------------------------------------------------------------
# input        $FileName | file path | string
#========================================================================================
sub _normalizeCartridgeRefsXMLQA {
    my $self = shift;
    my ($CartridgeRefsFile) = @_;

    die "\nmissing file name" unless defined $CartridgeRefsFile;

    #read in file content
    open (my $input_handle, '<', $CartridgeRefsFile) or die "Can't open file '$CartridgeRefsFile'";
    my @Content = <$input_handle>;
    close $input_handle or die "unable to close: $!";

    #reopen file to write back content
    open (my $output_handle, '>', $CartridgeRefsFile) or die "Can't open file '$CartridgeRefsFile'";

    foreach my $XMLline (@Content) {
        #normalize attribute
        $XMLline =~ s/(<StoreCartridge\b.+\bStore\s*=\s*\').+?(\')/$1ANONYMOUS$2/;
        print $output_handle $XMLline;
    }
    close $output_handle or die "unable to close: $!";
}

#global variable for _exportMainXML and _transForm
my %PARAMS;

#roadworks
sub _transForm {
  my ( $TagName, $hAttributes ) = @_;
  delete $hAttributes->{'GUID'}         if $PARAMS{'withoutguids'};
  delete $hAttributes->{'CreationDate'} if $PARAMS{'withoutcreationdate'};
  delete $hAttributes->{'FilePathPart'} if $PARAMS{'withoutfilepathpart'};
  return;
}

#========================================================================================
# function     _exportMainXML
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_exportMainXML( $FileName, $StoreName, $aArgs, $Path, $WriterPackage, $Recursive );
#----------------------------------------------------------------------------------------
# description  exports a xml file
#----------------------------------------------------------------------------------------
# input        $FileName | target file for output | string
# input        $StoreName | used Store | string
# input        $aArgs | | ref array
# input        $Path | | string
# input        $WriterPackage | | string
# input        $Recursive | | boolean
#========================================================================================
sub _exportMainXML {
    my $self = shift;
    my ( $FileName, $StoreName, $aArgs, $Path, $WriterPackage, $Recursive ) = @_;

    my $Password;
    my $Encoding = 'utf-8';
    $Recursive = 1 unless defined $Recursive;
    my $WithDefaults;
    my $WithoutGuids        = 1;
    my $WithoutCreationDate = 1;
    my $WithoutFilePathPart;

    %PARAMS = (
        'encoding'            => $Encoding,
        'recursive'           => $Recursive,
        'withdefaults'        => $WithDefaults,
        'withoutguids'        => $WithoutGuids,
        'withoutcreationdate' => $WithoutCreationDate,
        'withoutfilepathpart' => $WithoutFilePathPart,
        'writer'              => $WriterPackage,
    );

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub {
            my $subref = \&_transForm;
            my (@aObjects, @aNoObjects);

            foreach my $Object (@$aArgs) {
                my $ObjectPath = $Path;
                $ObjectPath =~ s,/$,,;
                $ObjectPath .= "/$Object";
                if (ExistsObjectByPath($ObjectPath)) {
                    push @aObjects, $Object;
                } else {
                    push @aNoObjects, $ObjectPath;
                }
            }
            if (scalar @aObjects) {
                ExportObjects($Path, $FileName, \%PARAMS, \@aObjects, $subref);
            }
            if (scalar @aNoObjects) {
                my $Content = 'Object not found: '.
                    join("\nObject not found: ", @aNoObjects). "\n";
                WriteFile("$FileName.error", \$Content, {create_directory=>1} );
            }
        },
    ) if ($self->_storeExists($StoreName));
}

#========================================================================================
# function     _removeOutputDir
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_removeOutputDir( $OutputFolder, $aRoles );
#----------------------------------------------------------------------------------------
# description  asks the user, wether the output directory should be deleted
#               yes will delete files, no will exit the script
#----------------------------------------------------------------------------------------
# input        $OutputFolder | target output directory for output | string
# input        $aRoles | all possible roles | ref array
# return       $Result | true or false | boolean
#========================================================================================
sub _removeOutputDir {
  my $self = shift;
  my ( $OutputFolder, $aRoles ) = @_;

  #test if output folder already exists
  if ( -e "$OutputFolder" ) {


    if($self->{'overwrite'}){

        DeleteDirectory( $OutputFolder, { 'recursive' => 1 } );
        return 1;
    }
    else{

        my $Request;
        do {

            #start user request, wether directories should be deleted
            print "output folder '$OutputFolder' already exists, all files will be deleted. Do you want to continue? <Y/N>\n";

            #save input
            $Request = <>;

            #delete line break
            chop($Request);

            #case 1: user chose no - exit script
            if ( uc($Request) eq 'N' ) {
                print "Canceled by User\n";
                exit(1);
            }

            #case 2: user chose yes - delete directories and leave subroutine
            elsif ( uc($Request) eq 'Y' ) {

                #delete all directories and subdirectories
                DeleteDirectory( $OutputFolder, { 'recursive' => 1 } );
            }

            #case 3: user chose neither no nor yes - loop will start again
            else {
                print "$Request is not a valid input\n";
            }
        } while ( uc($Request) ne 'Y' );
    }

  }
}

#========================================================================================
# function     _saveArray
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_saveArray( $FilePath, $aArray );
#----------------------------------------------------------------------------------------
# description  stores a two-dimensional array in a textfile
#----------------------------------------------------------------------------------------
# input        $FilePath | target file for output | string
# input        $aArray | two-dimensional array | ref array
#========================================================================================
sub _saveArray {
  my $self = shift;
  my ( $FilePath, $aArray ) = @_;

  #create directory, if Filepath doesn't exist
  unless ( -e "$FilePath" ) {

    #get path to the file
    my ($Directories) = ( fileparse( $FilePath, '\..+$' ) )[1];
    CreateDirectory($Directories);
  }
  open( my $Filehandle, '>', "$FilePath" ) or Error("Could not open $FilePath for writing! $!");

  #iterate over outer array
  for ( my $i = 0; $i < @$aArray; $i++ ) {

    #iterate over inner array
    for ( my $j = 0; $j < @{ $aArray->[$i] }; $j++ ) {

      #print inner values to file
      print $Filehandle "$aArray->[$i]->[$j]";

      #seperate values by comma unless it's the last value in line
      print $Filehandle ',' unless ( $j == @{ $aArray->[$i] } - 1 );
    }

    #end of line
    print $Filehandle "\n";
  }
  close($Filehandle) or Error("Could not close FilePath! $!");
}

#========================================================================================
# function     _saveShopChildDependencies
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_saveShopChildDependencies( $FilePath, $hShopChildDependencies, $Role ) ;
#----------------------------------------------------------------------------------------
# description  stores different shopchilddependencies in a textfile in dependence on $Role
#----------------------------------------------------------------------------------------
# input        $FilePath | target file for output | string
# input        $hShopChildDependencies | hash pointing to ShopChildOrderList and
#                    DependencyList | ref hash
#========================================================================================
sub _saveShopChildDependencies {
  my $self = shift;
  my ( $FilePath, $hShopChildDependencies ) = @_;

  #get two different hashes out of $hShopChildDependencies
  my $hOrderList      = $hShopChildDependencies->{'ShopChildOrderList'};
  my $hDependencyList = $hShopChildDependencies->{'DependencyList'};

  #create directory, if Filepath doesn't exist
  unless ( -e "$FilePath" ) {

    #get path to the file
    my ($Directories) = ( fileparse( $FilePath, '\..+$' ) )[1];
    CreateDirectory($Directories);
  }

  if ( defined $hOrderList ) {
    $self->_saveShopChildOrderList( $hOrderList, $FilePath );
  }
  $self->_saveDependencyList( $hDependencyList, $FilePath );

}

#========================================================================================
# function     _saveShopChildOrderList
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_saveShopChildOrderList( $hOrderList, $FilePath );
#----------------------------------------------------------------------------------------
# description  stores shopchildorderlist in a textfile
#----------------------------------------------------------------------------------------
# input        $hOrderList | hash containing the position as key and the value as inner key | ref hash
# input        $FilePath | target file for output | string
#========================================================================================
sub _saveShopChildOrderList {
  my $self = shift;
  my ( $hOrderList, $FilePath ) = @_;
  my $Filehandle;

  #create file
  open( $Filehandle, '>>', "$FilePath" ) or Error("Could not open $FilePath for writing! $!");    ## no critic (InputOutput::RequireBriefOpen)
  print $Filehandle ">ShopChildOrder:\n";

  #print position ($Key) and shop child ($Inkey) sorted by position in a textfile
  foreach my $Key ( sort { $a <=> $b } keys %$hOrderList ) {
    foreach my $Inkey ( keys %{ $hOrderList->{$Key} } ) {
      print $Filehandle "   $Key. $Inkey\n";
    }
  }
  close($Filehandle) or Error("Could not close FilePath! $!");

  return;

}

#========================================================================================
# function     _saveDependencyList
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_saveDependencyList( $hDependencyList, $FilePath );
#----------------------------------------------------------------------------------------
# description  stores dependencylist in a textfile
#----------------------------------------------------------------------------------------
# input        $FilePath | target file for output | string
# input        $hDependencyList | hash containing dependencylist | ref hash
#========================================================================================
sub _saveDependencyList {
  my $self = shift;
  my ( $hDependencyList, $FilePath ) = @_;
  my $Filehandle;

  open( $Filehandle, '>>', "$FilePath" ) or Error("Could not open $FilePath for writing! $!");    ## no critic (InputOutput::RequireBriefOpen)

  #for each shopchild in List print predecessors and successors
  foreach my $Key ( sort keys %$hDependencyList ) {
    print $Filehandle "\n>" . $Key . ":\n";

    #print all predecessors
    print $Filehandle '  - Predecessors: ';
    foreach my $Successor ( sort keys %{ $hDependencyList->{$Key}->{'Predecessors'} } ) {
      print $Filehandle $Successor . ' ';
    }

    #print all successors
    print $Filehandle "\n  - Successors: ";
    foreach my $Successor ( sort keys %{ $hDependencyList->{$Key}->{'Successors'} } ) {
      print $Filehandle $Successor . ' ';
    }
    print $Filehandle "\n";
  }
  close($Filehandle) or Error("Could not close FilePath! $!");

  return;
}

#========================================================================================
# function     _saveDBSchemaData
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_saveDBSchemaData( $FilePath, $hHash );
#----------------------------------------------------------------------------------------
# description  stores dbschemadata in a textfile
#----------------------------------------------------------------------------------------
# input        $FilePath | target file for output | string
# input        $aaData | two-dimensional array containing sql table | ref array
#========================================================================================
sub _saveDBSchemaData {
  my $self = shift;
  my ( $FilePath, $aaData ) = @_;
  my $Filehandle;

  #create directory, if Filepath doesn't exist
  unless ( -e "$FilePath" ) {

    #get path to the file
    my ($Directories) = ( fileparse( $FilePath, '\..+$' ) )[1];
    CreateDirectory($Directories);
  }

  #first array of $aaData contains column names
  my $aColumns = $aaData->[0];

  #get length of column names
  my @columnWidth = map { length $_ } @$aColumns;

  #get quantity of columns
  my $columnCount = scalar @$aColumns;

  #get maximum column width for every column
  for ( my $i = 0; $i < @$aaData; $i++ ) {
    for ( my $j = 0; $j < $columnCount; $j++ ) {

      #set current value to (null) if undefined
      $aaData->[$i]->[$j] = '(null)' unless defined $aaData->[$i]->[$j];

      #get current value
      my $value = $aaData->[$i]->[$j];

      #test for maximum
      if ( length $value > $columnWidth[$j] ) {
        $columnWidth[$j] = length $value;
      }
    }
  }
  open( $Filehandle, '>>', $FilePath ) or Error("Could not open FilePath for writing! $!");    ## no critic (InputOutput::RequireBriefOpen)

  #print a neatly formatted table
  #print headline
  for ( my $i = 0; $i < $columnCount; $i++ ) {

    #print every value filled with spaces
    printf $Filehandle '%-' . $columnWidth[$i] . 's', $aColumns->[$i];

    #if current value is last in row, print line break
    print $Filehandle ( ( $i + 1 == $columnCount ) ? "\n" : ' ' );
  }

  #print underline
  for ( my $i = 0; $i < $columnCount; $i++ ) {

    #print so many hyphens to match the maximum collumn width
    print $Filehandle '-' x $columnWidth[$i];

    #if current value is last in row, print line break
    print $Filehandle ( ( $i + 1 == $columnCount ) ? "\n" : ' ' );
  }

  #print values
  #print row
  for ( my $i = 1; $i < @$aaData; $i++ ) {

    #print value
    for ( my $j = 0; $j < $columnCount; $j++ ) {

      #print every value filled with spaces
      printf $Filehandle '%-' . $columnWidth[$j] . 's', $aaData->[$i]->[$j];

      #if current value is last in row, print line break
      print $Filehandle ( ( $j + 1 == $columnCount ) ? "\n" : ' ' );
    }
  }
  close($Filehandle) or Error("Could not close FilePath! $!");

}

#========================================================================================
# function     _saveFileData
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_saveFileData( $FilePath, $hFileData );
#----------------------------------------------------------------------------------------
# description  stores fileinformation (fingerprint and filepath) from a hash in a textfile
#----------------------------------------------------------------------------------------
# input        $FilePath | target file for output | string
# input        $hFileData | fileinformation, containing filepath as key and fingerprint as value | string
#========================================================================================
sub _saveFileData {
  my $self = shift;
  my ( $FilePath, $hFileData ) = @_;
  my $Filehandle;

  #create directory, if Filepath doesn't exist
  unless ( -e "$FilePath" ) {

    #get path to the file
    my ($Directories) = ( fileparse( $FilePath, '\..+$' ) )[1];
    CreateDirectory($Directories);
  }

  #open file in append mode
  open( $Filehandle, '>>', $FilePath ) or die $!;    ## no critic (InputOutput::RequireBriefOpen)

  #print hash, sorted numerically by position
  foreach my $Position ( sort { $a <=> $b } keys %$hFileData ) {
    foreach my $Filename ( keys %{ $hFileData->{$Position} } ) {
      print $Filehandle $self->_cleanFileData( $hFileData->{$Position}->{$Filename}, $Filename );
    }
  }
  close($Filehandle) or Error("Could not close FilePath! $!");

}

#========================================================================================
# function     _cleanFileData
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SystemInfoReportHandler->_cleanFileData( $Part1, $Part2 );
#----------------------------------------------------------------------------------------
# description  translates backslashes in slashes in $Part2, returns "$Part1 $Part2\n" if
#               $Part2 does not match exclude patterns like *.ctmpl
#----------------------------------------------------------------------------------------
# input        $Part1 | first part of line containing fingerprint | string
# input        $Part2 | second part of the line to be cleaned containing filepath | string
# return       "$Part1 $Part2\n" | $Part1: fingerprint, $Part2: cleaned filepath | string
#========================================================================================
sub _cleanFileData {

  #/Checkout/Patch/PatchDiff/clean.pl
  my $self = shift;
  my ( $Part1, $Part2 ) = @_;
  my @ExcludePatterns = qw (/\.6-* /\.6\.* ^\.6-* ^\.6\.* ~$ /\.packlist$
  /perllocal\.pod$ \.ept$ \.initial$ \.latest$
  ^ASPool\.db$ ^EbayCategories\.db$ ^version\.txt$
  ^CPAN$ ^CPAN/ ^DE_EPAGES/makefile$ ^Doc/Help$
  ^Doc/Help/ ^Doc/Manual/[^/]+/[^/]+\.pdf$ ^\.\/Cartridges\.xml$);

  #remove backslashes
  $Part2 =~ s,\\,/,g;

  #return empty string if line matches a pattern
  foreach my $Pattern (@ExcludePatterns) {
    if ( $Part2 =~ "$Pattern" ) {
      return '';
    }
  }

  #return string with slashes
  return "$Part1 $Part2\n";

}

sub _normalizeServerConfigQA {
    my $self = shift;
    my ($ServerConfigJson) = @_;

    # create the hash $ServerConfig
    my $ServerConfig;
    my $Dump = Data::Dumper->new([ServerConfig()->{'XMLConfig'}], ['ServerConfig']);

    # translate $ServerConfig into string $ServerConfigLines
    my $ServerConfigLines = $Dump->Dump;

    # replace IPs/Ports/MaxMemory/DefaultRequestRouter by 'ANONYMOUS'
    $ServerConfigLines =~ s/'(\d{1,3}\.){3}\d{1,3}(_\d+(-\d+)?)?'/'ANONYMOUS'/g;
    $ServerConfigLines =~ s/'(Ports)' => '[\d-]+'/'$1' => 'ANONYMOUS'/g;
    $ServerConfigLines =~ s/'(Bind)' => \d,/'$1' => 'ANONYMOUS',/g;
    $ServerConfigLines =~ s/.*'MaxMemory'.*\n//g;
    $ServerConfigLines =~ s/DefaultRequestRouter/ANONYMOUS/g;

    # translate $ServerConfigLines into $ServerConfig
    ## no critic qw(BuiltinFunctions::ProhibitStringyEval)
    eval $ServerConfigLines;
    ## use critic

    # translate $ServerConfig into sorted json string
    my $Content = JSON::XS->new->canonical->pretty->encode($ServerConfig);
    WriteFile($ServerConfigJson, \$Content);
}

sub _normalizeReportConfigFilesQA{
    my $self = shift;

    my ($QAReportConfFolder) = @_;

    #check for folder
    die "wrong format for folder '$QAReportConfFolder'"
        unless ($QAReportConfFolder =~ /(\\|\/)QA(\\|\/)Files(\\|\/)Config/);

    #normalize and clean up config files
    my $hInfo = GetInfoDirectory( $QAReportConfFolder, { recursive => 1 } );

    foreach my $Directory (@{ $hInfo->{'Directories'} }){
        DeleteDirectory($Directory, { 'recursive' => 1 }) if $Directory =~ m,/hotfixes$,;
    }

    foreach my $File (@{ $hInfo->{'Files'}}){

        if(
            $File =~ /ASPool\.db$/ or
            $File =~ /ASPoolDB\.conf\.mysql$/ or
            $File =~ /EbayCategories\.db$/ or
            $File =~ /ServerConfig\.xml$/ or
            $File =~ /version\.txt$/ or
            $File =~ /(\.conf\.latest|~)$/ or
            $File =~ /\.(env|ept|initial|backup|sample)$/
        ){
            DeleteFile($File);
        }
        elsif( $File =~ m,/StoreTypes.d/.*\.conf, ){
            my $aContent = GetFileContentLines($File);
            open(my $FileHandle, '>', $File) or die "$File: $!\n";
            foreach my $Line ( sort @$aContent ) {
                next if $Line =~ /^(de|en)=installed/;
                print $FileHandle $Line;
            }
            close($FileHandle) or die "$File: $!\n";
        }
        elsif ( $File =~ /log4perl\.conf/ or $File =~ /PdfFontMapping\.conf/ ){
            my $aContent = GetFileContentLines($File);
            open(my $FileHandle, '>', $File) or die "$File: $!\n";
            foreach my $Line ( sort @$aContent ) {
                next if $Line =~ /^\s*($|#)/;
                print $FileHandle $Line;
            }
            close($FileHandle) or die "$File: $!\n";
        }
        elsif ($File =~ /\.conf$/){
            #delete flex-site exceptions
            if ($self->_isFlexSiteInstallation()) {
                foreach my $Store (@{_getFlexSiteStoreNames()}) {
                    next unless $File =~ m,Database.d/$Store(\..+)?.conf,;
                    next unless -f $File;
                    MoveFile($File, "$QAReportConfFolder/Database.d/STORE.conf");
                    $File = "$QAReportConfFolder/Database.d/STORE.conf";
                }
            }

            #normalize config file
            $self->_normalizeReportConfigFileQA($File);
        }
    }

    $self->_normalizeServerConfigQA("$QAReportConfFolder/ServerConfig.json");

    return;
}

sub _normalizeReportConfigFileQA {
    my $self = shift;
    my ($File) = @_;

    my (@Anonymize, @Ignore);

    if (basename(dirname($File)) eq 'Database.d') {
        @Anonymize = qw(datasource BackupHost);
        @Ignore = qw(host FinalBackupType InitialBackupType);
        if(basename($File) eq 'ASPoolDB.conf') {
            @Ignore = qw(Backup BackupHost);
        }
    } elsif (basename($File) eq 'epages.conf') {
        @Anonymize = qw(SystemDomainName DefaultWebServiceProxy SiteUrl
                        UnityApiUrlScheme UnityUrlScheme);

    } elsif (basename($File) eq 'WebInterface.conf') {
        @Anonymize = qw(DEDICATED ExceptionHost APPLICATIONSERVER MAXMEMORY);

    } elsif (basename($File) eq 'ConfigServer.conf') {
        @Anonymize = qw(Password);
    }

    #canonicalize $File in QA directory
    _canonicalizeIni($File, [@Anonymize], [@Ignore]);
}

sub _normalizeReportConfigHashQA{
    my $self = shift;
    my ($hConfigReportHash) = @_;

    #for every position and file name in hash
    foreach my $Position (keys %{$hConfigReportHash}){
        foreach my $FileName (keys %{$hConfigReportHash->{$Position}}){

            #either remove exceptions
            if ( $FileName =~ /ASPool\.db$/ or
                 $FileName =~ /ASPoolDB\.conf\.mysql$/ or
                 $FileName =~ /EbayCategories\.db$/ or
                 $FileName =~ /version\.txt$/ or
                 $FileName =~ /hotfixes\// or
                 $FileName =~ /(\.conf\.latest|~)$/ or
                 $FileName =~ /\.(env|ept|initial|backup|sample)$/ ) {
                delete $hConfigReportHash->{$Position};
                next;
            }
            #or replace md5, but not the linux permissions
            $hConfigReportHash->{$Position}{$FileName} =~
                s/[[:xdigit:]]{32}$/######QA_UNCOMPAREABLE_MD5######/;

            #delete flex-site exceptions
            next unless $self->_isFlexSiteInstallation();
            foreach my $Store (@{_getFlexSiteStoreNames()}) {
                # find Database.d/$Store.conf or Database.d/$Store.DOMAIN.conf
                next unless $FileName=~ m,Database.d/$Store(\..+)?.conf,;
                $hConfigReportHash->{$Position}{'Database.d/STORE.conf'} =
                    $hConfigReportHash->{$Position}{$FileName};
                delete $hConfigReportHash->{$Position}{$FileName};
            }
        }
    }

    return;
}

sub _normalizeReportVarEpagesHashQA{
    my $self = shift;
    my ($hVarEpagesReportHash) = @_;

    #for every position and file name in hash
    foreach my $Position (keys %{$hVarEpagesReportHash}){
        foreach my $FileName (keys %{$hVarEpagesReportHash->{$Position}}){

            #remove exceptions
            if (
                $FileName =~ /CheckHotfixVersion\.PL$/ or
                $FileName =~ /Hotfix\.pm$/ or
                $FileName =~ /UpdateHotfixRepo\.PL$/
            ){

                delete $hVarEpagesReportHash->{$Position};
            }
        }
    }

    return;
}

sub _normalizeReportDBStructureHashQA{
    my $self = shift;
    my ($hDBSchemaFiles) = @_;

    foreach my $Store (keys %{$hDBSchemaFiles->{'QA'}}){
        $hDBSchemaFiles->{'QA'}->{$Store}->{'GlobalVariables'} = [
            [
                '###SYSTEM_REPORT: QA_UNCOMPAREABLE_CONTENT###',
                ''
            ]
        ];
        if($Store eq 'ASPoolDB' or $Store eq 'Store') {
            $hDBSchemaFiles->{'QA'}->{$Store}->{'Views'} = [
                [
                    '###SYSTEM_REPORT: QA_UNCOMPAREABLE_CONTENT###',
                    ''
                ]
            ];
        }
    }

    return;
}

sub _normalizeReportCartridgeFilesHashQA{
    my $self = shift;
    my ($hReportHash) = @_;

    #check for CPAN folder / Cartridges.xml
    foreach my $FileNumber (keys %$hReportHash){
        foreach my $File ( keys %{$hReportHash->{$FileNumber}}){
            if ($File =~ m,^Cartridges\.xml$,) {delete $hReportHash->{$FileNumber};}
            elsif ($File =~ m,^/CPAN/,) { delete $hReportHash->{$FileNumber}; }
        }
    }
    return;
}

sub _normalizeReportStoresFilesHashQA{
    my $self = shift;
    my ($hReportHash) = @_;

    #Stores exceptions
    my @Exeptions = (
      qr;Site/ShopCount\.xml;,
      qr;Site/ShopFeatures\.xml;,
      qr;Site/shops_Distributor\.csv;,
      qr;^Master/MasterShop;,
      qr;/export\.zip$;,
      qr;(/[0-9A-F]{4}){8}/;,
      qr;^_MONITOR_/[^.]+\.(html|response)$;,
    );

    $self->_deleteExeptionsFromReportHash($hReportHash, \@Exeptions);
    return;
}

sub _normalizeReportWebRootHashQA {
    my $self = shift;
    my ($hReportHash) = @_;

    #WebRoot exceptions
    my @Exeptions = (
      qr;PatchReport(/|.*\.html$);,
      qr;^[^/]+/Shops/;,
      qr;^_MONITOR_/[^.]+\.(html|response)$;,
      qr;^RssFeeds/http[^.]+\.xml$;,
      qr;javascript(/build)?/dojo/build.txt$;,
    );

    $self->_deleteExeptionsFromReportHash($hReportHash, \@Exeptions);
    return;
}

sub _normalizeReportCTMPLHashQA {
    my $self = shift;
    my ($hReportHash) = @_;

    #WebRoot exceptions
    my @Exeptions = (
    );

    $self->_deleteExeptionsFromReportHash($hReportHash, \@Exeptions);
    return;
}

sub _normalizeReportEpagesjHashQA {
    my $self = shift;
    my ($hReportHash) = @_;

    #Epagesj exceptions
    my @Exeptions = (
        qr;^(cores|logs|work|rpmsave)/;,
        qr;(epagesj\.conf|\.rpmsave|ep6WebService\.conf)$;,
    );

    $self->_deleteExeptionsFromReportHash($hReportHash, \@Exeptions);
    return;
}

sub _deleteExeptionsFromReportHash {
    my $self = shift;
    my ($hReportHash, $aExeptions) = @_;

    foreach my $FileNumber ( keys %$hReportHash ) {
        foreach my $File ( keys %{$hReportHash->{$FileNumber}} ) {

            #if current file is not comparable/important, delete the entry
            if( any { $File =~ $_ } @$aExeptions ) {
                delete $hReportHash->{$FileNumber};
            }
        }
    }
    return;
}

#========================================================================================
# function     _canonicalizeIni
# state        private
#----------------------------------------------------------------------------------------
# syntax       $AnonymousContent = _canonicalizeIni($IniFileName, $aAnonymize, $aIgnore);
#----------------------------------------------------------------------------------------
# description  Normalizes and anonymizes content of $IniFileName,
#               see functions _normalizeIni and _anonymizeIni.
#----------------------------------------------------------------------------------------
# input        $IniFileName | name of an ini file | string
# input        $aAnonymize | array of keys to anonymize | ref.array
# input        $aIgnore | array of keys to be ignored | ref.array
#========================================================================================
sub _canonicalizeIni {
    my ($FileName, $aAnonymize, $aIgnore) = @_;

    #return unless $FileName exists
    die "no such file: $FileName\n" unless -f $FileName;

    #return if file is empty
    return unless -s $FileName;

    my $Content = _anonymizeIni(_normalizeIni($FileName, $aIgnore), $aAnonymize);
    WriteFile($FileName, \$Content);

    return;
}

#========================================================================================
# function     _normalizeIni
# state        private
#----------------------------------------------------------------------------------------
# syntax       $aLines = _normalizeIni($IniFileName, $aIgnore);
#----------------------------------------------------------------------------------------
# description  Returns normalized lines $aLines from $IniFileName. Normalized lines are
#               sorted first by section then by key. Quotes are dropped in values.
#               Surrounding spaces are dropped in keys and values. Comments and empty lines
#               are dropped. Ignore line if key belongs to @$aIgnore.
#----------------------------------------------------------------------------------------
# input        $IniFileName | name of an ini file | string
# input        $aIgnore | array of keys to be ignored | ref.array
# return       $aLines | sorted, normalized array of lines from $IniFileName | ref.array
#========================================================================================
sub _normalizeIni {
    my ($FileName, $aIgnore) = @_;

    my $IniConfig = DE_EPAGES::Core::API::IniConfig->new(
        FileName => $FileName, CaseSensitive => 1 );

    my @Content;
    foreach my $SectionName (sort @{$IniConfig->sectionNames}) {
        next if $SectionName eq 'Patch';
        push @Content, "[$SectionName]";
        my $hSection = $IniConfig->section($SectionName);
        foreach my $Key (sort keys %$hSection) {
            next if $Key ~~ @$aIgnore;
            my $Value = $hSection->{$Key};
            if ($Value =~ s,^['],,) {
                $Value =~ s,[']$,,;
            } elsif ($Value =~ s,^["],,) {
                $Value =~ s,["]$,,;
            }
            push @Content, "$Key=$Value";
        }
    }
    return \@Content;
}

#========================================================================================
# function     _anonymizeIni
# state        private
#----------------------------------------------------------------------------------------
# syntax       $AnonymousContent = _anonymizeIni($aLines, $aAnonymize);
#----------------------------------------------------------------------------------------
# description  Returns anonymized string from $aLines. A value in key=value is set to
#               ANONYMOUS if key belongs to $aAnonymize.
#----------------------------------------------------------------------------------------
# example      my $Content = _anonymizeIni(_normalizeIni($FileName),
#                                           [qw(host hosts masterUrl)]);
#----------------------------------------------------------------------------------------
# input        $aLines | array of lines | ref.array
# input        $aAnonymize | array of keys to anonymize | ref.array
# return       $Content | anonymized string | string
#========================================================================================
sub _anonymizeIni {
    my ($aLines, $aAnonymize) = @_;

    my $AnonymousContent;
    my ($Key, $Value);
    foreach my $Line (@$aLines) {
        if (($Key, $Value) = $Line =~ /(.+?)=(.*)/) {
            $Value = 'ANONYMOUS' if $Key ~~ @$aAnonymize;
            $AnonymousContent .= "$Key=$Value\n" if $Value =~ /\S/;
        } else {
            $AnonymousContent .= "$Line\n";
        }
    }
    return $AnonymousContent;
}

#========================================================================================
# function     _sortCartridgesXML
# state        private
#----------------------------------------------------------------------------------------
# syntax       _sortCartridgesXML($XMLFile);
#----------------------------------------------------------------------------------------
# description  sorts Cartridges.xml
#----------------------------------------------------------------------------------------
# input        $XMLFile | path of Cartridges.xml file | string
#========================================================================================
sub _sortCartridgesXML {
    my $self = shift;
    my ($XMLFile) = @_;

    my $PerlExe = $^O eq 'MSWin32' ? $^X : $ENV{PERL};
    my $Script = "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Installer/Scripts/sortCartridgesXml.pl";
    SystemCommand("\"$PerlExe\" \"$Script\" -file \"$XMLFile\" -inplace", { Print=>1 });

    my $aContent = GetFileContentLines($XMLFile);
    open(my $FileHandle, '>', $XMLFile) or die "$XMLFile: $!\n";
    foreach my $Line ( sort @$aContent ) {
        next if $Line =~ /DE_EPAGES::Diagnostics/;
        print $FileHandle $Line;
    }
    close($FileHandle) or die "$XMLFile: $!\n";
}

#========================================================================================
# function     _getFlexSiteStoreNames
# state        private
#----------------------------------------------------------------------------------------
# syntax       $aPatchStores = _getFlexSiteStoreNames();
#----------------------------------------------------------------------------------------
# description  get a list of all flex site Stores
#----------------------------------------------------------------------------------------
# return       $aPatchStores | list of all flex site stores | ref. array
#========================================================================================
sub _getFlexSiteStoreNames {

    my @PatchStores;
    #get flex-site stores
    RunOnStore(
        Store => 'Site',
        Sub => sub {
            my @AllStores = @{LoadRootObject()->get('Stores')};

            foreach my $Store (@AllStores) {
                my $StoreName = $Store->get('Alias');
                next if $StoreName eq 'Site';
                push @PatchStores, $StoreName;
            }
        },
    );

    return \@PatchStores;
}

#========================================================================================
# function     _getFlexStoreShopType
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShopType = $SystemInfoReportHandler->_getFlexStoreShopType();
#----------------------------------------------------------------------------------------
# description  returns the actual current ShopType like 'Flex_6_17_23'
#----------------------------------------------------------------------------------------
# return       $ShopType | current shop type | string
#========================================================================================
sub _getFlexStoreShopType {
    my $self = shift;

    my $Store = $self->_storeExists('Store') ? 'Store' : 'Site';
    my $epVersion = CartridgeConfig($Store)->get('Cartridges', 'DE_EPAGES::Core');

    #use just first 3 parts of version
    $epVersion =~ s/^((\d+\.){2}\d+).*/$1/;
    #substitute . with _
    $epVersion =~ s/\./_/g;

    return "Flex_$epVersion";
}


#========================================================================================
# function     _isFlexSiteInstallation
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $IsFlexSite = $SystemInfoReportHandler->_isFlexSiteInstallation();
#----------------------------------------------------------------------------------------
# description  check if current installation is a flex installation
#----------------------------------------------------------------------------------------
# return       $IsFlexSite | 0, if it is no flex installation, 1 otherwise | boolean
#========================================================================================
sub _isFlexSiteInstallation {
    my $self = shift;
    return 0 unless $self->_storeExists('Site');
    return 'DE_EPAGES::FlexProvider' ~~ GetStoreTypeCartridges(StoreType('Site'));
}

1;
