#========================================================================================
# package      DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBStructure
# state        public
#----------------------------------------------------------------------------------------
# description  This the install handler for the DB structure
#========================================================================================
package DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerDBStructure;

use strict;
use warnings;
use feature 'switch';
use Getopt::Long;
use Time::HiRes qw ( time );
use DE_EPAGES::Database::API::Connection qw (RunOnStore GetCurrentDBHandle );
use DE_EPAGES::Core::API::Log qw (GetLog);
use DE_EPAGES::Core::API::File qw (ExistsFile);

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler = DE_EPAGES::Installer::API::SystemInfoReport::SystemInfoReportHandlerFiles->new( $aFiles, $aStore );
#----------------------------------------------------------------------------------------
# description  saves the report of passed sections in class variables
#----------------------------------------------------------------------------------------
# input        $aFiles | list of all sections | string
# input        $aStore | list of all stores | ref array
# input        $hOptions | options for object initialization | ref.hash
# return       $DBStructureReportHandler | system report handler files | object
#========================================================================================
sub new {
  my $class = shift;
  my ( $aFiles, $aStore, $hOptions ) = @_;
  $hOptions = {} unless defined $hOptions;
  my $self = bless $hOptions, $class;

  #foreach passed file section
  foreach my $Section (@$aFiles) {

    #get time
    my $time1 = time;

    print "fetching $Section data\n";

    #save info for current section
    $self->_getInfo( $Section, $aStore );

    #get runtime
    my $runtime = time - $time1;

    #save runtime in class variable
    $self->{'runtime'}->{$Section} = sprintf( '%.3f', $runtime );
  }
  return $self;
}

#========================================================================================
# function     _getInfo
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_getInfo( $Section, $aStore );
#----------------------------------------------------------------------------------------
# description  saves the report of the dbstructure section in class variables
#----------------------------------------------------------------------------------------
# input        $Section | subsection, regulate the section to be saved | string
# input        $aStore | list of all stores | ref array
#========================================================================================
sub _getInfo {
    my $self = shift;
    my ( $Section, $aStore ) = @_;

    my $SystemFile = $ENV{EPAGES_CONFIG} . "/Database.d/System.conf";

    if ( $Section eq 'DBSchema' ) {
        #for each passed store
        foreach my $Store (@$aStore) {
            #get info data for each store
            $self->_getInfoDBSchema($Store);
        }

        #get also info data for ASPoolDB
        $self->_getInfoDBSchema('ASPoolDB');

        #get also info data for epagesj
        $self->_getInfoDBSchema('epagesj');
        
        # only get info, if systemdb is installed
        if ( ExistsFile($SystemFile)) {
            #get also info data for systemdb
            $self->_getInfoDBSchema('System');
        }
    }
    elsif( $Section eq 'ASPoolDB' or $Section eq 'epagesj' ){
        #get info data for ASPoolDB or epagesj
        $self->_getInfoDBSchema($Section);
    }
    # check also, if systemdb is installed
    elsif ( $Section eq 'System' and ExistsFile($SystemFile) ) {
        $self->_getInfoDBSchema($Section);
    }
    else{
        #get sqlite info
        $self->_getInfoSQLLiteDBSchema($Section);
    }
}

#========================================================================================
# function     getReport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->getReport( $Section );
#----------------------------------------------------------------------------------------
# description  export the report of the DBStructure
#----------------------------------------------------------------------------------------
# input        $Section | name of section to be exported | string
# return       $hResult | hash containing RND, QA and AM report of passed section as keys | hash ref
#========================================================================================
sub getReport {
  my $self = shift;
  my ($Section) = @_;

  #get result in dependence on section
  my $hReturn = $self->{ $Section . 'Hash' };
  return $hReturn;
}

#========================================================================================
# function     _getInfoDBSchema
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_getInfoDBSchema( $Store );
#----------------------------------------------------------------------------------------
# description  saves the report of DBSchema as RND, QA and AM version in class variables
#----------------------------------------------------------------------------------------
# input        $Store | name of store | string
#========================================================================================
sub _getInfoDBSchema {
  my $self = shift;
  my ($Store) = @_;

  #get result for different categories
  my $Result1->{'Columns'}                = $self->_generateOutputColumns($Store);
  my $Result2->{'GlobalVariables'}        = $self->_generateOutputGlobalVariables($Store);
  my $Result3->{'KeyColumnUsage'}         = $self->_generateOutputKeyColumnUsage($Store);
  my $Result4->{'ReferentialConstraints'} = $self->_generateOutputReferentialConstraints($Store);
  my $Result5->{'Tables'}                 = $self->_generateOutputTables($Store);
  my $Result6->{'NonUniqueIndexes'}       = $self->_generateOutputNonUniqueIndexes($Store);
  my $Result7->{'Views'}                  = $self->_generateOutputViews($Store);

  #merge results to one resulthash
  my %Result = ( %$Result1, %$Result2, %$Result3, %$Result4, %$Result5, %$Result6, %$Result7 );

  #save result in class variable with section as first key an store as second key
  $self->{'DBSchemaHash'}->{'RND'}->{$Store} = \%Result;
  $self->{'DBSchemaHash'}->{'QA'}->{$Store}  = \%Result;
  $self->{'DBSchemaHash'}->{'AM'}->{$Store}  = \%Result;

  return;
}

#========================================================================================
# function     _getInfoSQLLiteDBSchema
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_getInfoSQLLiteDBSchema( $Section );
#----------------------------------------------------------------------------------------
# description  saves the report of ASPoolDB or EbayCategoriesDB
#               as RND, QA and AM version in class variables
#----------------------------------------------------------------------------------------
# input        $Section | indicates which db to export | string
#========================================================================================
sub _getInfoSQLLiteDBSchema {
  my $self = shift;
  my ($Section) = @_;
  my $aaResult;

  #set name of database
  my $Sqlitedb;
  given ($Section) {
    when ('EbayCategoriesDB') {
      $Sqlitedb = $ENV{'EPAGES_CONFIG'} . '/EbayCategories.db';
    }
  }

  #connect to database
  my $Dbh = DBI->connect("dbi:SQLite:dbname=$Sqlitedb");

  #prepare and execute SQL command
  my $Sth = $Dbh->prepare('SELECT tbl_name, type, name, sql FROM sqlite_master ORDER BY tbl_name,type,name');
  $Sth->execute();

  #build result hash
  my @Array;
  my $i = 0;
  while ( @Array = $Sth->fetchrow_array ) {
    push @{ $aaResult->[$i] }, $Array[0], $Array[1], $Array[2], $Array[3];
    $i++;
  }

  $Sth->finish;
  $Dbh->disconnect;

  #save result in class variable with section as first key
  $self->{ "$Section" . 'Hash' }->{'RND'} = $aaResult;
  $self->{ "$Section" . 'Hash' }->{'QA'}  = $aaResult;
  $self->{ "$Section" . 'Hash' }->{'AM'}  = $aaResult;
}

#========================================================================================
# function     _generateOutputColumns
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputColumns( $Store );
#----------------------------------------------------------------------------------------
# description  generate information about all columns used in passed store
#----------------------------------------------------------------------------------------
# input        $Store | name of store used for export | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputColumns {

  #/Patch/PatchDiff/DBSchema/dbschema_mysql.pl
  my $self = shift;
  my ($Store) = @_;
  my $hResult;

  RunOnStore(
    Store      => "$Store",
    DBPassword => undef,
    Sub        => sub {

      #COLUMNS
      $hResult = $self->_generateOutputFile( <<END_SQL );
SELECT DISTINCT
  TABLE_NAME,
  COLUMN_NAME,
  ORDINAL_POSITION,
  COLUMN_DEFAULT,
  IS_NULLABLE,
  DATA_TYPE,
  COLUMN_DEFAULT
  CHARACTER_MAXIMUM_LENGTH,
  CHARACTER_OCTET_LENGTH,
  NUMERIC_PRECISION,
  NUMERIC_SCALE,
  CHARACTER_SET_NAME,
  COLLATION_NAME,
  COLUMN_TYPE,
  COLUMN_KEY,
  EXTRA,
  PRIVILEGES,
  COLUMN_COMMENT
FROM INFORMATION_SCHEMA.COLUMNS
WHERE TABLE_SCHEMA = DATABASE()
ORDER BY
  TABLE_NAME,
  ORDINAL_POSITION
END_SQL
    }
  );
  return $hResult;
}

#========================================================================================
# function     _generateOutputGlobalVariables
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputGlobalVariables( $Store );
#----------------------------------------------------------------------------------------
# description  generate information about global variables used in passed store
#----------------------------------------------------------------------------------------
# input        $Store | name of store used for export | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputGlobalVariables {
  my $self = shift;
  my ($Store) = @_;
  
  my $Dbi;
  my $hResult;
  RunOnStore(
    Store      => "$Store",
    DBPassword => undef,
    Sub        => sub {
      $Dbi = GetCurrentDBHandle();

      #GLOBAL_VARIABLES
      if ( $self->_existsInformationSchemaTable( $Dbi, 'GLOBAL_VARIABLES' ) ) {
        $hResult = $self->_generateOutputFile( <<END_SQL );
SELECT *
FROM INFORMATION_SCHEMA.GLOBAL_VARIABLES
WHERE VARIABLE_NAME NOT IN(
    'BASEDIR',
    'CHARACTER_SETS_DIR',
    'DATADIR',
    'GENERAL_LOG_FILE',
    'INNODB_DATA_FILE_PATH',
    'INNODB_DATA_HOME_DIR',
    'INNODB_LOG_GROUP_HOME_DIR',
    'LANGUAGE',
    'LOG_ERROR',
    'PID_FILE',
    'PLUGIN_DIR',
    'PORT',
    'RELAY_LOG_INDEX',
    'RELAY_LOG_INFO_FILE',
    'SLAVE_LOAD_TMPDIR',
    'SLOW_QUERY_LOG_FILE',
    'SOCKET',
    'TIMESTAMP',
    'TMPDIR',
    'HOSTNAME'
)
ORDER BY VARIABLE_NAME
END_SQL
      }

      else {
        GetLog->warn('PatchDiff Warning: table \'INFORMATION_SCHEMA.GLOBAL_VARIABLES\' does not exist');
      }

    }
  );
  return $hResult;
}

#========================================================================================
# function     _generateOutputKeyColumnUsage
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputKeyColumnUsage( $Store );
#----------------------------------------------------------------------------------------
# description  generate information about used key columns in passed store
#----------------------------------------------------------------------------------------
# input        $Store | name of store used for export | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputKeyColumnUsage {
  my $self = shift;
  my ($Store) = @_;
  my $hResult;
  RunOnStore(
    Store      => "$Store",
    DBPassword => undef,
    Sub        => sub {

      #KEY_COLUMN_USAGE
      $hResult = $self->_generateOutputFile( <<END_SQL );
SELECT
  CONSTRAINT_NAME,
  TABLE_NAME,
  COLUMN_NAME,
  ORDINAL_POSITION,
  REFERENCED_TABLE_NAME,
  POSITION_IN_UNIQUE_CONSTRAINT,
  REFERENCED_COLUMN_NAME
FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
WHERE CONSTRAINT_SCHEMA = DATABASE()
ORDER BY
  CONSTRAINT_NAME,
  TABLE_NAME,
  ORDINAL_POSITION
END_SQL
    }
  );
  return $hResult;
}

#========================================================================================
# function     _generateOutputReferentialConstraints
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputReferentialConstraints( $Store );
#----------------------------------------------------------------------------------------
# description  generate information about constraints used in passed store
#----------------------------------------------------------------------------------------
# input        $Store | name of store used for export | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputReferentialConstraints {
  my $self = shift;
  my ($Store) = @_;
  my $Dbi;
  
  my $hResult;
  RunOnStore(
    Store      => "$Store",
    DBPassword => undef,
    Sub        => sub {
      $Dbi = GetCurrentDBHandle();

      #REFERENTIAL_CONSTRAINTS
      if ( $self->_existsInformationSchemaTable( $Dbi, 'REFERENTIAL_CONSTRAINTS' ) ) {
        $hResult = $self->_generateOutputFile( <<END_SQL );
SELECT
  CONSTRAINT_NAME,
  TABLE_NAME,
  REFERENCED_TABLE_NAME,
  UNIQUE_CONSTRAINT_NAME,
  MATCH_OPTION,
  UPDATE_RULE,
  DELETE_RULE
FROM INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS
WHERE CONSTRAINT_SCHEMA = DATABASE()
ORDER BY
  CONSTRAINT_NAME,
  TABLE_NAME
END_SQL
      }
      else {
        GetLog->warn('PatchDiff Warning: table \'INFORMATION_SCHEMA.REFERENTIAL_CONSTRAINTS\' does not exist, outdatet mysql version');
      }
    }
  );
  return $hResult;
}

#========================================================================================
# function     _generateOutputTables
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputTables( $Store );
#----------------------------------------------------------------------------------------
# description  generate information about tables used in passed store
#----------------------------------------------------------------------------------------
# input        $Store | name of store used for export | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputTables {
  my $self = shift;
  my ($Store) = @_;
  my $hResult;
  RunOnStore(
    Store      => "$Store",
    DBPassword => undef,
    Sub        => sub {

      #TABLES
      $hResult = $self->_generateOutputFile( <<END_SQL );
SELECT
  TABLE_NAME,
  TABLE_TYPE,
  ENGINE,
  VERSION,
  ROW_FORMAT,
  MAX_DATA_LENGTH
FROM INFORMATION_SCHEMA.TABLES
WHERE TABLE_SCHEMA = DATABASE()
ORDER BY TABLE_NAME
END_SQL
    }
  );
  return $hResult;
}

#========================================================================================
# function     _generateOutputNonUniqueIndexes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputNonUniqueIndexes( $Store );
#----------------------------------------------------------------------------------------
# description  generate information about non unique indexes used in passed store
#----------------------------------------------------------------------------------------
# input        $Store | name of store used for export | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputNonUniqueIndexes {
  my $self = shift;
  my ($Store) = @_;
  my $hResult;
  RunOnStore(
    Store      => "$Store",
    DBPassword => undef,
    Sub        => sub {

      #Non-unique indexes
      $hResult = $self->_generateOutputFile( <<END_SQL );
SELECT
   TABLE_NAME,
   COLUMN_NAME,
   SEQ_IN_INDEX,
   INDEX_NAME,
   COLLATION,
   NULLABLE,
   INDEX_TYPE,
   COMMENT
FROM INFORMATION_SCHEMA.STATISTICS
WHERE TABLE_SCHEMA = DATABASE()
  AND NON_UNIQUE = 1
ORDER BY
  TABLE_NAME,
  INDEX_NAME,
  SEQ_IN_INDEX
END_SQL
    }
  );
  return $hResult;
}

#========================================================================================
# function     _generateOutputViews
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputViews( $Store );
#----------------------------------------------------------------------------------------
# description  generate information about views used in passed store
#----------------------------------------------------------------------------------------
# input        $Store | name of store used for export | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputViews {
  my $self = shift;
  my ($Store) = @_;
  my $hResult;
  RunOnStore(
    Store      => "$Store",
    DBPassword => undef,
    Sub        => sub {

      #Non-unique indexes
      $hResult = $self->_generateOutputFile( <<END_SQL );
SELECT
  TABLE_NAME,
  VIEW_DEFINITION,
  CHECK_OPTION,
  IS_UPDATABLE,
  DEFINER,
  SECURITY_TYPE
FROM INFORMATION_SCHEMA.VIEWS
WHERE TABLE_SCHEMA = DATABASE()
ORDER BY
  TABLE_NAME
END_SQL
    }
  );
  return $hResult;
}

#========================================================================================
# function     _generateOutputFile
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_generateOutputFile( $SQL );
#----------------------------------------------------------------------------------------
# description  executes passed SQL statement and returns edited result as hash
#----------------------------------------------------------------------------------------
# input        $SQL | sql statement | string
# return       $hResult | resulthash  | ref hash
#========================================================================================
sub _generateOutputFile {
  my $self = shift;
  my ($SQL) = @_;
  my $Dbh;
  
  $Dbh = GetCurrentDBHandle()->dbh;

  $Dbh->{RaiseError} = 1;
  my $Sth = $Dbh->prepare($SQL);

  $Sth->execute;
  my $hResult;

  $hResult = $self->_showResultSetAsText($Sth);

  $Sth->finish;
  return $hResult;
}

#========================================================================================
# function     _showResultSetAsText
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_showResultSetAsText( $Sth );
#----------------------------------------------------------------------------------------
# description  convert SQL tables statement to two dimensional array
#----------------------------------------------------------------------------------------
# input        $Sth | DBI statement handle | object
# return       $aaResult | two-dimensional array containing sql table | ref array
#========================================================================================
sub _showResultSetAsText {
  my $self = shift;
  my ($Sth) = @_;

  my $aaResult;

  #get column names
  my $aColumns = $Sth->{NAME};

  #add names to array
  push @{ $aaResult->[0] }, @$aColumns;

  if ( 0 == $Sth->{NUM_OF_FIELDS} ) {
    return;
  }

  #fetch all rows
  my $aValues;
  my $i = 1;

  #save rows in an array
  while ( $aValues = $Sth->fetchrow_arrayref ) {

    #every line is an array of single values
    my @Line;

    #for each column
    foreach my $Index ( 0 .. scalar(@$aColumns) - 1 ) {

      #get value at position $Index
      my $Value = $aValues->[$Index];

      #fill current line with values
      push @Line, $Value;
    }

    #fill result array with lines
    push @{ $aaResult->[$i] }, @Line;
    $i = $i + 1;
  }

  return $aaResult;
}

#========================================================================================
# function     _existsInformationSchemaTable
# state        private
#----------------------------------------------------------------------------------------
# syntax       $DBStructureReportHandler->_existsInformationSchemaTable( $Dbi, $TableName );
#----------------------------------------------------------------------------------------
# description  roadworks
#----------------------------------------------------------------------------------------
# input        $Dbi | database handle | object
# input        $TableName
# return
#========================================================================================
sub _existsInformationSchemaTable {
  my $self = shift;
  my ( $Dbi, $TableName ) = @_;

  my $aaResult = $Dbi->execute( <<END_SQL, [$TableName] );
SELECT table_name
FROM information_schema.tables
WHERE table_schema = 'INFORMATION_SCHEMA' and table_name=?
END_SQL

  return scalar @$aaResult > 0 ? 1 : 0;    #???
}

1;
