#========================================================================================
# package      DE_EPAGES::Ipayment::API::SOAP::WebServiceClient
# state        public
# description  object which manages the web service request
#========================================================================================
package DE_EPAGES::Ipayment::API::SOAP::WebServiceClient;

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::TestParameter qw ( TestObject );
use DE_EPAGES::Ipayment::API::SOAP::Request::Validator;
use DE_EPAGES::Ipayment::API::Logger;

use strict;



#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( $CompiledClient );
#----------------------------------------------------------------------------------------
# description  Create an instance of class WebServiceClient
#----------------------------------------------------------------------------------------
# input        $CompiledClient | CompiledClient | DE_EPAGES::Ipayment::API::SOAP::CompiledClient
# return       $self | WebServiceClient | DE_EPAGES::Ipayment::API::SOAP::WebServiceClient
#========================================================================================
sub new {
    my $class = shift;
    my ($CompiledClient) = @_;
    TestObject('CompiledClient',$CompiledClient,'DE_EPAGES::Ipayment::API::SOAP::CompiledClient');
    my $self  = bless {
        'CompiledClient' => $CompiledClient,
    }, $class;
    $self->{'Validator'} = DE_EPAGES::Ipayment::API::SOAP::Request::Validator->new();
    return $self;
}

#========================================================================================
# function     setValidator
# state        public
#----------------------------------------------------------------------------------------
# syntax       setValidator( $Validator );
#----------------------------------------------------------------------------------------
# description  Sets a validator
#----------------------------------------------------------------------------------------
# input        $Validator | Validator | DE_EPAGES::Ipayment::API::SOAP::Request::Validator
#========================================================================================
sub setValidator {
    my $self = shift;
    my ($Validator) = @_;
    TestObject('Validator', $Validator, 'DE_EPAGES::Ipayment::API::SOAP::Request::Validator');
    $self->{'Validator'} = $Validator;
    return undef;
}

#========================================================================================
# function     doRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       doRequest( $HashBuilderObject );
#----------------------------------------------------------------------------------------
# description  Does the call to the web service with the data in the hash builder object
#----------------------------------------------------------------------------------------
# input        $HashBuilderObject | hash builder object | DE_EPAGES::Ipayment::API::SOAP::Request::HashBuilder
# return       $Response | TODO | refHash
#========================================================================================
sub doRequest {
    my $self = shift;
    my ( $HashBuilderObject ) = @_;

    TestObject('HashBuilderObject',$HashBuilderObject,'DE_EPAGES::Ipayment::API::SOAP::Request::HashBuilder');
    my $hRequestData = $HashBuilderObject->getRequestData();
    my $Response;

    if ( $self->{'Validator'}->isValidHash($hRequestData) ){
        my $Trace;
        my $Call = $self->{'CompiledClient'}->createClient();
        eval {
            ( $Response, $Trace ) = $Call->($hRequestData);
        };
        if (ExistsError()) {
            my $Error = GetError();
            Error($Error) if ( $Error->code eq 'NativeDie');
            $self->_LogCommunicationWithoutPasswords($Error->code(), $hRequestData);
            return undef;
        }
        $self->_LogCommunicationWithoutPasswords( 'SOAPRequest', ${$Trace->request()->{'_content'}} );
        $self->_LogCommunicationWithoutPasswords( 'SOAPResponse', $Trace->response()->{'_content'} );
    } else {
        $self->_LogCommunicationWithoutPasswords('Data invalid, checked before call', $hRequestData);
    }

    return $Response;
}


#========================================================================================
# function     _LogCommunicationWithoutPasswords
# state        private
#----------------------------------------------------------------------------------------
# syntax       _LogCommunicationWithoutPasswords( $Name, $Trace );
#----------------------------------------------------------------------------------------
# description  replace the login data from request and log
#----------------------------------------------------------------------------------------
# input        $Name | name for logging entry | String
# input        $Trace | the trace of the WSDL11 request | String
#========================================================================================
sub _LogCommunicationWithoutPasswords {
    my $self = shift;
    my ($Name, $LogData) = @_;

    if(ref $LogData eq 'HASH'){
        $LogData->{'accountData'}->{'trxpassword'} = '---replaced---' if exists($LogData->{'accountData'}->{'trxpassword'});
        $LogData->{'accountData'}->{'adminactionpassword'} = '---replaced---'if exists($LogData->{'accountData'}->{'adminactionpassword'});
    }else{
        $LogData =~s/(<trxpassword>).*?(<\/trxpassword>)/$1---replaced---$2/msxig;
        $LogData =~s/(<adminactionpassword>).*?(<\/adminactionpassword>)/$1---replaced---$2/msxig;
    }

    $self->_LogPayment('Ipayment', $Name, $LogData, 'createSession');
    return;
}

#========================================================================================
# function     _LogPayment
# state        private
#----------------------------------------------------------------------------------------
# syntax       _LogPayment($PaymentTypeAlias, $Name, $Trace, $Section);
#----------------------------------------------------------------------------------------
# description  wrapp core function for unit tests
#----------------------------------------------------------------------------------------
# input        $PaymentTypeAlias | Payment Type Alias | String
# input        $Name | Name | String
# input        $DebugStuff | the trace of the WSDL11 request | String
# input        $Section | name of the section to be logged (optional - defaults to name of calling sub) | string
# input        $Filter | (optional) filter (perl regular expression | ref.array.string
# return       $Logger
#========================================================================================
sub _LogPayment {
    my $self = shift;
    my ($PaymentTypeAlias, $Name, $DebugStuff, $Section, $Filter) = @_;
    my $Logger = DE_EPAGES::Ipayment::API::Logger->instance();
    return $Logger->logPayment($PaymentTypeAlias, $Name, $DebugStuff, $Section, $Filter);
}

1;