#========================================================================================
# package      DE_EPAGES::Order::API::OrderSearchService
# state        public
#----------------------------------------------------------------------------------------
# description  Collects Order Search data
#========================================================================================
package DE_EPAGES::Order::API::OrderSearchService;

use DE_EPAGES::WebService::API::Util qw (  ConvertDateForEP6 );
use DE_EPAGES::Core::API::Log qw ( LogDebug );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::Search::API::SearchDefinition qw ( GetSearchDefinition );

use strict;



#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( );
#----------------------------------------------------------------------------------------
# description  Constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Order::API::OrderSearchService
#========================================================================================
sub new {
    my $class = shift;
    my ($Shop, $ResultsPerPage, $Page, $hQuery) = @_;
    
    my $self  = bless {
        'Shop' => $Shop,
        'ResultsPerPage' => $ResultsPerPage,
        'Page' => $Page,
        'Query' => $hQuery,
        'SearchName' => 'SimpleSearch',
        'OrderBy' => 'CreationDate',
        'SearchForm' => {},
        'SearchParameters' => {}

    }, $class;
    return $self;
}

#========================================================================================
# function     GetSearchParameters
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hSearchParameters = $OrderSearchService->GetSearchParameters();
#----------------------------------------------------------------------------------------
# description  GetSearchParameters returns keys and values to be used in DAL search
#----------------------------------------------------------------------------------------
# return       $hSearchParameters | DAL related info for our order search | ref.hash
#========================================================================================
sub GetSearchParameters {
    my $self = shift;

    $self->{'SearchForm'}->{'Search'} = $self->{'SearchName'};
    delete $self->{'SearchForm'}->{$_} foreach grep { not defined $self->{'SearchForm'}->{$_} } keys % {$self->{'SearchForm'}};

    if($self->{'OrderBy'} eq 'LastUpdated'){
        $self->{'SearchParameters'}->{'UriParameters'}{'Search'} = 'LastUpdatedSearch';
    }

    foreach my $Parameter (keys % {$self->{'SearchForm'}} ) {
        if ($Parameter =~ /String/) {
            # replace * with %
            $self->{'SearchParameters'}->{'Parameters'}{$Parameter} = $self->{'SearchForm'}->{$Parameter};
        } elsif ($Parameter =~ /(.*)On$/) {
            # status = 0/1/2 egal/yes/no
            my $Status = $1;
            if ($self->{'SearchForm'}->{$Parameter} == 1) {
                $self->{'SearchParameters'}->{'Parameters'}{'Is'.$Status} = 1;
            } elsif ($self->{'SearchForm'}->{$Parameter} == 2) {
                $self->{'SearchParameters'}->{'Parameters'}{'Not'.$Status} = 1;
            }
        } else {
            $self->{'SearchParameters'}->{'Parameters'}{$Parameter} = $self->{'SearchForm'}->{$Parameter};
        }

    }

    if (defined $self->{'SearchParameters'}->{'Parameters'}{'ProductAliasString'}) {
        #ProductLineItemClassIDs
        my $LineItemProduct = $self->_LoadClassByAlias('LineItemProduct');
        my $Classes = $LineItemProduct->subClasses();
        unshift @$Classes, $LineItemProduct;
        $self->{'SearchParameters'}->{'Parameters'}{'ProductLineItemClassIDs'} = [ map { $_->id() } @$Classes ];
    }

    LogDebug('SearchParameters', $self->{'SearchParameters'});
    return $self->{'SearchParameters'};
}

#========================================================================================
# function     DefineSearchParameters
# state        public
#----------------------------------------------------------------------------------------
# syntax       $OrderSearchService->DefineSearchParameters();
#----------------------------------------------------------------------------------------
# description  Defines SearchParameters used in DAL
#========================================================================================
sub DefineSearchParameters {
    my $self = shift;

    $self->{'SearchParameters'} = {
        'ObjectID'=> $self->{'Shop'}->id(),
        'Parameters' => {
            'Shop.ID'    => $self->{'Shop'}->id(),
            'IsVisible'  => 1,
        },
        'PageSize'=>$self->{'ResultsPerPage'},
        'OrderBy'    => $self->{'OrderBy'},
        'OrderDesc'  => 1,
        'Object' => undef,
        'SearchDefinition' => undef
    };

    $self->{'SearchParameters'}{'Object'} = $self->{'Shop'};

    my $SearchDefinition = $self->_GetSearchDefinition($self->_LoadClassByAlias('CustomerOrder'), $self->{'SearchName'});
    $self->{'SearchParameters'}{'SearchDefinition'} = $SearchDefinition;
}

#========================================================================================
# function     ConvertEpagesJValuesToEpages6
# state        public
#----------------------------------------------------------------------------------------
# syntax       $OrderSearchService->ConvertEpagesJValuesToEpages6();
#----------------------------------------------------------------------------------------
# description  ConvertEpagesJValuesToEpages6 changes syntax of keys and values 
#               from epagesj style to epages6 style
#========================================================================================
sub ConvertEpagesJValuesToEpages6 {
    my $self = shift;

    LogDebug('Query', $self->{'Query'});

    foreach my $Key (keys %{$self->{'Query'}}) {

        # Check for (.*)On aka filters on status
        if ($Key =~ /(.*)On$/) {   ## no critic (ProhibitCascadingIfElse)

            #camel-case for epagesj
            my $NewKey = ucfirst($Key);
            my $Value = $self->{'Query'}->{$Key}->[0];
            if(isValid($Value)){
                $self->{'SearchForm'}->{$NewKey} = $self->{'Query'}->{$Key}->[0] eq 'true'?'1':'2';
            }

        } elsif ($Key =~ /(.*)Before$/) {
            
            my $NewKey = ucfirst($Key);
            my $Value = $self->{'Query'}->{$Key}->[0];
            if(isValid($Value)){
                # if value exists convert date for EP6
                $self->{'SearchForm'}->{$NewKey} = ConvertDateForEP6($Value);
            }

        } elsif ($Key =~ /(.*)After$/) {

            my $NewKey = ucfirst($Key);
            my $Value = $self->{'Query'}->{$Key}->[0];
            if(isValid($Value)){
                # if value exists convert date for EP6
                $self->{'SearchForm'}->{$NewKey} = ConvertDateForEP6($Value);
            }

        } elsif($Key eq 'customerId'){

            # Search orders for a customer guid
            my $Customer = $self->{'Shop'}->loadObjectByGuid( $self->{'Query'}->{$Key}->[0], 'Customer' );
            $self->{'SearchForm'}->{'CustomerID'} = $Customer->id();

        } elsif($Key eq 'productId'){

            # Search orders for a product guid
            my $Product = $self->{'Shop'}->loadObjectByGuid( $self->{'Query'}->{$Key}->[0], 'Product' );
            $self->{'SearchForm'}->{'ProductAliasString'} = $Product->alias();

        } elsif($Key eq 'currency'){

                $self->{'SearchForm'}->{'CurrencyID'} = $self->{'Query'}->{$Key}->[0];
        } elsif($Key eq 'updated'){

            if($self->{'Query'}->{$Key}->[0] eq 'true'){
                $self->{'OrderBy'}    = 'LastUpdated';
            }

        } elsif($Key eq 'updatedFrom'){
                $self->{'SearchForm'}->{'UpdatedDateFrom'} = ConvertDateForEP6($self->{'Query'}->{$Key}->[0]);
                $self->{'OrderBy'}    = 'LastUpdated';
        }

        # Set searchname based on LastUpdated.  We have saperate sql query for UpdateDate lookup since it requires additional join.
        if($self->{'OrderBy'} eq 'LastUpdated') {
            $self->{'SearchName'} = 'LastUpdatedSearch';
        } else {
            $self->{'SearchName'} = 'RESTSearch';
        }
    }
}

sub _LoadClassByAlias {
    my $self = shift;
    my ($ClassName) = @_;
    return LoadClassByAlias($ClassName);
}

sub _GetSearchDefinition {
    my $self = shift;
    my ($ResultClass, $SearchName) = @_;
    return GetSearchDefinition($ResultClass, $SearchName);
}

1;