#========================================================================================
# package      DE_EPAGES::Order::API::TLE::OrderDocument
# description  TLE functions for order documents
# state        public
#========================================================================================
package DE_EPAGES::Order::API::TLE::OrderDocument;

use strict;

use DE_EPAGES::Core::API::File qw ( ExistsFile );
use DE_EPAGES::Core::API::Image qw ( GetImageInfo ConvertCmToPx );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::TestParameter qw( TestObject );
use DE_EPAGES::Order::API::Constants qw( PDF_DENSITY );
use DE_EPAGES::Order::API::OrderDocument qw (
    GetLegalDocumentPDFAttachment
    GetLegalDocumentPDFWebFileName
);
use DE_EPAGES::Presentation::API::SiteTLEProcessor qw( ReplaceTLEs  );
use List::Util qw( max );
use POSIX;

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function registers this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;
    GetLog->debug( 'OrderDocument.RegisterHandlerProc' );

    __PACKAGE__->new()->register( $Params->{'Processor'} );
    return;
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::Object::API::TLE::OrderHandler;
# example      my $PrefixHandler  = new DE_EPAGES::Object::API::TLE::OrderHandler;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  this function register this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    GetLog->debug( 'OrderDocument.register' );

    $Processor->registerHandler('FunctionHandler', $self, 'CustomerInformationPDFWebFileName');
    $Processor->registerHandler('FunctionHandler', $self, 'UnityOrderDocumentsFooter');
    $Processor->registerHandler('FunctionHandler', $self, 'UnityOrderDocumentsInvoiceText');
    $Processor->registerHandler('FunctionHandler', $self, 'UnityOrderDocumentsPackingSlipText');
    $Processor->registerHandler('FunctionHandler', $self, 'UnityOrderDocumentsCreditNoteText');
    $Processor->registerHandler('FunctionHandler', $self, 'GetLogoPrintDimensions');
    return;
}

#========================================================================================
# function     CustomerInformationPDFWebFileName
# state        public
#----------------------------------------------------------------------------------------
# syntax       #CustomerInformationPDFWebFileName(#Shop.Object, #LineItemContainer.LanguageID)
#----------------------------------------------------------------------------------------
# description  Returns the translated web path to the customer information PDF.
#----------------------------------------------------------------------------------------
# input        #Shop.Object | actual shop object | object
# input        #LanguageID | line item container language id | integer
#----------------------------------------------------------------------------------------
# return        #WebFileName | translated web path to the customer information PDF | string
#========================================================================================
sub CustomerInformationPDFWebFileName { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;

    my ($Processor, $aParams) = @_;
    my ($Shop, $OrderLanguageID) = @$aParams;
    my $CustomerInformation = $Shop->get('CustomerInformation');

    # create PDF if necessary
    my $Success = GetLegalDocumentPDFAttachment($Shop, $OrderLanguageID, $CustomerInformation);

    return $Success ? GetLegalDocumentPDFWebFileName($Shop, $OrderLanguageID, $CustomerInformation) : undef;
}

#========================================================================================
# function     UnityOrderDocumentsFooter
# state        public
#----------------------------------------------------------------------------------------
# syntax       #UnityOrderDocumentsFooter(#Order)
#----------------------------------------------------------------------------------------
# description  Returns the tle replaced OrderDocumentsFooter.
#----------------------------------------------------------------------------------------
# input        #Order | order object | object
#========================================================================================
sub UnityOrderDocumentsFooter {
    my $self = shift;
    my ($Processor, $aParams) = @_;

    return $self->_processShopAttribute($Processor, $aParams, 'OrderDocumentsFooter');
}

#========================================================================================
# function     UnityOrderDocumentsInvoiceText
# state        public
#----------------------------------------------------------------------------------------
# syntax       #UnityOrderDocumentsInvoiceText(#Order)
#----------------------------------------------------------------------------------------
# description  Returns the tle replaced OrderDocumentsInvoiceText1.
#----------------------------------------------------------------------------------------
# input        #Order | order object | object
#========================================================================================
sub UnityOrderDocumentsInvoiceText {
    my $self = shift;
    my ($Processor, $aParams) = @_;

    return $self->_processShopAttribute($Processor, $aParams, 'OrderDocumentsInvoiceText1');
}

#========================================================================================
# function     UnityOrderDocumentsPackingSlipText
# state        public
#----------------------------------------------------------------------------------------
# syntax       #UnityOrderDocumentsPackingSlipText(#Order)
#----------------------------------------------------------------------------------------
# description  Returns the tle replaced OrderDocumentsPackingSlipText1.
#----------------------------------------------------------------------------------------
# input        #Order | order object | object
#========================================================================================
sub UnityOrderDocumentsPackingSlipText {
    my $self = shift;
    my ($Processor, $aParams) = @_;

    return $self->_processShopAttribute($Processor, $aParams, 'OrderDocumentsPackingSlipText1');
}

#========================================================================================
# function     UnityOrderDocumentsCreditNoteText
# state        public
#----------------------------------------------------------------------------------------
# syntax       #UnityOrderDocumentsCreditNoteText(#Order)
#----------------------------------------------------------------------------------------
# description  Returns the tle replaced OrderDocumentsCreditNoteText1.
#----------------------------------------------------------------------------------------
# input        #Order | order object | object
#========================================================================================
sub UnityOrderDocumentsCreditNoteText {
    my $self = shift;
    my ($Processor, $aParams) = @_;

    return $self->_processShopAttribute($Processor, $aParams, 'OrderDocumentsCreditNoteText1');
}

#========================================================================================
# function     GetLogoPrintDimensions
# state        public
#----------------------------------------------------------------------------------------
# syntax       #GetLogoPrintDimensions(#Shop.Object, #MaxWidth, #MaxHeight)
#----------------------------------------------------------------------------------------
# description  Calculates the logo dimensions to fit into the given measures
#----------------------------------------------------------------------------------------
# input        #Shop | shop object | object
# input        #MaxWidthCm | max width in cm | integer
# input        #MaxHeightCm | max width in cm | integer
# return       #LogoDimensions | logo dimensions, keys: | hash.ref
#                                   <ul>
#                                      <li>Height</li>
#                                      <li>Width</li>
#                                   </ul>
#========================================================================================
sub GetLogoPrintDimensions {
    my $self = shift;
    my ($Processor, $aParams) = @_;
    my ($Shop, $MaxWidthCm, $MaxHeightCm) = @$aParams;

    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');

    my $hInfoShop = $Shop->get(['Logo', 'PublicPath']);

    return unless defined $hInfoShop->{'Logo'};

    # construct logo file name
    my $LogoPath = $hInfoShop->{'PublicPath'} . '/' . $hInfoShop->{'Logo'};

    return unless ExistsFile($LogoPath);

    # convert image dimensions to pixel
    my $MaxWidthPixels = ConvertCmToPx($MaxWidthCm, PDF_DENSITY);
    my $MaxHeightPixels = ConvertCmToPx($MaxHeightCm, PDF_DENSITY);

    # define scale
    my $hImageInfo = GetImageInfo($LogoPath);
    my $Scale = max( $hImageInfo->{'Width'} / $MaxWidthPixels, $hImageInfo->{'Height'} / $MaxHeightPixels );

    # calculate rescaled dimensions
    my %LogoDimensions;
    if ( $Scale < 1 ) {
        # no rescaling
        $LogoDimensions{'Width'} = $hImageInfo->{'Width'};
        $LogoDimensions{'Height'} = $hImageInfo->{'Height'};
    } else {
        $LogoDimensions{'Width'} = floor($hImageInfo->{'Width'} / $Scale);
        $LogoDimensions{'Height'} = floor($hImageInfo->{'Height'} / $Scale);
    }

    return \%LogoDimensions;
}


#========================================================================================
# function     _processShopAttribute
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_processShopAttribute($Processor, $aParams, $ShopAttribute)
#----------------------------------------------------------------------------------------
# description  Returns the tle replaced shop attribute in the context of the given order
#               object.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
# input        $aParams   | tle params | array.ref
# input        $ShopAttribute   | name of the shop attribute | string
# return       $Text   | tle processed shop attribute | string
#========================================================================================
sub _processShopAttribute {
    my $self = shift;
    my ($Processor, $aParams, $ShopAttribute) = @_;
    my ($Order) = @$aParams;
    TestObject('CustomerOrder', $Order, 'DE_EPAGES::Order::API::Object::CustomerOrder');

    my $Shop = $Order->getSite();
    my $TLEText = $Shop->get($ShopAttribute);

    # use locale from the original processor (important for formatters)
    my $Locale = $Processor->tle('Locale') // $Processor->tle('INPUT.Locale');

    my $hVarsObject = $Order->tleHash;
    $hVarsObject->{'INPUT'} = {
            'ObjectID' => $Order->id,
            'Locale' => $Locale
    };

    return ReplaceTLEs($TLEText, $hVarsObject);
}

1;
