#========================================================================================
# package      DE_EPAGES::Order::Hooks::ShopMailType
# state        public
# base         Exporter
# description  hook functions for ShopMailType.
#========================================================================================
package DE_EPAGES::Order::Hooks::ShopMailType;
use strict;

use DE_EPAGES::Object::API::XML::AttributeImportHandler qw( ImportTranslation );
use DE_EPAGES::Order::API::Cartridge;
use DE_EPAGES::Order::API::OrderDocument qw ( GetLegalDocumentPDFAttachment );

use constant VIEWACTION_NO_TABS => "MBO-ViewEmailSettings";


#========================================================================================
# function     OnShopMailTypeCreated
# state        public
#----------------------------------------------------------------------------------------
# syntax       OnShopMailTypeCreated($hParams);
#----------------------------------------------------------------------------------------
# description  Triggered by hook API_ShopMailTypeCreated.
#               Import translations for new created shopmailtypes
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>ShopMailType - shop mailtype object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnShopMailTypeCreated {
    my ($hParams) = @_;

    return unless $hParams->{'ShopMailType'};
    my $ShopMailType = $hParams->{'ShopMailType'};

    # import loalization for this new created shop mailtype
    my $Cartridge = DE_EPAGES::Order::API::Cartridge->new;

    if ($ShopMailType->alias eq 'CustomerOrderStatusReturnedOn') {
        ImportTranslation($Cartridge->xmlDirectory().'/Translation.ShopMailTypeOrderStatusReturnedOn',$ShopMailType->getSite()->pathString);
    }

    # add pdf documents for TaC and right of withdrawal as default for order status mails of Unity shops (UNITY-5067)
    my $hInfoShop = $ShopMailType->getSite()->get(['IsUnityRegistered', 'TermsAndConditions', 'CustomerInformation']);

    if ($hInfoShop->{'IsUnityRegistered'} && $ShopMailType->get('MailTypeTemplate')->get('MailTypeTemplateGroup')->alias eq 'OrderStatus') {
        $ShopMailType->set({
            'UnityAttachLegalDocumentsPDF' => [$hInfoShop->{'TermsAndConditions'}, $hInfoShop->{'CustomerInformation'}]
        });
    }

}

#========================================================================================
# function     OnSendTestMail
# state        public
#----------------------------------------------------------------------------------------
# syntax       OnSendTestMail($hParams);
#----------------------------------------------------------------------------------------
# description  Triggered by hook UI_OnSendTestMail.
#               Add withdrawal rights PDF as attachment in test emails.
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>ShopMailType - shop mailtype object</li>
#                   <li>Shop - shop object</li>
#                   <li>LanguageID - current language id</li>
#                   <li>MailParams -  mail params</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnSendTestMail {
    my ($hParams) = @_;

    my $LegalDocs = $hParams->{'Shop'}->isUnity()
        ? $hParams->{'ShopMailType'}->get('UnityAttachLegalDocumentsPDF')
        : $hParams->{'ShopMailType'}->get('AttachCustomerInformationPDF')
            ? [$hParams->{'Shop'}->get('CustomerInformation')]
            : undef;

    foreach my $LegalDocument (@$LegalDocs) {
        my $hLegalDocumentPDFAttachment
            = GetLegalDocumentPDFAttachment($hParams->{'Shop'},$hParams->{'LanguageID'}, $LegalDocument);
        push @{$hParams->{'MailParams'}->{'Attachments'}}, $hLegalDocumentPDFAttachment
            if $hLegalDocumentPDFAttachment;
    }

    return;
}

#========================================================================================
# function     ShowMBOTabs
# state        public
#----------------------------------------------------------------------------------------
# syntax       ShowMBOTabs($hParams);
#----------------------------------------------------------------------------------------
# description  Triggered by hook UI_ShowMBOTabs.
#               Switch the display of MBO-Tabs include at VIEWACTION_NO_TABS.
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used keys are
#               <ul>
#                   <li>ViewAction - ViewAction (string)</li>
#                   <li>ShowMBOTabs - indicator (boolean</li>
#                   <li>Shop - Shop (object)</li>
#               </ul> | ref.hash.*
#========================================================================================
sub ShowMBOTabs {
    my ($hParams) = @_;

    if ($hParams->{'ViewAction'} eq VIEWACTION_NO_TABS and $hParams->{'Shop'}->get('FeatureMaxValue')->{'OrderDocument'}) {
        $hParams->{'ShowMBOTabs'} = 0;
    }

    return;
}

1;
