#==============================================================================
# package      DE_EPAGES::Presentation::UI::Login
# state        public
# description  UI functions for user login
#==============================================================================
package DE_EPAGES::Presentation::UI::Login;
use base DE_EPAGES::Presentation::UI::Object;

use strict;

use DateTime;
use DateTime::Duration;
use HTTP::Status;

use DE_EPAGES::Core::API::Error                 qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::EpagesConfig          qw ( EpagesConfig );
use DE_EPAGES::Core::API::Log                   qw ( LogDebug GetLog );
use DE_EPAGES::Core::API::Password              qw ( CryptPassword CheckPassword );
use DE_EPAGES::Core::API::PerlTools             qw ( GetCurrentDateTime VerifySecureCode AddDuration );
use DE_EPAGES::Core::API::Scramble              qw ( Unscramble );
use DE_EPAGES::Core::API::TestParameter         qw ( TestObject );
use DE_EPAGES::Core::API::Url                   qw ( UrlEncodeValue);
use DE_EPAGES::Object::API::Factory             qw ( Transaction LoadRootObject );
use DE_EPAGES::Presentation::API::User          qw ( GetUsersByEMail GetVerfiedUsersByEMail );
use DE_EPAGES::Trigger::API::Trigger            qw ( TriggerHook );
use DE_EPAGES::WebInterface::API::Cookie        qw ( SetCookie );
use DE_EPAGES::Presentation::API::LoginChecker  qw ( NewLoginChecker ) ;

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'NoResults_LocaleID'  => 'Locale "#LocaleID" doesn\'t exist at site "#SitePath"',
        'SSLRequiredForLogin' => 'SSL required for login'
    };
}
#ERRORDEFINITION END

sub ViewSimpleLoginForm {
    my $self = shift;
    my ($Servlet) = @_;
    LogDebug( 'ViewLoginForm', $Servlet->params );

    my $TemplateDir = "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation/Templates";
    $Servlet->vars( 'TemplateFile', "$TemplateDir/SimpleLogin.html" );
    $Servlet->vars( 'LastObjectID'  , $Servlet->object->id );
    $Servlet->vars( 'LastViewAction', $Servlet->params('ViewAction') );
    $Servlet->vars( 'ViewAction'    , 'ViewLoginForm' );
}

sub ViewLoginForm {
    my $self = shift;
    my ($Servlet) = @_;
    LogDebug( 'ViewLoginForm', $Servlet->params );


    $Servlet->vars( 'LastObjectID'  , $Servlet->object->id );
    $Servlet->vars( 'LastViewAction', $Servlet->params('ViewAction') );

    # Check if MBO is called but no HTTPs is used; if SSL enabled in TBO, redirect
    # to a https connection (EPG-13656).
    my $System = LoadRootObject();
    my $AdminExtension = $System->get('URLAdminExtension');
    my $Extension = $Servlet->serverVariables('SCRIPT_NAME') =~ /\.(\w+)$/ ? $1 : '';

    if ( $Extension eq $AdminExtension
        && !$Servlet->serverVariables('SERVER_PORT_SECURE')
        && $System->get('HasSSLCertificate'))
    {
        my $Object = $Servlet->object;
        my $Site = $Object->getSite();

        my $LocaleID = $Servlet->locale;
        my $Location = $Site->get('WebServerScriptAdminUrlSSL') . "/$LocaleID/";
        my @Params;
        my $ViewAction = $Servlet->params('ViewAction');
        push @Params, 'ViewAction='.$ViewAction if $ViewAction ne 'View';
        push @Params, 'ObjectPath='.UrlEncodeValue($Object->get('Path'));
        $Location .= '?'.join('&', @Params) if scalar(@Params) > 0;

        $Servlet->redirect( $Location );
    }

    return;
}

sub SaveLoginForm {
    my $self = shift;
    my ($Servlet) = @_;
    LogDebug( 'SaveLoginForm', $Servlet->params );

    if (EpagesConfig()->get('Security','ForcePOSTLogin',0)) {
        $Servlet->forcePOSTForNonTestShops();
    }
    # EPG-32268: don't allow non-SSL logins if SSL is available
    $Servlet->testSecureConnection();

    my $RegistrationObjectID = $Servlet->params('RegistrationObjectID');

    my $Site = $Servlet->object->getSite();
    my $RegistrationObject;
    if (defined $RegistrationObjectID) {
        $RegistrationObject = $Site->loadObject( $RegistrationObjectID, 'Object' );
    }
    else {
        $RegistrationObject = $Site;
    }

    my $LoginForm = $Servlet->form->form($RegistrationObject, 'Login');

    my $LoginChecker = NewLoginChecker( $Servlet ) ;

    my $LoginCheckerResponse = $LoginChecker->startLoginAttempt() ;
    if ( $LoginCheckerResponse->{'CancelLogin'} ) {
        GetLog('LoginChecker')->info( 'LoginChecker deny with Reason: '.$LoginCheckerResponse->{'Reason'} );
        $Servlet->form->executeFormError({
            'Reason'=> $LoginCheckerResponse->{'Reason'},
            'Form'  => 'Login',
            'Name'  => 'Login',
        });
    }
    # search for users alias or users email with the given email address and use the
    # active user with the latest registration if more than one email address was found
    my $aUsers = [];
    foreach my $User (@{GetUsersByEMail($LoginForm->{'Login'},$RegistrationObject)}) {
        push @$aUsers, $User;
    }

    if (not scalar(@$aUsers)) {
        $LoginChecker->triggerLoginFailure('LOGIN_FAILED') ;
        $Servlet->form->addFormError({
            'Reason'=> 'LOGIN_FAILED',
            'Form'  => 'Login',
            'Name'  => 'Login',
            'RegistrationObject' => $RegistrationObject->pathString,
        });
        $Servlet->form->addFormError({
            'Reason'=> 'LOGIN_FAILED',
            'Form'  => 'Login',
            'Name'  => 'Password',
            'RegistrationObject' => $RegistrationObject->pathString,
        });
        $Servlet->form->executeFormError;
    }
    my $User = $aUsers->[0];

    $LoginChecker->setUser( $User ) ;

    # check activation
    if ( not $User->get('IsActivated') ) {
        $LoginChecker->triggerLoginFailure('LOGIN_INACTIVE') ;
        $Servlet->form->executeFormError({
            'Reason'=> $User->get('IsEmailVerified') ? 'LOGIN_INACTIVE' : 'LOGIN_NOT_VERIFIED',
            'Form'  => 'Login',
            'Name'  => 'Login',
            'RegistrationObject' => $RegistrationObject->pathString,
        });
    }
    # check LoginFailureCount and LoginFailureTimeLimit
    my $System = LoadRootObject();
    if (defined $System->get('LoginFailureTimeLimit')) {
        my $duration = DateTime::Duration->new( minutes => $System->get('LoginFailureTimeLimit') );
        my $checkTime = GetCurrentDateTime() - $duration;
        if ($User->get('LastLoginFailure') <= $checkTime) {
            $User->set({'CountLoginFailure' => 0});
        }
        if ($System->get('LoginFailureCount') <= $User->get('CountLoginFailure')) {
            $LoginChecker->triggerLoginFailure('LOGIN_LOCKED') ;
            $Servlet->form->executeFormError({
                'Reason'=> 'LOGIN_LOCKED',
                'Form'  => 'Login',
                'Name'  => 'Login',
            });
        }
    }

    # check the password
    my $UserPassword = $User->get('Password') // '';

    # only for unity:
    # we have the plain text password "epages" for csv imported users (the user is not able to login with "epages")
    # the user gets an info about the outdated password and should use the
    # forgot password functionality to create a new one
    if($Site->class->alias eq 'Shop' and $Site->get('IsUnityRegistered')){
        $Servlet->form->executeFormError({
            'Reason'=> 'PASSWD_OUTDATED',
            'Form'  => 'Login',
            'Name'  => 'Login',
        }) if $UserPassword eq 'epages';
    }

    eval {
        unless (CheckPassword($LoginForm->{'Password'}, $UserPassword)) {
            my $CountLoginFailure = (defined $User->get('CountLoginFailure')) ? $User->get('CountLoginFailure') : 0;
            $User->set({
                'CountLoginFailure' => $CountLoginFailure + 1,
                'LastLoginFailure'  => GetCurrentDateTime(),
            });
            $LoginChecker->triggerLoginFailure('LOGIN_FAILED') ;
            $Servlet->form->addFormError({
                'Reason'=> 'LOGIN_FAILED',
                'Form'  => 'Login',
                'Name'  => 'Login',
                'RegistrationObject' => $RegistrationObject->pathString,
            });
            $Servlet->form->addFormError({
                'Reason'=> 'LOGIN_FAILED',
                'Form'  => 'Login',
                'Name'  => 'Password',
                'RegistrationObject' => $RegistrationObject->pathString,
            });
            $Servlet->form->executeFormError;
        }
    };
    if (ExistsError()) {
        my $Error = GetError();
        Error($Error) unless $Error->code eq 'NativeDie' and $Error->shortMessage eq 'Wide character in crypt';
        $User->set({
            'CountLoginFailure' => $User->get('CountLoginFailure') + 1,
            'LastLoginFailure'  => GetCurrentDateTime(),
        });
        $LoginChecker->triggerLoginFailure('PASSWORD_MISMATCH') ;
        $Servlet->form->executeFormError({
            'Reason'=> 'PASSWORD_MISMATCH',
            'Form'  => 'Login',
            'Name'  => 'Password',
        }) ;
    }

    $User->set({'CountLoginFailure' => 0});
    $self->LoginUser($Servlet, $User);

    $LoginChecker->triggerLoginSuccessful() ;

    $Servlet->vars('ViewObjectID', $User->id) if $Servlet->form->value('UseUserObject');
}

#========================================================================================
# function     LoginUser
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->LoginUser( $Servlet, $User );
#----------------------------------------------------------------------------------------
# description  Registers the given user at the current session.
#               Sets the user's preferred locale and currency.
#               Sets the user GUID as cookie if login cookies are enabbled for the site
#               and for this user.
#               Deletes the AnonymousUser object if the user already had a basket.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet | object
# input        $User | user object | object
#========================================================================================
sub LoginUser {
    my $self = shift;
    my ($Servlet, $User) = @_;
    TestObject('User', $User);

    my $OldUser = $Servlet->user;
    my $IntSessionID = $Servlet->params( 'IntSessionID');
    my $OldIntSessionID = $IntSessionID;
    Transaction( sub {
        my $trigger = defined $IntSessionID ? $IntSessionID : 0;
        my %hSession = ( UserID => $User->id, IntSessionID => $IntSessionID );
        TriggerHook('API_BeforeUpdateSession', \%hSession ) if $trigger;
        $hSession{'OldIntSessionID'} = $IntSessionID;
        $IntSessionID = $Servlet->registerSession( $User );
        $hSession{'IntSessionID'} = $IntSessionID;
        TriggerHook('API_AfterUpdateSession', \%hSession ) if $trigger;

        if ( $User->get('IsLoginCookieAllowed') and $User->getSite()->get('LoginCookie') ) {
            my $GUID = $User->get('GUID');
            my $UserCookie = $Servlet->cookies('UserGUID');
            if (not defined $UserCookie or $UserCookie ne $GUID) {
                my %Options = (
                    'Max-Age'=> 60 * 60 * 24 * 365 * 5, ## no critic qw( ProhibitMagicNumbers ) = 5 years
                    'Path'   => $Servlet->params('SCRIPT_NAME'),
                    'Comment'=> $User->class->attribute('IsLoginCookieAllowed')->get('Description'),
                    'Expires'=> AddDuration(DateTime->now(), 'years', 5)->strftime('%a, %d-%b-%Y %I:%M:%S %Z'),
                    'HttpOnly' => 1,
                );
                SetCookie($Servlet->response, 'UserGUID', $GUID, \%Options);
            }
        }

        # remove anonymous user
        $OldUser->delete if defined $OldUser and $OldUser->get('IsAnonymous');
    } );

    TriggerHook('UI_AfterLoginUser', { 'Servlet' => $Servlet, 'OldIntSessionID' => $OldIntSessionID });
}

#========================================================================================
# function     Logout
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->Logout( $Servlet );
#----------------------------------------------------------------------------------------
# description  Unassigns the currently logged on user from the session.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet | object
#========================================================================================
sub Logout {
    my $self = shift;
    my ($Servlet) = @_;
    LogDebug( 'Logout', $Servlet->params );

    TriggerHook('UI_BeforeLogoutUser', { 'Servlet' => $Servlet } );

    $Servlet->logout;
    return;
}

#========================================================================================
# function     SaveUserPassword
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->SaveUserPassword( $Servlet, $User );
#----------------------------------------------------------------------------------------
# description  Saves user password. If no User is specified the current user saves his
#               password.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet | object
# input        $User | user (optional)| object
#========================================================================================
sub SavePassword {
    my $self = shift;
    my ($Servlet, $User) = @_;

    $Servlet->testSecureConnection();
    $Servlet->testPostRequest();

    $User = $Servlet->user unless defined $User;
    my $hValues = $Servlet->form->form($User, 'SavePassword');
    my $NewPassword ;

    if( defined $hValues->{'NewPassword'} && $hValues->{'NewPassword'} =~ /\S/ ) {
        $NewPassword = $hValues->{'NewPassword'} ;

        if (!$Servlet->form->value('SupportStaff')) {
            my $OldUserPassword = $User->get('Password');
            eval {
                unless (CheckPassword($hValues->{'OldPassword'}, $OldUserPassword)) {
                    $Servlet->form->addFormError({
                        'Reason'=> 'WRONG_OLD_PWD',
                        'Form'  => 'Save',
                        'Name'  => 'OldPassword',
                    }) ;
                }
            };
            if (ExistsError()) {
                my $Error = GetError();
                Error($Error) unless $Error->code eq 'NativeDie' and $Error->shortMessage eq 'Wide character in crypt';
                $Servlet->form->addFormError({
                    'Reason'=> 'WRONG_OLD_PWD',
                    'Form'  => 'Save',
                    'Name'  => 'OldPassword',
                }) ;
            }
        }
        if ($hValues->{'PasswordConf'} ne $NewPassword) {
            $Servlet->form->addFormError({
                'Reason'=> 'PWD_NOT_SAME',
                'Form'  => 'Save',
                'Name'  => 'NewPassword',
            }) ;
            $Servlet->form->addFormError({
                'Reason'=> 'PWD_NOT_SAME',
                'Form'  => 'Save',
                'Name'  => 'PasswordConf',
            }) ;
        }
        #crypt password if new password submitted
        eval { $hValues->{'Password'} = CryptPassword($NewPassword); };
        if (ExistsError()) {
            my $Error = GetError();
            Error($Error) unless $Error->code eq 'NativeDie' and $Error->shortMessage eq 'Wide character in crypt';
            $Servlet->form->addFormError({
                'Reason'=> 'PASSWORD_WRONG',
                'Form'  => 'Save',
                'Name'  => 'NewPassword',
            }) ;
        }
        TriggerHook(
            'UI_BeforeSavePassword',
            {
                'Servlet'           => $Servlet,
                'User'              => $User,
                'Password'          => $NewPassword,
                'FormName'          => 'Save',
                'FormField'         => 'NewPassword',
            }
        );
        $Servlet->form->executeFormError;
    }
    delete $hValues->{$_} foreach qw(NewPassword PasswordConf OldPassword);
    $User->set($hValues);
    if ( defined $NewPassword ) {
        TriggerHook(
            'UI_AfterSavePassword',
            {
                'Servlet'           => $Servlet,
                'User'              => $User,
                'Password'          => $NewPassword
            }
        );
    }

    # crypt password if new password submitted
    $Servlet->vars('PWD_CHANGED',1);

    $Servlet->logout(1);
    $self->LoginUser($Servlet, $User);
    $Servlet->params('SessionID', $Servlet->cookies('SessionID'));

    return;
}

#========================================================================================
# function     SaveUserSettings
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->SaveUserSettings( $Servlet );
#----------------------------------------------------------------------------------------
# description  Saves user preferences such as name, e-mail, locale and password.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet | object
#========================================================================================
sub SaveUserSettings {
    my $self = shift;
    my ($Servlet) = @_;

    # test all params before
    my $User = $Servlet->object;
    my $hValues = $Servlet->form->form($User, 'SaveSettings');

    # test if another user has the same login
    if( exists $hValues->{'Alias'} ) { # EPG-7799: in some projects, the input field "Alias" may be removed
        my $UserFolder = $User->get('Parent');
        if( $hValues->{'Alias'} ne $User->alias && $UserFolder->existsChild($hValues->{'Alias'}) ) {
            $Servlet->form->executeFormError({
                'Reason'=> 'USER_EXISTS',
                'Form'  => 'Form',
                'Name'  => 'Alias',
            });
        }
    }

    # save password
    $self->SavePassword( $Servlet, $User ) if defined $hValues->{'NewPassword'};

    # set locale and switch locale for the current session
    my $LocaleID = $hValues->{'LocaleID'};

    my $Site = $User->getSite();
    Error( 'NoResults_LocaleID', {
        LocaleID => $LocaleID,
        SitePath => $User->getSite()->get('Path')
    }) unless $Site->existsSiteLocale($LocaleID);

    my $Locale = $Site->siteLocale($LocaleID);

    # only set local when user changes his own settings
    if (!$Servlet->form->value('SupportStaff')) {
        $hValues->{'CurrencyID'} = $Locale->get('Locale')->{'CurrencyID'};
        $Servlet->setLocale($LocaleID);
        $Servlet->vars('CurrencyID', $Locale->get('Locale')->{'CurrencyID'} );
    }

    delete $hValues->{$_} foreach qw(NewPassword PasswordConf OldPassword SiteLocaleID);
    $User->set($hValues);
    return;
}

#========================================================================================
# function     ConfirmUserAccount
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->ConfirmUserAccount( $Servlet );
#----------------------------------------------------------------------------------------
# description  This confirms the user's account by verifying a security code.
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet | object
#========================================================================================
sub ConfirmUserAccount {
    my $self = shift;
    my ($Servlet) = @_;

    $Servlet->testSecureConnection();

    my $Site = $Servlet->object;

    my $Alias = Unscramble($Servlet->params('Alias'));
    my $Email = Unscramble($Servlet->params('EMail'));
    $Servlet->params('Alias', $Alias);
    $Servlet->params('EMail', $Email);

    my $SecurityCode = $Servlet->params('SecurityCode');
    Error('RequestManipulated') unless VerifySecureCode($Alias.$Email, $SecurityCode);

    my $Folder = $Site->folder('Users');

    if ($Folder->existsChild($Alias)) {
        my $User = $Folder->child($Alias);
        die 'Object ' . $User->id . ' is not a User' unless $User->instanceOf('User');

        if ($User->get('IsEmailVerified')) {
            # EPG-18125: already confirmed -> show login form
            $Servlet->vars('EmailAlreadyVerified', 1);
            Error('PermissionDenied');
        }

        # check if the email is now in use
        if (scalar @{GetVerfiedUsersByEMail($Email, $User->getSite)}) {
            $Servlet->vars('EmailAlreadyUsed', 1);
            return;
        }

        $User->set({
            'IsEmailVerified'   => 1,
            'IsActivated'       => 1,
            'EMail'             => $Email,
        });
        $User->get('BillingAddress')->set({'EMail' => $Email});

        $Servlet->vars('EMAIL_VERIFIED', 1);
        $self->LoginUser($Servlet, $User);

        TriggerHook('UI_ConfirmUserAccount', {
            'Servlet' => $Servlet,
            'User' => $User,
            'EMail' => $Email
        });
    }
    else {
        $Servlet->form->executeFormError({
            'Reason'=> 'UserNotFound',
            'Name'  => 'UserNotFound',
        }) ;
    }

    return;
}

1;