#========================================================================================
# package      DE_EPAGES::Product::API::ProductService
# state        public
#----------------------------------------------------------------------------------------
# description  Collects the products of a shop
#========================================================================================
package DE_EPAGES::Product::API::ProductService;

use DE_EPAGES::Core::API::PerlTools qw ( isValid trim );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadClassByAlias Transaction );
use DE_EPAGES::Object::API::Object::Locale;
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Shop::API::Resource::Formatter;
use DE_EPAGES::Shop::API::ShopTaxClass;
use DE_EPAGES::Product::API::Product qw ( TestUPCEAN );
use DE_EPAGES::Product::API::Resource::CustomAttributes;
use DE_EPAGES::Product::API::Resource::ProductSlideshow;
use DE_EPAGES::Product::API::Resource::ProductVariation;
use DE_EPAGES::Product::API::Resource::ProductLowestPrice;
use DE_EPAGES::Product::API::Resource::ProductCategories;
use DE_EPAGES::Product::DAL::Product qw ( DAL_GetUpdatedProducts );
use DE_EPAGES::Product::API::Tools qw ( MergeUpdatedProductsResults ConvertDateTimeForEpJ );
use DE_EPAGES::WebService::API::Util qw (ReadJSONBoolean);

use strict;



#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( );
#----------------------------------------------------------------------------------------
# description  Constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Product::API::ProductService
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

sub getAllUpdatedProducts {
    my $self = shift;
    my ($Shop, $AttributesUpdatedSince, $StockLevelUpdatedSince) = @_;

    my $ShopId = $Shop->id();
    my $aProductIDsWithChangedAttributes = [];
    my $aProductIDsWithChangedStocklevel = [];

    if(defined $AttributesUpdatedSince) {
        $aProductIDsWithChangedAttributes = $self->_DAL_GetUpdatedProducts($ShopId, 'RestApiResponseAttributes', $AttributesUpdatedSince);
    }

    if(defined $StockLevelUpdatedSince) {
        $aProductIDsWithChangedStocklevel = $self->_DAL_GetUpdatedProducts($ShopId, 'RestApiStockLevel', $StockLevelUpdatedSince);
    }

    my $aUpdatedProducts = MergeUpdatedProductsResults($aProductIDsWithChangedAttributes, $aProductIDsWithChangedStocklevel);

    my @UpdatedProducts = map { [$self->_LoadObject($_->[0]), ConvertDateTimeForEpJ($_->[1])] } @$aUpdatedProducts;

    return \@UpdatedProducts;
}



sub getProductDetailsData {
    my $self = shift;
    my ($Shop, $Product, $LocaleID, $CurrencyID) = @_;
    my $Context = $self->_getShopContext($Shop);
    my $APIBaseURL = $Context->apiBaseUrl();
    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    my $hShopData = $self->getShopData($Shop, $LocaleID, $CurrencyID);
    my $hProductData = $self->getProductData($Product, $hShopData);

    my $aLinks = $hProductData->{'links'};
    $hProductData->{'links'} = $self->_addProductDetailsLinks($Product, $APIBaseURL, $LanguageID, $aLinks);
    $hProductData->{'shippingMethodsRestrictedTo'} = $self->_getShippingMethods($Product->get('ShippingMethods'), $APIBaseURL, $LanguageID),

    my $aCategories = $Product->get('AllProductCategories');
    my $MainCategory = $aCategories->[0];
    if (defined $MainCategory ) {
        push @{$hProductData->{'links'}}, {
            'rel' => 'main-category',
            'href' => $self->_getCategoryHref($APIBaseURL, $MainCategory->guid()),
            'title' => $MainCategory->get( 'NavigationString', $LanguageID ),
        };
    }

    return $hProductData;
}

sub getShopData {
    my $self = shift;
    my ($Shop, $LocaleID, $CurrencyID) = @_;

    my $Context = $self->_getShopContext($Shop);
    my $Locale = $self->_loadLocale($LocaleID);

    my $hShopData = {
        'Formatter' => $self->_createFormatter($Shop, $LocaleID),
        'Context' => $self->_getShopContext($Shop),
        'TaxModel' => $Shop->get('TaxModel'),
        'Locale' => $Locale,
        'LanguageID' => $Locale->get('LanguageID'),
        'SiteLocale' => $Shop->siteLocale($LocaleID),
        'APIBaseURL' => $Context->apiBaseUrl(),
        'ImageBaseURL' => $Context->imageBaseUrl(),
        'CurrencyID' => $CurrencyID,
        'TaxMatrix' => $Shop->get('TaxMatrix'),
    };
    return $hShopData;
}

sub getProductData {
    my $self = shift;
    my ($Product, $hNonSpecificProductData) = @_;

    my $Formatter = $hNonSpecificProductData->{'Formatter'};
    my $TaxModel = $hNonSpecificProductData->{'TaxModel'};
    my $LanguageID = $hNonSpecificProductData->{'LanguageID'};
    my $SiteLocale = $hNonSpecificProductData->{'SiteLocale'};
    my $APIBaseURL = $hNonSpecificProductData->{'APIBaseURL'};
    my $ImageBaseURL = $hNonSpecificProductData->{'ImageBaseURL'};
    my $CurrencyID = $hNonSpecificProductData->{'CurrencyID'};
    my $TaxMatrix = $hNonSpecificProductData->{'TaxMatrix'};

    my @Attributes = qw(
        IsVisible
        AvailabilityText
        AvailabilityIcon
        DeliveryPeriod
        Description
        DisplayTitle
        IsSpecialOffer
        Weight WeightUnit
        Text NameOrAlias
        ImageHotDeal ImageMedium ImageLarge ImageSmall ImageThumbnail
        Path GUID
        PriceQuantity OrderUnit
        RefUnit RefAmount RefContentAmount
        ShippingMethods
        WebPath
        WebUrl
        EnergyLabelsString
        EnergyLabelHighestEfficiencyClass
        EnergyLabelSourceFile
        ProductDataSheet
        Alias
        UPCEAN
        Manufacturer ManufacturerSKU
        MainCharacteristics
        TaxClassDigital
        TaxClass
        Keywords
        StockLevel StockLevelAlert
        IsMaster
        SuperProduct
        VariationTypeValues
        Length Width Height
        YoutubeVideoId
    );

    my $hProduct = $Product->getWithParent(\@Attributes, $LanguageID);
    $ImageBaseURL = $ImageBaseURL . $hProduct->{'WebPath'};

    my $ListPrice = $Product->getListPrice($CurrencyID, $TaxModel);
    my $ManufacturerPrice = $Product->getManufacturerPrice($CurrencyID, $TaxModel);
    my $EcoParticipation = $Product->getEcoParticipation($CurrencyID, $TaxModel);
    my $Deposit = $Product->getDeposit($CurrencyID, $TaxModel);
    my $PriceWithDeposits = defined $ListPrice ? $ListPrice + ($EcoParticipation // 0) + ($Deposit // 0) : undef;
    my $ReferencePrice = $Product->getReferencePrice($CurrencyID, $ListPrice);
    my $LowestPrice = $Product->lowestSubProductPrice($CurrencyID, $TaxModel);

    my $RefPriceFormatted;
    my $RefUnit = $hProduct->{'RefUnit'};
    if( defined $ReferencePrice && defined $RefUnit ) {
        $RefPriceFormatted = sprintf '%d %s = %s',
            $hProduct->{'RefAmount'},
            $RefUnit->get('Abbreviation') // $RefUnit->get('NameOrAlias', $LanguageID),
            $SiteLocale->formatMoney($ReferencePrice, $CurrencyID);
    }

    my $hQuantity = $Formatter->quantity( $hProduct->{'PriceQuantity'}, $hProduct->{'OrderUnit'} );
    my $hAmount   = $Formatter->quantity( $hProduct->{'RefContentAmount'}, $hProduct->{'RefUnit'} );


    my $hTaxData = {
        'LanguageID' => $LanguageID,
        'TaxClassDigital' => $hProduct->{'TaxClassDigital'},
        'TaxClass' => $hProduct->{'TaxClass'},
        'TaxMatrix' => $TaxMatrix,
    };

    my $aSearchKeywords = $self->_createSearchKeywords($hProduct->{'Keywords'});

    my $ProductVariationType = 'regular';
    my $aProductVariationSelection = undef;
    if($hProduct->{'IsMaster'}){
        $ProductVariationType = 'master';
    }elsif($hProduct->{'SuperProduct'}){
        $ProductVariationType = 'variation';
        $aProductVariationSelection = $self->_getVariationAttributes($Product, $hProduct->{'VariationTypeValues'});
    }

    # this is common information for all types of products, including master and variations
    my $hResult = {
        'productVariationType' => $ProductVariationType,
        'productVariationSelection' => $aProductVariationSelection,
        'visible' => $Formatter->bool($hProduct->{'IsVisible'}),
        'availabilityText' => $hProduct->{'AvailabilityText'},
        'availability' => $self->_getAvailability($hProduct->{'AvailabilityIcon'}),
        'deliveryWeight' => $Formatter->quantity( $hProduct->{'Weight'}, $hProduct->{'WeightUnit'} ),
        'deliveryPeriod' => $hProduct->{'DeliveryPeriod'},
        'description' => $hProduct->{'Text'},
        'title' => $hProduct->{'DisplayTitle'},
        'forSale' => $Formatter->bool( $Product->isMaster ? 0 : $Product->canAddToBasket ),
        'productImage' => $hProduct->{'ImageLarge'},
        'images' => $self->_getImages($hProduct, $ImageBaseURL),
        'specialOffer' => $Formatter->bool($hProduct->{'IsSpecialOffer'}), # TODO: move to Catalog cartridge
        'name' => $hProduct->{'NameOrAlias'},
        'priceInfo' => {
            'basePrice' => !defined($RefPriceFormatted) ? undef : {
                'formatted' => $RefPriceFormatted,
                'quantity' => $hAmount,
                'refPrice' => $Formatter->price( $ReferencePrice, $CurrencyID, $TaxModel ),
                'refQuantity' => $Formatter->quantity( $hProduct->{'RefAmount'}, $hProduct->{'RefUnit'} ),
            },
            'depositPrice' => $Formatter->price( $Deposit, $CurrencyID, $TaxModel ),
            'ecoParticipationPrice' => $Formatter->price( $EcoParticipation, $CurrencyID, $TaxModel ),
            'manufacturerPrice' => $Formatter->price( $ManufacturerPrice, $CurrencyID, $TaxModel ),
            'price' => $Formatter->price( $ListPrice, $CurrencyID, $TaxModel ),
            'priceWithDeposits' => $Formatter->price( $PriceWithDeposits, $CurrencyID, $TaxModel ),
            'lowestPrice' => $Formatter->price( $LowestPrice, $CurrencyID, $TaxModel ),
            'quantity' => $hQuantity,
            'taxClass' => $self->_calculateTaxClass($hTaxData),
        },
        'productId' => $hProduct->{'GUID'},
        'sfUrl'     => $hProduct->{'WebUrl'},
        'shortDescription' => $hProduct->{'Description'},
        'energyLabelsString' => $hProduct->{'EnergyLabelsString'},
        'energyLabelsMaxEfficiencyString' => $hProduct->{'EnergyLabelHighestEfficiencyClass'},
        'energyLabelSourceFile' => $hProduct->{'EnergyLabelSourceFile'} ? $ImageBaseURL .'/'. $hProduct->{'EnergyLabelSourceFile'} : undef,
        'productDataSheet' => $hProduct->{'ProductDataSheet'} ? $ImageBaseURL .'/'. $hProduct->{'ProductDataSheet'} : undef,
        'productNumber' => $hProduct->{'Alias'},
        'upcEan' => $hProduct->{'UPCEAN'},
        'manufacturer' => $hProduct->{'Manufacturer'},
        'manufacturerProductNumber' => $hProduct->{'ManufacturerSKU'},
        'essentialFeatures' => $hProduct->{'MainCharacteristics'},
        'searchKeywords' => $aSearchKeywords,
        'stocklevel' => $hProduct->{'StockLevel'},
        'minStocklevel' => $hProduct->{'StockLevelAlert'},
        'productLength' => $hProduct->{'Length'},
        'productHeight' => $hProduct->{'Height'},
        'productWidth' => $hProduct->{'Width'},
        'links' => $self->_createLinks($Product, $APIBaseURL, $LanguageID),
        'videos' => [],
    };

    # add videos in an array of hashes for upcomming changes (maybe we have more than one video and platform)
    # so we have to change only this resource and the perl backend and not the REST part in epagesj
    if($hProduct->{'YoutubeVideoId'}){
        push @{$hResult->{'videos'}}, {'type' => 'YouTube', 'source' => $hProduct->{'YoutubeVideoId'}};
    }

    return $hResult;
}

sub _getVariationAttributes{
    my $self = shift;
    my ($Product, $aVariationTypeValues) = @_;
    if (scalar @$aVariationTypeValues == 0){
        return; # this is not a master product
    }
    my $aVariationAttributeNames = [map { $_->{'Attribute'}->alias } @{$aVariationTypeValues}];
    my $hVariationAttributeValues = $Product->getWithParent($aVariationAttributeNames);
    return [
            map {
                    {
                        'name' => $_,
                        'value' => $hVariationAttributeValues->{$_}->alias()
                    }
            } @{$aVariationAttributeNames}
        ];
}


sub _calculateTaxClass {
    my $self = shift;
    my ($hTaxData) = @_;
    my $TaxClassDigital = $hTaxData->{'TaxClassDigital'};

    if(defined ($TaxClassDigital)) {
        return $self->_calculateTaxClassDigital($TaxClassDigital, $hTaxData->{'LanguageID'});
    } else {
        return $self->_calculateTaxClassNormal($hTaxData);
    }
}


sub _calculateTaxClassDigital {
    my $self = shift;

    my ($TaxClassDigital, $LanguageID) = @_;
    return {
        'taxClassId' => $TaxClassDigital->guid(),
        'name' => $TaxClassDigital->get('NameOrAlias', $LanguageID),
        'percentage'=> undef,
    };
}


sub _calculateTaxClassNormal {
    my $self = shift;
    my ($hTaxData) = @_;
    my $TaxClass = $hTaxData->{'TaxClass'};
    my $TaxMatrix = $hTaxData->{'TaxMatrix'};

    return unless defined ($TaxClass) && defined ($TaxMatrix);

    my $TaxArea = $TaxMatrix->defaultTaxArea();
    return {
        'taxClassId' => $TaxClass->guid(),
        'name' => $TaxClass->get('NameOrAlias', $hTaxData->{'LanguageID'}),
        'percentage'=> $TaxMatrix->taxRate($TaxArea, $TaxClass) * 100, ## no critic (ProhibitMagicNumbers)
    };
}


sub createProduct {
    my $self = shift;
    my ($Shop, $LocaleID, $CurrencyID, $hRequestBody) = @_;

    my $Locale = $self->_loadLocale($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');
    my $TaxModel = $Shop->get('TaxModel');

    my $Alias = $hRequestBody->{'productNumber'};
    if(!isValid($Alias)) {
        Error('BadRequest', {
            'Reason' => 'The attribute productNumber is mandatory!',
            'ShopAlias' => $Shop->alias(),
        });
    }

    my $ProductFolderClass = $self->_LoadClassByAlias('ProductFolder');
    my $ProductFolder = $Shop->folder('Products', { 'Class' => $ProductFolderClass } );
    my $ProductClass = $Shop->get('ProductClass');

    my $Product;
    eval {
        $self->_Transaction(
            sub {

                $Product = $ProductClass->insertObject( {
                    'Alias' => $Alias,
                    'Parent' => $ProductFolder,
                });

                my $aSearchKeywords = $hRequestBody->{'searchKeywords'} // [];
                my $Keywords = join(', ', @$aSearchKeywords);

                $Product->set({
                    'Name' => $hRequestBody->{'name'},
                    'Text' => $hRequestBody->{'description'},
                    'DeliveryPeriod' => $hRequestBody->{'deliveryPeriod'},
                    'Description' => $hRequestBody->{'shortDescription'},
                    'Manufacturer' => $hRequestBody->{'manufacturer'},
                    'MainCharacteristics' => $hRequestBody->{'essentialFeatures'},
                    'IsVisible' => ReadJSONBoolean($hRequestBody->{'visible'}),
                    'Keywords' => $Keywords,
                    'StockLevel' => $hRequestBody->{'stocklevel'},
                    'ManufacturerSKU' => $hRequestBody->{'manufacturerProductNumber'},
                    'Length' => $hRequestBody->{'productLength'},
                    'Width' => $hRequestBody->{'productWidth'},
                    'Height' => $hRequestBody->{'productHeight'},
                    'StockLevelAlert' => $hRequestBody->{'minStocklevel'},
                }, $LanguageID);

                if(isValid($hRequestBody->{'upc'})) {
                    TestUPCEAN('Value', $hRequestBody->{'upc'});
                    $Product->set({'UPCEAN' => $hRequestBody->{'upc'}});
                }
                if(isValid($hRequestBody->{'ean'})) {
                    TestUPCEAN('Value', $hRequestBody->{'ean'});
                    $Product->set({'UPCEAN' => $hRequestBody->{'ean'}});
                }
                if(isValid($hRequestBody->{'taxClassId'})) {
                    my $TaxClass = $self->_loadTaxClass($hRequestBody->{'taxClassId'}, $Shop);
                    if(defined $TaxClass) {
                        $self->_setTaxClass($Product, $TaxClass);
                    } else {
                        Error('TaxClass could not be found!');
                    }
                }

                $Product->setListPrice($CurrencyID, $TaxModel, $hRequestBody->{'price'});
                $Product->setManufacturerPrice($CurrencyID, $TaxModel, $hRequestBody->{'manufacturerPrice'});
                $Product->setDeposit($CurrencyID, $TaxModel, $hRequestBody->{'depositPrice'});
            }
        );
    };
    if (ExistsError()) {
        my $Error = GetError();
        if ( $Error->code() eq 'OBJECT_EXISTS' ) {
            Error('BadRequest', {
                'Reason' => 'Product with given alias exists already!',
                'ShopAlias' => $Shop->alias(),
            });
        } elsif ( $Error->code() eq 'FORMAT_NOT_UPCEAN' ) {
            Error('BadRequest', {
                'Reason' => 'UPC or EAN is not valid!',
                'ShopAlias' => $Shop->alias(),
            });
        } else {
            Error($Error);
        }
    }

    my $hShopData = $self->getShopData($Shop, $LocaleID, $CurrencyID);
    return $self->getProductData($Product, $hShopData);
}


sub _addProductDetailsLinks {
    my $self = shift;
    my ($Product, $APIBaseURL, $LanguageID, $Links) = @_;

    # now add specific information that is not available for all products

    my @Links = @{$Links};
    if (defined $Product->getWithParent('ImagesSlideShowString')) {
        push @Links, {
            'rel' => 'slideshow',
            'href' => DE_EPAGES::Product::API::Resource::ProductSlideshow->href($APIBaseURL, $Product->guid()),
        };
    }

    if (scalar grep { $_->{'Attribute'}->get('IsVisible') && $_->{'Attribute'}->type ne 'PreDefCustomerString' } @{$Product->get('UserDefinedAttributeValues', $LanguageID)}) {
        push @Links, {
            'rel' => 'custom-attributes',
            'href' => DE_EPAGES::Product::API::Resource::CustomAttributes->href($APIBaseURL, $Product->guid),
        };
    }

    return \@Links;
}


sub _createLinks {
    my $self = shift;
    my ($Product, $APIBaseURL, $LanguageID) = @_;

    my @Links = ();

    # add a link to self.
    push @Links, {
        'rel' => 'self',
        'href' => $self->_buildHref($APIBaseURL, $Product->guid),
    };

    # now add specific information that is not available for all products

    my $SuperProduct = $Product->superProduct();
    if( $SuperProduct ) {
        # this is a sub product - add a link to variations selection.
        push @Links, {
            'rel' => 'variations',
            'href' => DE_EPAGES::Product::API::Resource::ProductVariation->href($APIBaseURL, $SuperProduct->guid),
        };

        # add a link to master product.
        push @Links, {
            'rel' => 'master',
            'href' => $self->_buildHref($APIBaseURL, $SuperProduct->guid),
        };
    }
    elsif ($Product->isMaster) {
        # add a link to variations selection
        push @Links, {
            'rel' => 'variations',
            'href' => DE_EPAGES::Product::API::Resource::ProductVariation->href($APIBaseURL, $Product->guid()),
        };

        # add a link to the lowest price
        if ($Product->existSubProductPrices) {
            push @Links, {
                'rel' => 'lowest-price',
                'href' => DE_EPAGES::Product::API::Resource::ProductLowestPrice->href($APIBaseURL, $Product->guid()),
            };
        }
    }
    # add a link to categories.
    push @Links, {
        'rel' => 'categories',
        'href' => DE_EPAGES::Product::API::Resource::ProductCategories->href($APIBaseURL, $Product->guid),
    };

    # add a link to crossselling products.
    # use cross-selling products of the master product for variations (UNITY-4850)
    if(scalar(@{$Product->getWithParent('CrossSelling')})){
        push @Links, {
            'rel' => 'cross-selling',
            'href' => DE_EPAGES::Product::API::Resource::Product->hrefToProductCrossSelling($APIBaseURL, $Product->guid),
        };
    }

    return \@Links;
}



sub _loadTaxClass {
    my $self = shift;
    my ($GUID, $Shop) = @_;

    my $ShopTaxClass = $self->_createShopTaxClass($Shop);
    my $TaxClass = $ShopTaxClass->loadTaxClassbyGUID($GUID);

    return $TaxClass;
}



sub _setTaxClass {
    my $self = shift;
    my ($Product, $TaxClass) = @_;

    if($TaxClass->instanceOf('TaxClass') || $TaxClass->instanceOf('TaxClassUS')) {
        $Product->set({'TaxClass' => $TaxClass});

        # Note: if there is a TaxClassDigital, the TaxClass of a product will be ignored!
        $Product->set({'TaxClassDigital' => undef});
    }

    if($TaxClass->instanceOf('TaxClassDigital')) {
        $Product->set({'TaxClassDigital' => $TaxClass});
    }

    return;
}



sub _getAvailability {
    my $self = shift;
    my ($AvailabilityIcon) = @_;

    return undef unless (defined $AvailabilityIcon);
    $AvailabilityIcon =~ s/^Product//;
    $AvailabilityIcon =~ s/Icon$//;
    return $AvailabilityIcon;
}



sub _getShippingMethods {
    my $self = shift;
    my ($aShippingMethods, $APIBaseURL, $LanguageID) = @_;

    # empty list is equal to no restriction. This should be reflected as null in the REST API
    if(!scalar @$aShippingMethods) {
        return undef;
    }
    return [
        map {
            {
                href => "$APIBaseURL/api/shipping-methods/".$_->guid,
                rel => 'related',
                title => $_->get('NameOrAlias', $LanguageID)
            }
        }
        @$aShippingMethods
    ];
}



sub _createSearchKeywords {
    my $self = shift;
    my ($KeywordString) = @_;

    return [] unless defined $KeywordString;

    my @SplittedKeywords = split(/,/, $KeywordString); # split by ','
    my @TrimmedKeywords = map {trim($_)} @SplittedKeywords; # remove leading and trailing spaces
    my @Keywords = grep {length($_) > 0} @TrimmedKeywords; # remove empty strings
    my $aKeywords = [@Keywords];

    return $aKeywords;
}



sub _getImages {
    my $self = shift;
    my ($hProduct, $BaseURL) = @_;

    return [
        map { $self->_getImage($_, $hProduct->{$_}, $BaseURL) }
        grep { defined $hProduct->{$_} }
        qw ( ImageHotDeal ImageMedium ImageLarge ImageSmall ImageThumbnail )
    ];
}



sub _getImage {
    my $self = shift;
    my ($Classifier, $Value, $BaseUrl) = @_;

    # Remove 'Image'-prefix (ImageHotdeal -> Hotdeal)
    $Classifier =~ s/^Image//;

    return {
        'classifier' => $Classifier,
        'url' => $BaseUrl . '/' . $Value,
    };
}



sub _DAL_GetUpdatedProducts {
    my $self = shift;
    my ($ShopId, $AttributeSet, $UpdatedAfter) = @_;
    return DAL_GetUpdatedProducts($ShopId, $AttributeSet, $UpdatedAfter);
}



sub _LoadObject {
    my $self = shift;
    my ($ObjectID) = @_;
    print $ObjectID."\n";
    return LoadObject($ObjectID);
}



sub _createFormatter {
    my $self = shift;
    my ($Shop, $LocaleID) = @_;
    return DE_EPAGES::Shop::API::Resource::Formatter->new($Shop, $LocaleID);
}



sub _buildHref {
    my $self = shift;
    my ($APIBaseURL, $ProductGUID) = @_;
    return "$APIBaseURL/api/products/$ProductGUID";
}



sub _loadLocale {
    my $self = shift;
    my ($LocaleID) = @_;
    return DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
}



sub _getShopContext {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);
}



sub _getCategoryHref {
    my $self = shift;
    my ( $APIBaseURL, $GUID ) = @_;
    return "$APIBaseURL/api/categories/$GUID";
}



#========================================================================================
# function     Function
# state        public
#----------------------------------------------------------------------------------------
# syntax       Function( $Product );
#----------------------------------------------------------------------------------------
# description  Deletes the given product.
#
#               If the product IS a sub-product, only the the sub-product will be deleted,
#               the main product automatically does not contain this sub-product anymore
#               without explicitly removing it from the main product as well.
#
#               If the product HAS some sub-products, they will be deleted as well.
#
#               All mappings between the product and shippings will be deleted as well.
#
#               The RemoteSearch cartridge implememts the hook OnDeleteProduct and
#               updates the search index automatically when deleting the product.
#----------------------------------------------------------------------------------------
# input        $Product | A product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub deleteProduct {
    my $self = shift;
    my ($Product) = @_;

    $Product->delete();

    return;
}



sub _LoadClassByAlias {
    my $self = shift;
    my ($Alias) = @_;
    return LoadClassByAlias($Alias);
}



sub _Transaction {
    my $self = shift;
    my ($cCode, $cRollback) = @_;
    Transaction($cCode, $cRollback);
    return;
}



sub _createShopTaxClass {
    my $self = shift;
    my ($Shop) = @_;
    return DE_EPAGES::Shop::API::ShopTaxClass->new($Shop);
}

1;
