#========================================================================================
# package      DE_EPAGES::Search::API::Object::PagerSearch
# state        public
# base         DE_EPAGES::Presentation::API::Object::Pager
#----------------------------------------------------------------------------------------
# description  object interface for Search.
#----------------------------------------------------------------------------------------
# example      # without Servlet
#               my $ProductClass = LoadClassByAlias('Product');
#               my $SearchDefinition = GetSearchDefinition($ProductClass, 'MBO-AllStrings');
#               my $Search = DE_EPAGES::Search::API::Object::PagerSearch->new({
#                   'SearchDefinition' => $SearchDefinition,
#                   'ObjectID' => $Shop->id,
#                   'OrderBy' => 'Alias',
#                   'Parameters' => {
#                       'Shop.ID' => $Shop->id,
#                       'SearchLanguageID' => $LanguageID,
#                       'NameAttributeID' => $ProductClass->attribute('Name')->id,
#                       'SearchString' => $SearchString,
#                   },
#               });
#               my $aProducts = $Search->allItems;
# example      # with Servlet
#               my $ProductClass = LoadClassByAlias('Product');
#               my $SearchDefinition = GetSearchDefinition($ProductClass, 'MBO-AllStrings');
#               my $Search = DE_EPAGES::Search::API::Object::PagerSearch->new({
#                   'SearchDefinition' => $SearchDefinition,
#                   'Servlet' => $Servlet,
#                   'OrderBy' => 'Alias',
#                   'Parameters' => {
#                       'Shop.ID' => $Shop->id,
#                       'SearchLanguageID' => $LanguageID,
#                       'NameAttributeID' => $ProductClass->attribute('Name')->id,
#                       'SearchString' => $SearchString,
#                   },
#                   'UriParameters'  => {
#                       'Search' => 'MBO-AllStrings',
#                   },
#               });
#               $Servlet->vars('Pager', $Search->page->{'Pager'});
#========================================================================================
package DE_EPAGES::Search::API::Object::PagerSearch;
use base qw( DE_EPAGES::Presentation::API::Object::Pager Exporter);
use strict;

our @EXPORT_OK = qw (
    ReplaceWildcards
    ReplaceWildcardsWithoutPosOne
);

use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Core::API::Log qw (LogDebug);
use DE_EPAGES::Core::API::TestParameter qw ( TestInteger );
use DE_EPAGES::Database::API::Connection qw (GetCurrentDBHandle);
use DE_EPAGES::Search::DAL::PagerSearch qw (DAL_ExecutePagerSearch DAL_GetSQLSelect DAL_GetSQLSimpleSelect
                                            DAL_GetSQLLimit DAL_GetSQLGroup DAL_GetSQLSort DAL_GetSQLSortMax);

sub ErrorDefinition {
    return {
        'NoMaster' => 'Search #SearchDefinition at class #ClassPath hasn\'t a defined master table.',
        'NoSort'   => 'Search #SearchDefinition at class #ClassPath hasn\'t a sort definition or sort by \'#OrderBy\' isn\'t available.',
    }
}

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Pager = $Package->new($hInfo);
#----------------------------------------------------------------------------------------
# description  Creates a new search pager object.
#----------------------------------------------------------------------------------------
# input        $hInfo | parameters with used keys
#                       <ul>
#                         <li>SearchDefinition - search definition - object</li>
#                         <li>Servlet - (optional) servlet - object</li>
#                         <li>Page - (optional) page number. Set this value to 0 to
#                               query all results, not only one page.
#                               If $hInfo->{Servlet} is defined, the default value is 1,
#                               otherwise the default value is 0. - integer</li>
#                       </ul>
#                       See <function DE_EPAGES::Presentation::API::Object::Pager::new>
#                       for additional parameters | ref.hash
# return       $Pager | pager |object
#========================================================================================
sub new  {
    my $class = shift;
    my ($hInfo) = @_;

    $hInfo->{'Page'} //= $hInfo->{'Servlet'} ? 1 : 0;

    my $self = $class->SUPER::new($hInfo);
    $self->statistic;

    return $self;
}

#========================================================================================
# function     statistic
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Pager->statistic;
#----------------------------------------------------------------------------------------
# description  Calls <function countPages> and <function countItems> to collect required
#               information for all pages.
#========================================================================================
sub statistic {
    my $self = shift;

    return if $self->get('StatisticFinished');

    $self->execute( $self->get('Page') );

    $self->SUPER::statistic;
    return;
}

#========================================================================================
# function     execute
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = $Package->execute($PageNumber);
#----------------------------------------------------------------------------------------
# description  executes the search.
#----------------------------------------------------------------------------------------
# input        $PageNumber | 1-based page number.
#                       Use 0 to query all rows, not only one page | integer
#========================================================================================
sub execute {
    my $self = shift;
    my ($PageNumber) = @_;

    $PageNumber = $PageNumber // 1;

    my $SQL = $self->_getSqlQuery($PageNumber);
    
    LogDebug('PAGERSQL', $SQL);
    LogDebug('Parameter', $self->get('ParameterValues'));
    LogDebug('CurrentPage', $PageNumber);

    my $hValues = DAL_ExecutePagerSearch($PageNumber, $SQL, $self->get('ParameterValues'));
    $self->{$_} = $hValues->{$_} foreach (keys %$hValues); # sets attributes CurrentPage, Items and ItemCount

    return;
}

#========================================================================================
# function     ReplaceWildcards
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $SQLString = ReplaceWildcards($SearchString);
#----------------------------------------------------------------------------------------
# description  replaces * with % and adds on begin and end a %.
# input        $SearchString | search string can include * | string
# return       $SQLString | string for sql statement | string
#========================================================================================
sub ReplaceWildcards {
    my ($String) = @_;
    my $dbi = GetCurrentDBHandle();
    local $_ = $dbi->replaceLikeWildcards($String);
    s/\*/%/g;
    return '%'.$_.'%';
}

#========================================================================================
# function     ReplaceWildcardsWithoutPosOne
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $SQLString = ReplaceWildcardsWithoutPosOne($SearchString);
#----------------------------------------------------------------------------------------
# description  replaces * with % and adds on end a %.
# input        $SearchString | search string can include * | string
# return       $SQLString | string for sql statement | string
#========================================================================================
sub ReplaceWildcardsWithoutPosOne {
    my ($String) = @_;
    my $dbi = GetCurrentDBHandle();
    local $_ = $dbi->replaceLikeWildcards($String);
    s/\*/%/g;
    return $_.'%';
}

#========================================================================================
# function     allItems
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aItems = $Pager->allItems;
#----------------------------------------------------------------------------------------
# description  Returns all objects of all pages.
#----------------------------------------------------------------------------------------
# return       $aItems | list of objects | ref.array.object
#========================================================================================
sub allItems {
    my $self = shift;

    return $self->items(0);
}

#========================================================================================
# function     allItemIDs
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aItemIDs = $Pager->allItemIDs;
#----------------------------------------------------------------------------------------
# description  Returns the object id of all found objects of all pages.
#----------------------------------------------------------------------------------------
# return       $allItemIDs | list of object ids | ref.array.integer
#========================================================================================
sub allItemIDs {
    my $self = shift;

    return $self->itemIDs(0);
}

#========================================================================================
# function     masterFrom
#----------------------------------------------------------------------------------------
# state        private
# syntax       my $MasterFrom = $Object->masterFrom;
#----------------------------------------------------------------------------------------
# description  returns the from part of the master table
#----------------------------------------------------------------------------------------
# return       $MasterFrom | search from of search definition result class | hash
#========================================================================================
sub masterFrom {
    my $self = shift;
    my $Froms = $self->get('SearchDefinition')->get('ResultClass')->get('SearchFroms');
    my ($Master) = grep { $_->{'IsMaster'} } @$Froms;
    return $Master;
}

#========================================================================================
# function     sqlFrom
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $From = $Object->sqlFrom;
#----------------------------------------------------------------------------------------
# description  returns the from part of sql statement
#----------------------------------------------------------------------------------------
# return       $From | part of sql statement | string
#========================================================================================
sub sqlFrom {
    my $self = shift;
    my $SearchDefinition = $self->get('SearchDefinition');
    my $From = $SearchDefinition->get('FromPart');
    my $hSearchParameters = $self->get('Parameters');
    my @ParameterValues = ();
    $From =~ s/\#([\w\.]+)/_ReplaceParameter($1, $hSearchParameters, \@ParameterValues)/ge;
    $self->set('ParameterValues', \@ParameterValues);
    return $From;
}

#========================================================================================
# function     sqlWhere
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $Where = $Object->sqlWhere;
#----------------------------------------------------------------------------------------
# description  returns the where part of sql statement
#----------------------------------------------------------------------------------------
# return       $Where | part of sql statement | string
#========================================================================================
sub sqlWhere {
    my $self = shift;
    my $SearchDefinition = $self->get('SearchDefinition');
    my @Wheres = ();
    push @Wheres , $SearchDefinition->get('WherePart') if defined $SearchDefinition->get('WherePart');

    my $aDefinitionParameters = $SearchDefinition->get('Parameters');
    my $hSearchParameters = $self->get('Parameters');
    foreach my $DefinitionParameter (sort {$a->{'Position'} <=> $b->{'Position'}} @$aDefinitionParameters) {
        if (exists $hSearchParameters->{$DefinitionParameter->{'Parameter'}}) {
            push @Wheres , $DefinitionParameter->{'WherePart'};
        }
    }
    return '' if 0 == scalar @Wheres;

    my $Where = "\nWHERE " . join "\n  AND ", map { '('.$_.')' } @Wheres;
    my @ParameterValues = ();
    $Where =~ s/\#([\w\.]+)/_ReplaceParameter($1, $hSearchParameters, \@ParameterValues)/ge;
    my $esc = GetCurrentDBHandle()->escapeFromExtension;
    my $collation = GetCurrentDBHandle()->getSearchCollation;
    $Where =~ s/(like \?)/$1 $collation $esc/gi;
    my $FromParameter = $self->get('ParameterValues');
    push @$FromParameter, @ParameterValues;
    $self->set('ParameterValues', $FromParameter);
    return $Where;
}

#========================================================================================
# function     sqlSort
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $OrderBy = $Object->sqlSort;
#----------------------------------------------------------------------------------------
# description  returns the OrderBy part of sql statement
#----------------------------------------------------------------------------------------
# return       $OrderBy | part of sql statement | string
#========================================================================================
sub sqlSort {
    my $self = shift;
    my $OrderBy   = $self->orderBy;
    my $OrderDesc = $self->orderDesc;

    # If no sort by given dont sort.
    return '' if not defined $OrderBy or $OrderBy eq '';

    my $SearchDefinition = $self->get('SearchDefinition');
    my ($hSortParameter) = grep { $_->{'Parameter'} eq $OrderBy }
                           @{$SearchDefinition->get('SortParameters')};
    if (not defined $hSortParameter) {
        Error('NoSort', {
            'ClassPath' => $SearchDefinition->get('ResultClass')->pathString,
            'SearchDefinition'=> $SearchDefinition->alias,
            'OrderBy' => $OrderBy,
        });
    }

    my $OrderByPart = $hSortParameter->{'OrderByPart'};
    my $IgnoreCase = $OrderBy =~ /(name|alias)/i ? 1 : 0;
    return DAL_GetSQLSort( $OrderByPart, $OrderDesc, $IgnoreCase );
}

#========================================================================================
# function     sqlSortMax
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $OrderBy = $Object->sqlSortMax();
#----------------------------------------------------------------------------------------
# description  returns the OrderBy MAX() part of sql statement
#----------------------------------------------------------------------------------------
# return       $OrderBy | part of sql statement | string
#========================================================================================
sub sqlSortMax {
    my $self = shift;
    my $OrderBy   = $self->orderBy;
    my $OrderDesc = $self->orderDesc;

    # If no sort by given dont sort.
    return '' if not defined $OrderBy or $OrderBy eq '';

    my $SearchDefinition = $self->get('SearchDefinition');
    my ($hSortParameter) = grep { $_->{'Parameter'} eq $OrderBy }
                           @{$SearchDefinition->get('SortParameters')};
    if (not defined $hSortParameter) {
        Error('NoSort', {
            'ClassPath' => $SearchDefinition->get('ResultClass')->pathString,
            'SearchDefinition'=> $SearchDefinition->alias,
            'OrderBy' => $OrderBy,
        });
    }

    my $OrderByPart = $hSortParameter->{'OrderByPart'};
    my $IgnoreCase = $OrderBy =~ /(name|alias)/i ? 1 : 0;
    return DAL_GetSQLSortMax( $OrderByPart, $OrderDesc, $IgnoreCase );
}

#========================================================================================
# function     sqlSelect
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $Select = $Object->sqlSelect;
#----------------------------------------------------------------------------------------
# description  returns the select part of the sql statement for result columns
#----------------------------------------------------------------------------------------
# return       $Select | sql select statement | string
#========================================================================================
sub sqlSelect {
    my $self = shift;

    my $Master = $self->masterFrom;
    Error('NoMaster', {
        'ClassPath'       => $self->get('SearchDefinition')->get('ResultClass')->pathString,
        'SearchDefinition'=> $self->get('SearchDefinition')->alias,
    }) unless defined $Master;
    return sprintf( DAL_GetSQLSelect() , @$Master{'TableShortCut', 'PrimaryKey'});
}

#========================================================================================
# function     sqlSimpleSelect
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $Select = $Object->sqlSimpleSelect();
#----------------------------------------------------------------------------------------
# description  returns the select part of the sql statement for result columns without distinct
#----------------------------------------------------------------------------------------
# return       $Select | sql select statement | string
#========================================================================================
sub sqlSimpleSelect {
    my $self = shift;

    my $Master = $self->masterFrom;
    Error('NoMaster', {
        'ClassPath'       => $self->get('SearchDefinition')->get('ResultClass')->pathString,
        'SearchDefinition'=> $self->get('SearchDefinition')->alias,
    }) unless defined $Master;
    return sprintf( DAL_GetSQLSimpleSelect() , @$Master{'TableShortCut', 'PrimaryKey'});
}


#========================================================================================
# function     sqlGroup
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $Group = $Object->sqlGroup();
#----------------------------------------------------------------------------------------
# description  returns the group part of the sql statement for result columns
#----------------------------------------------------------------------------------------
# return       $Group | sql group statement | string
#========================================================================================
sub sqlGroup {
    my $self = shift;

    my $Master = $self->masterFrom;
    Error('NoMaster', {
        'ClassPath'       => $self->get('SearchDefinition')->get('ResultClass')->pathString,
        'SearchDefinition'=> $self->get('SearchDefinition')->alias,
    }) unless defined $Master;
    return sprintf( DAL_GetSQLGroup() , @$Master{'TableShortCut', 'PrimaryKey'});
}

#========================================================================================
# function     countItems
#----------------------------------------------------------------------------------------
# state        public
# syntax       $Object->countItems;
#----------------------------------------------------------------------------------------
# description  returns how many items should be paged.
#----------------------------------------------------------------------------------------
# return       $cItems | count items | integer
#========================================================================================
sub countItems {
    my $self = shift;
    return $self->{'ItemCount'};
}

#========================================================================================
# function     itemIDs
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $aItems = $Object->items($PageNumber);
#----------------------------------------------------------------------------------------
# description  Returns the ObjectIDs of all objects of a specified page.
#               Returns all ObjectIDs if $PageNumber is 0.
#----------------------------------------------------------------------------------------
# input        $PageNumber | current page number | integer
# return       $aItemIDs | list of object ids | ref.array.int
#========================================================================================
sub itemIDs {
    my $self = shift;
    my ($PageNumber) = @_;
    TestInteger('PageNumber',$PageNumber);

    $self->execute($PageNumber) unless ( defined $self->{'CurrentPage'} && $self->{'CurrentPage'} == $PageNumber );
    return $self->{'Items'};
}

#========================================================================================
# function     items
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $aItems = $Object->items($PageNumber);
#----------------------------------------------------------------------------------------
# description  Returns all objects of a specified page.
#               Returns all items if $PageNumber is 0.
#----------------------------------------------------------------------------------------
# input        $PageNumber | current page number | integer
# return       $aItems | array of objects | ref.array.object
#========================================================================================
sub items {
    my $self = shift;
    my ($PageNumber) = @_;
    TestInteger('PageNumber',$PageNumber);

    my $Class = $self->get('SearchDefinition')->get('ResultClass');
    return [
        map { $Class->loadObject($_) }
        @{ $self->itemIDs($PageNumber) }
    ];
}

#========================================================================================
# function     sqlLimit
#----------------------------------------------------------------------------------------
# state        public
# syntax       my $Limit = $Object->sqlLimit($PageNumber);
#----------------------------------------------------------------------------------------
# description  returns the limit part of the sql statement for result columns.
#               Returns '' if $PageNumber is 0.
#----------------------------------------------------------------------------------------
# input        $PageNumber | page number | string
# return       $Limit | sql limit statement | string
#========================================================================================
sub sqlLimit {
    my $self = shift;
    my ($PageNumber) = @_;
    TestInteger( 'PageNumber', $PageNumber );

    return '' if $PageNumber == 0;

    my $Offset = $self->firstItemOfPage($PageNumber);
    return DAL_GetSQLLimit($Offset, $self->get('PageSize'));
}

sub _getSqlQuery {
    my $self = shift;
    my ($PageNumber) = @_;

    if ($self->orderBy eq 'LastUpdated') {
        
        return join("\n",
            $self->sqlSimpleSelect(),
            $self->sqlFrom(),
            $self->sqlWhere(),
            $self->sqlGroup(),
            $self->sqlSortMax(),
            $self->sqlLimit($PageNumber)
        );
    }

    return join("\n",
            $self->sqlSelect(),
            $self->sqlFrom(),
            $self->sqlWhere(),
            $self->sqlSort(),
            $self->sqlLimit($PageNumber)
        );
}

#========================================================================================
# function     _ReplaceParameter
#----------------------------------------------------------------------------------------
# state        private
# syntax       _ReplaceParameter($Parameter, $hValues, $aValues);
#----------------------------------------------------------------------------------------
# description  used for replace #Parameter at where clauses with ? placeholder
#----------------------------------------------------------------------------------------
# input        $Parameter | name of parameter | string
# input        $hValues | hash of parameter | ref.hash
# input        $aValues | array of values for sql statement | ref.array.*
# return       $Parameter | parameter for where clause | string
#========================================================================================
sub _ReplaceParameter {
    my ($Parameter, $hValues, $aValues) = @_;
    return "#$Parameter" unless defined $hValues->{$Parameter};
    if (ref $hValues->{$Parameter} eq 'ARRAY') {
        push @$aValues, @{$hValues->{$Parameter}};
        return join ',', map { '?' } @{$hValues->{$Parameter}};
    }
    push @$aValues, $hValues->{$Parameter};
    return '?';
}

1;
