#========================================================================================
# package      DE_EPAGES::Search::DAL::PagerSearch
# state        public
# base         DE_EPAGES::Presentation::API::Object::Pager
# description  object interface for Search.
#========================================================================================
package DE_EPAGES::Search::DAL::PagerSearch;
use base qw (Exporter);
use strict;

our @EXPORT_OK = qw (
    DAL_ExecutePagerSearch
    DAL_GetSQLSelect
    DAL_GetSQLSimpleSelect
    DAL_GetItems
    DAL_GetSQLLimit
    DAL_GetSQLGroup
    DAL_GetSQLSort
    DAL_GetSQLSortMax
);

use DE_EPAGES::Core::API::Log qw (GetLog );
use DE_EPAGES::Core::API::PerlTools qw ( UsePackage );
use DE_EPAGES::Database::API::Connection qw (GetCurrentDBHandle);

my $DO_CACHE = 1;

sub _CallDriver {
    my ($FunctionName, @Parameter) = @_;
    my $dbi = GetCurrentDBHandle();
    my $Package = 'DE_EPAGES::Search::DAL::'.$dbi->driverName.'::PagerSearch';
    UsePackage($Package);
    GetLog->debug("call $Package->$FunctionName()");
    return $Package->$FunctionName($dbi, @Parameter);
}

#========================================================================================
# function     DAL_ExecutePagerSearch
#----------------------------------------------------------------------------------------
# syntax       $aValues = DAL_ExecutePagerSearch($PageNumber, $SQL, $ParameterValues);
#----------------------------------------------------------------------------------------
# description  Queries results for the given SQL command and stores the results in an
#               internal array. The resultset depends on the current page which the user
#               is viewing. Only the results for this page are queries (at least for MySQL)
#----------------------------------------------------------------------------------------
# input        $PageNumber | the current page, starting with 1 | integer
# input        $SQL | The SQL query | string
# input        $ParameterValues | parameters for the SQL query | ref.array
# return       $hValues | with following keys:
#               <ul>
#                  <li>CurrentPage   - current displayed page (starting with 1) - int
#                  <li>Items         - result item ids - ref.array
#                  <li>ItemCount     - number of result items - int
#               </ul> | reference to hash
#========================================================================================
sub DAL_ExecutePagerSearch {
    my ($PageNumber, $SQL, $ParameterValues) = @_;
    return _CallDriver('ExecutePagerSearch', ($PageNumber, $SQL, $ParameterValues));
}

#========================================================================================
# function     DAL_GetSQLSelect
#----------------------------------------------------------------------------------------
# syntax       $SQL = DAL_GetSQLSelect();
#----------------------------------------------------------------------------------------
# description  Returns the SQL select statement for PagerSearch depending on the
#               current database system
#----------------------------------------------------------------------------------------
# return       $SQL | the SQL query | string
#========================================================================================
sub DAL_GetSQLSelect {
    return _CallDriver('GetSQLSelect');
}

#========================================================================================
# function     DAL_GetSQLSimpleSelect
#----------------------------------------------------------------------------------------
# syntax       $SQL = DAL_GetSQLSimpleSelect();
#----------------------------------------------------------------------------------------
# description  Returns the SQL select statement without DISTINCT for PagerSearch depending on the
#               current database system
#----------------------------------------------------------------------------------------
# return       $SQL | the SQL query | string
#========================================================================================
sub DAL_GetSQLSimpleSelect {
    return _CallDriver('GetSQLSimpleSelect');
}

#========================================================================================
# function     DAL_GetSQLLimit
#----------------------------------------------------------------------------------------
# syntax       $SQL = DAL_GetSQLLimit($Offset, $PageSize);
#----------------------------------------------------------------------------------------
# description  Returns the SQL limit statement depending on the current database system
#----------------------------------------------------------------------------------------
# input        $Offset | the number of the first item of the page in the result array | integer
# input        $LastItemOfPage | the number of items which are displayed on one page | integer
# return       $SQL | the SQL limit statement | string
#========================================================================================
sub DAL_GetSQLLimit {
    my ($Offset, $PageSize) = @_;
    return _CallDriver('GetSQLLimit', ($Offset, $PageSize));
}

#========================================================================================
# function     DAL_GetSQLGroup
#----------------------------------------------------------------------------------------
# syntax       $SQL = DAL_GetSQLGroup();
#----------------------------------------------------------------------------------------
# description  Returns the SQL GROUP BY statement for PagerSearch depending on the
#               current database system
#----------------------------------------------------------------------------------------
# return       $SQL | the SQL query | string
#========================================================================================
sub DAL_GetSQLGroup {
    return _CallDriver('GetSQLGroup');
}

#========================================================================================
# function     DAL_GetSQLSort
#----------------------------------------------------------------------------------------
# syntax       $SQL = DAL_GetSQLSort($OrderByPart, $OrderDesc, $IgnoreCase);
#----------------------------------------------------------------------------------------
# description  Returns the ORDER BY clause of the SQL statement
#----------------------------------------------------------------------------------------
# input        $OrderByPart | sort criteria | string
# input        $OrderDesc | order descending if true, otherwise ascending | boolean
# input        $IgnoreCase | don't sort case sensitive if true | boolean
# return       $SQL | partial SQL statement | string
#========================================================================================
sub DAL_GetSQLSort {
    my ($OrderByPart, $OrderDesc, $IgnoreCase) = @_;
    return _CallDriver('GetSQLSort', ($OrderByPart, $OrderDesc, $IgnoreCase));
}

#========================================================================================
# function     DAL_GetSQLSortMax
#----------------------------------------------------------------------------------------
# syntax       $SQL = DAL_GetSQLSortMax($OrderByPart, $OrderDesc, $IgnoreCase);
#----------------------------------------------------------------------------------------
# description  Returns the ORDER BY MAX(...) clause of the SQL statement specific for dates
#----------------------------------------------------------------------------------------
# input        $OrderByPart | sort criteria | string
# input        $OrderDesc | order descending if true, otherwise ascending | boolean
# input        $IgnoreCase | don't sort case sensitive if true | boolean
# return       $SQL | partial SQL statement | string
#========================================================================================
sub DAL_GetSQLSortMax {
    my ($OrderByPart, $OrderDesc, $IgnoreCase) = @_;
    return _CallDriver('GetSQLSortMax', ($OrderByPart, $OrderDesc, $IgnoreCase));
}

1;
