#==============================================================================
# package      DE_EPAGES::Tax::UI::TaxMatrix
# state        private
#==============================================================================
package DE_EPAGES::Tax::UI::TaxMatrix;
use base DE_EPAGES::Tax::UI::Shop;

use strict;
use DE_EPAGES::Address::API::Address qw( CheckZipcode CompareZipCodes );
use DE_EPAGES::Core::API::Error qw ( Error GetError ExistsError );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadRootObject );
use DE_EPAGES::Tax::API::Table::TaxAreaCountry qw( GetInfoTaxAreaCountry ExistsTaxAreaCountry );
use DE_EPAGES::Tax::API::Table::TaxRate qw ( UpdateTaxRate );## no critic qw(RequireCorrectImports)
use DE_EPAGES::Tax::API::Table::TaxRateDigital qw ( UpdateTaxRateDigital );## no critic qw(RequireCorrectImports)
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

#==============================================================================
# Event handlers
#==============================================================================

sub Save {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;

    # use a postfix for having the normal or digital tax matrix
    my $Postfix = $Matrix->instanceOf('TaxMatrixDigital') ? 'Digital' : '';

    my $aTaxClasses = $Matrix->get('TaxClasses'.$Postfix);
    my $aTaxAreas   = $Matrix->get('TaxAreas'.$Postfix);
    my $hForm = $Servlet->form->form($Matrix, 'SaveTaxMatrix'.$Postfix);

    # save matrix or taxmatrix digital
    my @TaxRates = map { $_->{'TaxRate'.$Postfix} } @{$hForm->{'TaxRates'.$Postfix}};
    foreach my $TaxArea (@$aTaxAreas) {
        foreach my $TaxClass (@$aTaxClasses) {
            my $TaxRate = shift @TaxRates;
            ## no critic qw(ProhibitNoStrict)
            no strict 'refs';
            # call UpdateTaxRate or UpdateTaxRateDigital
            &{"UpdateTaxRate".$Postfix}(({
                'TaxArea'.$Postfix.'ID'  => $TaxArea->id,
                'TaxClass'.$Postfix.'ID' => $TaxClass->id,
                'Rate'=> $TaxRate / 100.00 ## no critic qw( ProhibitMagicNumbers )
            }));
            use strict 'refs';
            ## use critic
        }
    }
    $Matrix->getSite->updatePageCache;
    return;
}

sub UnitySaveTaxMatrix {
    my $self = shift;
    my ($Servlet) = @_;
    my $Form = $Servlet->form;

    # check if we can reset the tax matrix
    if ($Form->value('ResetTaxMatrix')) {
        my $Matrix = $Servlet->object;
        my $SystemTaxMatrix = LoadRootObject()->child( $Matrix->alias );

        # throw error in case the tax matrix has been customized
        if($self->NewTaxAreaAdded($SystemTaxMatrix, $Matrix) || $self->NewTaxClassAdded($SystemTaxMatrix, $Matrix)) {

            # unset reset flag
            $Servlet->params('ResetTaxMatrix', 0);

            $Form->executeFormError({
                'Reason'=> 'CannotResetTaxMatrix',
                'Form'  => 'SaveTaxMatrix',
                'Name'  => 'ResetTaxMatrix'
            });
        }
    }

    $self->Save($Servlet);
}

sub NewTaxArea {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix   = $Servlet->object;
    my $hForm = $Servlet->form->form($Matrix, 'NewTaxArea');

    my $TaxArea = undef;
    if (! $Matrix->existsChild( $hForm->{'NewAlias'} ) ){
        $TaxArea = $Matrix->insertChild('TaxArea',{'Alias'=>$hForm->{'NewAlias'}});
    } else{
        $Servlet->vars('TAXAREA_EXISTS',1);
            $Servlet->form->executeFormError({
            'Reason'=> 'TAXAREA_EXISTS',
            'Form'  => 'NewTaxArea',
            'Name'  => 'NewAlias',
            'ItemNo'=> 0,
            'Value' => $hForm->{'NewAlias'},
            }) ;
        return;
    }

    foreach my $LanguageEntry (@{$hForm->{'Languages'}}) {
        my %Info;
        $Info{$_} = $LanguageEntry->{'New'.$_} foreach map { substr($_, 3) } keys %$LanguageEntry;
        delete $Info{'LanguageID'};
        $TaxArea->set(\%Info, $LanguageEntry->{'NewLanguageID'});
    }
    return;
}

sub NewTaxAreaUnity {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;
    my $hForm = $Servlet->form->form($Matrix, 'NewTaxAreaUnity');

    my $TaxArea = undef;

    my $NewAlias = $Servlet->params('NewName');
    my $TestAlias = $NewAlias;
    my $Counter = 1;

    while ($Matrix->existsChild($TestAlias)) {
        $TestAlias = $NewAlias . '-' . $Counter;
        $Counter++;
    }

    $NewAlias = $TestAlias;
    $TaxArea = $Matrix->insertChild('TaxArea', {'Alias' => $NewAlias, 'Position' => $Matrix->nextPosition()});

    foreach my $LanguageEntry (@{$hForm->{'Languages'}}) {
        my %Info;
        $Info{$_} = $LanguageEntry->{'New' . $_} foreach map { substr($_, 3) } keys %$LanguageEntry;
        delete $Info{'LanguageID'};
        $TaxArea->set(\%Info, $LanguageEntry->{'NewLanguageID'});
    }

    return;
}

sub SaveTaxAreas {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;
    my $Shop   = $Matrix->getSite();

    NewTaxArea($self,$Servlet) if $Servlet->params('NewAlias') ne '';

    my $Form   = $Servlet->form->form($Matrix, 'SaveTaxAreas');

    my $DefaultTaxArea = $Shop->loadObject( $Servlet->params('DefaultObject'), 'TaxArea' );
    $Servlet->testPermission( $DefaultTaxArea, 'Delete' );
    $DefaultTaxArea->get('TaxMatrix')->set({ 'DefaultTaxArea' => $DefaultTaxArea });

    foreach my $LanguageEntry (@{$Form->{'Languages'}}) {
        my %Info;
        $Info{$_} = $LanguageEntry->{$_} foreach keys %$LanguageEntry;
        delete $Info{'LanguageID'};
        delete $Info{'TaxAreaID'};

        my $TaxArea = $Shop->loadObject($LanguageEntry->{'TaxAreaID'}, 'TaxArea');
        $Servlet->testPermission( $TaxArea, 'Delete' );

        $TaxArea->set(\%Info, $LanguageEntry->{'LanguageID'});
    }

    $Matrix->getSite->updatePageCache;
    return;
}

sub SaveTaxAreaCountries {
    my $self = shift;
    my ($Servlet) = @_;
    my $Form = $Servlet->form;

    my $Matrix = $Servlet->object;

    my $hForm = $Form->form($Matrix, 'SaveTaxAreaCountries');

    # change fallback
    my $FallbackTaxArea = $Matrix->getSite()->loadObject( $Servlet->params('FallbackObject'), 'TaxArea' );
    $Servlet->testPermission( $FallbackTaxArea, 'Delete' );
    my $ChangeFallback = ($Matrix->fallbackTaxArea->id != $FallbackTaxArea->id) ? 1 : 0;
    $FallbackTaxArea->get('TaxMatrix')->set({ 'FallbackTaxArea' => $FallbackTaxArea }) if $ChangeFallback;

    # save taxarea for existing countries
    $self->_saveTaxAreaForCountries($Form, $Matrix, $ChangeFallback, $hForm->{'Countries'});

    $self->_saveTaxAreaCountries($Form, $Matrix, $hForm);

    $Form->executeFormError();

    $Matrix->updatePageCache;
    return;
}

sub _saveTaxAreaForCountries {
    my $self = shift;
    my ($Form, $Matrix, $ChangeFallback, $aCountries) = @_;

    my $Shop = $Matrix->getSite();

    foreach my $hCountryEntry (@$aCountries){
        my $CountryID = $hCountryEntry->{'CountryID'};
        my $TaxAreaID = $hCountryEntry->{'TaxArea'};
        my $TaxArea = $Shop->loadObject($TaxAreaID, 'TaxArea');
        my $OldTaxArea = $Matrix->taxAreaByCountry($CountryID);
        if ($OldTaxArea->id != $TaxAreaID) {                                      # TaxArea has changed
            $OldTaxArea->removeCountry($CountryID);
            $TaxArea->addCountry($CountryID) unless $TaxArea->get('IsFallback');
        } elsif ($ChangeFallback && !$TaxArea->existsCountry($CountryID)) {
            $TaxArea->addCountry($CountryID) unless $TaxArea->get('IsFallback');  # add to TaxArea
        }
    }
}

sub _saveTaxAreaCountries {
    my $self = shift;
    my ($Form, $Matrix, $hForm) = @_;

    my $aTaxAreaCountries = $hForm->{'TaxAreaCountries'};
    my $Shop = $Matrix->getSite();

    # check zipcodes
    my %TaxAreaCountriesPerCountry;
    my $Index = -1;
    foreach my $TaxAreaCountryEntry (@$aTaxAreaCountries) {
        $Index++;
        my $CountryID = $TaxAreaCountryEntry->{'BaseCountryID'};
        my $TaxAreaCountryID = $TaxAreaCountryEntry->{'TaxAreaCountryID'};
        my $TaxAreaID = $TaxAreaCountryEntry->{'TaxAreaCountryTaxArea'};
        my $ZipcodeFrom = $TaxAreaCountryEntry->{'ZipcodeFrom'};
        my $ZipcodeTo = $TaxAreaCountryEntry->{'ZipcodeTo'};
        my $RegionName = $TaxAreaCountryEntry->{'RegionName'};

        if (!defined $ZipcodeFrom || $ZipcodeFrom eq '') {
            $Form->addFormError({
                Reason => 'MissingZipcodeFrom',
                Index => $Index,
                LoopName => 'TaxAreaCountries',
                Form => 'SaveTaxAreaCountries',
                Name => 'ZipcodeFrom',
            } );
            next;
        }
        $ZipcodeTo = $ZipcodeFrom if !defined $ZipcodeTo || $ZipcodeTo eq '';

        if ( CompareZipCodes($ZipcodeFrom,$ZipcodeTo) == 1 ) {  #$ZipcodeFrom > $ZipcodeTo
            my $tmp = $ZipcodeFrom;
            $ZipcodeFrom = $ZipcodeTo;
            $ZipcodeTo = $tmp;
        }

        if (!exists($TaxAreaCountriesPerCountry{$CountryID})) {
            $TaxAreaCountriesPerCountry{$CountryID} = [];
        }

        next unless _checkZipcodeValues($Form, $ZipcodeFrom, $ZipcodeTo, $Index, $CountryID
         , $TaxAreaCountriesPerCountry{$CountryID}, {'ZipcodeFrom' => 'ZipcodeFrom', 'ZipcodeTo' => 'ZipcodeTo'} );

        push @{$TaxAreaCountriesPerCountry{$CountryID}}, {
            'TaxAreaCountryID' => $TaxAreaCountryID,
            'TaxAreaID' => $TaxAreaID,
            'ZipcodeFrom' => $ZipcodeFrom,
            'ZipcodeTo' => $ZipcodeTo,
            'RegionName' => $RegionName,
        };
    }

    #save the newly added zipcodes
    $aTaxAreaCountries = $hForm->{'NewTaxAreaCountries'};
    $Index = -1;
    foreach my $TaxAreaCountryEntry (@$aTaxAreaCountries) {
        $Index++;
        my $CountryID = $TaxAreaCountryEntry->{'ZipcodeCountryID'};
        my $TaxAreaID = $TaxAreaCountryEntry->{'NewTaxArea'};
        my $ZipcodeFrom = $TaxAreaCountryEntry->{'NewZipcodeFrom'};
        my $ZipcodeTo = $TaxAreaCountryEntry->{'NewZipcodeTo'};
        my $RegionName = $TaxAreaCountryEntry->{'NewRegionName'};

        $ZipcodeTo = $ZipcodeFrom if !defined $ZipcodeTo || $ZipcodeTo eq '';

        next if (!defined $ZipcodeFrom || $ZipcodeFrom eq '');

        if ( CompareZipCodes($ZipcodeFrom,$ZipcodeTo) == 1 ) {  #$ZipcodeFrom > $ZipcodeTo
            my $tmp = $ZipcodeFrom;
            $ZipcodeFrom = $ZipcodeTo;
            $ZipcodeTo = $tmp;
        }

        if (!exists($TaxAreaCountriesPerCountry{$CountryID})) {
            $TaxAreaCountriesPerCountry{$CountryID} = [];
        }

        next unless _checkZipcodeValues($Form, $ZipcodeFrom, $ZipcodeTo, $Index, $CountryID
         , $TaxAreaCountriesPerCountry{$CountryID}, {'ZipcodeFrom' => 'NewZipcodeFrom', 'ZipcodeTo' => 'NewZipcodeTo'} );

        my $TaxArea = $Shop->loadObject($TaxAreaID, 'TaxArea');
        $TaxArea->addCountry($CountryID, $ZipcodeFrom, $ZipcodeTo, $RegionName) if $ZipcodeFrom && $ZipcodeTo;  # add to TaxArea
    }

    $Form->executeFormError();

    #save taxareacountries
    foreach my $hTaxAreaCountries (values %TaxAreaCountriesPerCountry) {
        foreach my $hTaxAreaCountry (@$hTaxAreaCountries) {
            my $TaxAreaCountryID = $hTaxAreaCountry->{'TaxAreaCountryID'};
            delete $hTaxAreaCountry->{'TaxAreaCountryID'};
            $Matrix->updateTaxAreaCountry($TaxAreaCountryID, $hTaxAreaCountry);
        }
    }

}

sub _checkZipcodeValues {
    my ($Form, $ZipcodeFrom, $ZipcodeTo, $Index, $CountryID, $ahTaxAreaCountries, $hKeyNames) = @_;

    #check the zipcodes validity
    if (!CheckZipcode($ZipcodeFrom, $CountryID) || !CheckZipcode($ZipcodeTo, $CountryID)) {
        $Form->addFormError({
            Reason => 'ZIPCODE_NOT_VALID',
            Index => $Index,
            LoopName => 'TaxAreaCountries',
            Form => 'SaveTaxAreaCountries',
            Name => $hKeyNames->{'ZipcodeFrom'},
        });
        $Form->addFormError({
            Reason => 'ZIPCODE_NOT_VALID',
            Index => $Index,
            LoopName => 'TaxAreaCountries',
            Form => 'SaveTaxAreaCountries',
            Name => $hKeyNames->{'ZipcodeTo'},
        });
        return 0;
    }




    # look after overlapping with other codes
    my $Overlapps = 0;
    foreach my $OtherTaxAreaCountryEntry (@$ahTaxAreaCountries) {

        my $OtherZipcodeFrom = $OtherTaxAreaCountryEntry->{'ZipcodeFrom'};
        my $OtherZipcodeTo = $OtherTaxAreaCountryEntry->{'ZipcodeTo'};

        if ( (CompareZipCodes($ZipcodeFrom,$OtherZipcodeFrom) != -1)&& (CompareZipCodes($ZipcodeTo,$OtherZipcodeTo) != 1) ||
             (CompareZipCodes($ZipcodeFrom,$OtherZipcodeFrom) != 1) && (CompareZipCodes($OtherZipcodeFrom,$ZipcodeTo) != 1) ||
             (CompareZipCodes($ZipcodeFrom,$OtherZipcodeTo) != 1)   && (CompareZipCodes($OtherZipcodeTo,$ZipcodeTo) != 1) ){
             $Overlapps = 1;
        }

    }

    if ($Overlapps) {
        $Form->addFormError({
            Reason  => 'ExistingZipcode',
            Index    => $Index,
            LoopName => 'TaxAreaCountries',
            Form => 'SaveTaxAreaCountries',
            Name    => $hKeyNames->{'ZipcodeFrom'},
        } );
        $Form->addFormError({
            Reason  => 'ExistingZipcode',
            Index    => $Index,
            LoopName => 'TaxAreaCountries',
            Form => 'SaveTaxAreaCountries',
            Name    => $hKeyNames->{'ZipcodeTo'},
        } );
        return 0;
    }

    return 1;
}

sub NewTaxClass {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix   = $Servlet->object;
    my $hForm = $Servlet->form->form($Matrix, 'NewTaxClass');

    my $TaxClass = undef;
    if (! $Matrix->existsChild( $hForm->{'NewAlias'} ) ){
        $TaxClass = $Matrix->insertChild('TaxClass',{'Alias'=>$hForm->{'NewAlias'}});
    } else{
        $Servlet->vars('TAXCLASS_EXISTS',1);
            $Servlet->form->executeFormError({
            'Reason'=> 'TAXCLASS_EXISTS',
            'Form'  => 'NewTaxClass',
            'Name'  => 'NewAlias',
            'ItemNo'=> 0,
            'Value' => $hForm->{'NewAlias'},
            }) ;
        return;
    }

    foreach my $LanguageEntry (@{$hForm->{'Languages'}}) {
        my %Info;
        $Info{$_} = $LanguageEntry->{'New'.$_} foreach map { substr($_, 3) } keys %$LanguageEntry;
        delete $Info{'LanguageID'};
        $TaxClass->set(\%Info, $LanguageEntry->{'NewLanguageID'});
    }
    return;
}

sub NewTaxClassUnity {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;
    my $hForm = $Servlet->form->form($Matrix, 'NewTaxClassUnity');

    my $TaxClass = undef;

    my $NewAlias = $Servlet->params('NewName');
    my $TestAlias = $NewAlias;
    my $Counter = 1;

    while ($Matrix->existsChild($TestAlias)) {
        $TestAlias = $NewAlias . '-' . $Counter;
        $Counter++;
    }

    $NewAlias = $TestAlias;
    $TaxClass = $Matrix->insertChild('TaxClass', {'Alias' => $NewAlias, 'Position' => $Matrix->nextPosition()});

    foreach my $LanguageEntry (@{$hForm->{'Languages'}}) {
        my %Info;
        $Info{$_} = $LanguageEntry->{'New' . $_} foreach map { substr($_, 3) } keys %$LanguageEntry;
        delete $Info{'LanguageID'};
        $TaxClass->set(\%Info, $LanguageEntry->{'NewLanguageID'});
    }

    return;
}

sub SaveTaxClasses {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;
    my $Shop   = $Matrix->getSite();

    NewTaxClass($self,$Servlet) if $Servlet->params('NewAlias') ne '';

    my $Form   = $Servlet->form->form($Matrix, 'SaveTaxClasses');

    my $DefaultTaxClass = $Shop->loadObject( $Servlet->params('DefaultObject'), 'TaxClass' );
    $Servlet->testPermission( $DefaultTaxClass, 'Delete' );
    $DefaultTaxClass->get('TaxMatrix')->set({ 'DefaultTaxClass' => $DefaultTaxClass });

    foreach my $LanguageEntry (@{$Form->{'Languages'}}) {
        my %Info;
        $Info{$_} = $LanguageEntry->{$_} foreach keys %$LanguageEntry;
        delete $Info{'LanguageID'};
        delete $Info{'TaxClassID'};

        my $TaxClass = $Shop->loadObject($LanguageEntry->{'TaxClassID'}, 'TaxClass');
        $Servlet->testPermission( $TaxClass, 'Delete' );

        $TaxClass->set(\%Info, $LanguageEntry->{'LanguageID'});
    }
    return;
}

sub DeleteTaxAreas {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;
    my $Form = $Servlet->form;
    my $Shop = $Matrix->getSite();

    #get list objects to delete; get all TaxAreas
    my $ListedTaxAreaIDs = $Servlet->form->values->{'ListObjectID'};
    my $AllTaxAreas = $Matrix->get('TaxAreas');

    $Form->executeFormError( {  Reason  => 'NoTaxAreaSelected',
                                Name    => 'ListObjectID' } ) if( not defined $ListedTaxAreaIDs);
    $Form->executeFormError( {  Reason  => 'DontDeleteAllTaxAreas',
                                Name    => 'ListObjectID' } ) if( scalar(@$AllTaxAreas) == scalar(@$ListedTaxAreaIDs) );
    # check if TaxAreas can be deleted
    my $CanDelete = 1;
    foreach my $TaxAreaID (@$ListedTaxAreaIDs) {
        my $TaxArea = $Shop->loadObject( $TaxAreaID, 'TaxArea' );
        $CanDelete = 0 if $TaxArea->get('IsDefault') == 1;
        $CanDelete = 0 if $TaxArea->get('IsFallback') == 1;
    }
    if( !$CanDelete ) {
        $Form->executeFormError( {  Reason  => 'DontDeleteTaxArea',
                                    Name    => 'ListObjectID' } );
    }
    # delete
    eval {
        $self->DeleteListedObjects( $Servlet );
    };
    if( ExistsError() ) {
        my $Error = GetError();
        my $Code = $Error->code;
        if( $Code eq 'DontDeleteTaxArea' ) {
            $Matrix->tleHash->{'ErrorMessage_DontDeleteTaxArea'} = $Error->shortMessage;
            $Form->executeFormError( {  Reason  => 'DontDeleteTaxArea',
                                        Name    => 'ListObjectID' } );
        } else {
            Error($Error);
        }
    }
    # set new default if old default was deleted
    $AllTaxAreas = $Matrix->get('TaxAreas');
    my $WithDefault = 0;
    foreach my $TaxArea (@$AllTaxAreas) {
        $WithDefault = 1 if( $TaxArea->get('IsDefault') );
    }
    if( not $WithDefault ) {
        my $NewDefaultArea = shift @$AllTaxAreas;
        $NewDefaultArea->set({IsDefault => 1});
    }
    if ($Servlet->vars('DELETE') == 0){
        $Matrix->getSite->updatePageCache;
    }
    return;
}

sub DeleteTaxClasses {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;
    my $Form = $Servlet->form;
    my $Shop = $Matrix->getSite();

    #get list objects to delete; get all TaxClasses
    my $ListedTaxClassIDs = $Servlet->form->values->{'ListObjectID'};
    my $AllTaxClasses = $Matrix->get('TaxClasses');

    $Form->executeFormError( {  Reason  => 'NoTaxClassSelected',
                                Name    => 'ListObjectID' } ) if( not defined $ListedTaxClassIDs);
    $Form->executeFormError( {  Reason  => 'DontDeleteAllTaxClasses',
                                Name    => 'ListObjectID' } ) if( scalar(@$AllTaxClasses) == scalar(@$ListedTaxClassIDs) );

    # check if TaxClasses can be deleted
    foreach my $TaxClassID (@$ListedTaxClassIDs) {
        my $TaxClass = $Shop->loadObject( $TaxClassID, 'TaxClass' );
        if ($TaxClass->get('IsDefault')) {
            $Form->addFormError({
                Reason  => 'DontDeleteDefaultTaxClass',
                Name    => 'ListObjectID',
                Value   => 'IsDefault'
                } );
        } else {
            eval { TriggerHook('UI_TestDeleteTaxClass', {'TaxClass' => $TaxClass }); };
            if( ExistsError() ) {
                my $Error = GetError();
                Error( $Error ) unless $Error->code eq 'DontDeleteTaxClass';
                my $UsedClass = $Error->vars->{'UsedClass'};
                $Error->addVars( {'DeleteTaxClass' => $TaxClass});
                $Error->addVars( {'UsedClass'=> LoadClassByAlias($UsedClass)}) if defined $UsedClass;
                $Form->mapFormError({
                    Name    => 'ListObjectID',
                } , $Error);
            }
        }
    }
    $Form->executeFormError();

    # delete
    eval {
        $self->DeleteListedObjects( $Servlet );
    };
    if( ExistsError() ) {
        my $Error = GetError();
        my $Code = $Error->code;
        if( $Code eq 'DontDeleteTaxClass' ) {
            $Form->executeFormError( {  Reason  => 'DontDeleteTaxClass',
                                        Name    => 'ListObjectID' } ) if( $Code eq 'DontDeleteTaxClass' );
        } else {
            Error($Error);
        }
    }

    # set new default if old default was deleted
    $AllTaxClasses = $Matrix->get('TaxClasses');
    my $WithDefault = 0;
    foreach my $TaxClass (@$AllTaxClasses) {
        $WithDefault = 1 if( $TaxClass->get('IsDefault') );
    }
    if( not $WithDefault ) {
        my $NewDefaultClass = shift @$AllTaxClasses;
        $NewDefaultClass->set({IsDefault => 1});
    }

    return;
}

sub DeleteTaxAreaCountry {
    my $self = shift;
    my ($Servlet) = @_;

    my $Matrix = $Servlet->object;
    my $Form = $Servlet->form;

    my $TaxAreaCountryID = $Form->value('TaxAreaCountryID');
    $Matrix->removeTaxAreaCountry($TaxAreaCountryID) if (ExistsTaxAreaCountry($TaxAreaCountryID) && GetInfoTaxAreaCountry($TaxAreaCountryID)->{'TaxMatrixID'} == $Matrix->id);
}

1;